// tridiag.h - Diagonalisation code
// Copyright (C) 2009-2015 Rok Zitko

#ifndef _diag_h_
#define _diag_h_

// LAPACK routines

#ifdef __cplusplus
extern "C"
{
#endif

  // Compute all eigenvalues and eigenvectors of a double symmetric matrix A
  void dsyev(char *jobz, char *uplo, int *n, double *a, int *lda,
	     double *w, double *work, int *lwork, int *info);

  // DSYEVR computes selected eigenvalues and, optionally,
  // eigenvectors of a real symmetric matrix T.  Eigenvalues and
  // eigenvectors can be selected by specifying either a range of
  // valu

  // Whenever possible, DSYEVR calls DSTEGR to compute the
  // eigenspectrum using Relatively Robust Representations.  DSTEGR
  // computes eigenvalues by the dqds algorithm, while orthogonal
  // eigenvectors are computed from various "good" L D L^T
  // representations (also known as Relatively Robust
  // Representations). Gram-Schmidt orthogonalization is avoided as
  // far as possible. es or a range of indices for the desired
  // eigenvalues.

  void dsyevr(char *jobz, char *range, char *uplo,
	      int *n, double *a, int *lda,
	      double *vl, double *vu, int *il, int *iu,
	      double *abstol, int *m, double *w,
	      double *z, int *ldz, int *isuppz,
	      double *work, int *lwork, int *iwork, int *liwork,
	      int *info);

  void zheev(char *jobz, char *uplo, int *n, cmpl *a, int *lda,
	     double *w, cmpl *work, int *lwork, double *rwork, int *info);

  void zheevr(char *jobz, char *range, char *uplo,
	      int *n, cmpl *a, int *lda,
	      double *vl, double *vu, int *il, int *iu,
	      double *abstol, int *m, double *w,
	      cmpl *z, int *ldz, int *isuppz,
	      cmpl *work, int *lwork, cmpl *rwork, int *lrwork, int *iwork,
	      int *liwork, int *info);
		  
   
#ifdef __cplusplus
}
#endif

bool logletter(char);

// Used by diagonalise_dsyev and diagonalise_dsyevr.
void copy_values(t_eigen *eigenvalues, EVEC &diagvalue, int M)
{
   // make sure eigenvalues are in ascending order!!
   t_eigen *ptr = adjacent_find(eigenvalues, eigenvalues+M, _1 > _2);
   if (ptr != eigenvalues+M) {
      ptrdiff_t index = ptr-eigenvalues;
      cerr << "WARNING: Values are not in ascending order "
	<< "[index=" << index << ", M=" << M << "]: "
	<< *ptr << " " << *(ptr+1) << endl;
      cerr << "This likely indicates a bug in LAPACK dsyev* routines." << endl;
   }

   diagvalue.resize(M);
   copy(eigenvalues, eigenvalues + M, diagvalue.begin());
}

// Perform diagonalisation: wrappers for LAPACK
// m: matrix to be diagonalised
// diag: eigenvalues and eigenvectors
// jobz: 'N' for values only, 'V' for values and vectors

#ifdef NRG_REAL
int diagonalise_dsyev(Matrix &m,
                      Eigen &diag,
                      char jobz = 'V')
{
  Timing t(false);

  my_assert(m.size1() >= 1 && m.size1() == m.size2()); // must be square

  const unsigned int dim = m.size1();

#ifdef NRG_MPI
   nrglog('A', "LAPACK, dim=" << dim << " (dsyev)" << " on node " << myrank);
#else
   nrglog('A', "LAPACK, dim=" << dim << " (dsyev)");
#endif   
   
  t_matel *ham = bindings::traits::matrix_storage(m);
  t_eigen eigenvalues[dim]; // eigenvalues on exit

  char UPLO = 'L'; // lower triangle of a is stored
  int NN = dim; // the order of the matrix
  int LDA = dim; // the leading dimension of the array a
  int INFO = 0; // 0 on successful exit

  int LWORK0 = -1; // length of the WORK array
  double WORK0[1];

  // Step 1: determine optimal LWORK
  dsyev(&jobz, &UPLO, &NN, ham, &LDA, (double*)eigenvalues,
	WORK0, &LWORK0, &INFO);

  my_assert(INFO == 0);

  int LWORK = int(WORK0[0]);
  nrglog('d', "LWORK=" << LWORK);
  my_assert(LWORK > 0);

  const int minLWORK = MAX(1, 3*dim-1); // cf. LAPACK 3.1 dsyev.f
  if (LWORK < minLWORK) {
     cerr << "Buggy dsyev. Fixing LWORK." << endl;
     LWORK = minLWORK;
  }

  double * WORK = new double[LWORK];
  if (!WORK) 
    my_error("Failed to allocate WORK.");

  // Step 2: perform the diagonalisation
  dsyev(&jobz, &UPLO, &NN, ham, &LDA, (double*)eigenvalues,
	WORK, &LWORK, &INFO);

  if (INFO!=0)
    my_error("eigensolver failed. INFO=%i", INFO);

  diag = Eigen(dim, dim);
  copy_values(eigenvalues, diag.value, dim);

  if (jobz == 'V') {
     for (unsigned int r = 0; r < dim; r++) {
#define IJ(i, j) (dim*(i)+(j))
        for (unsigned int j = 0; j < dim; j++)
           diag.vektor(r, j) = ham[ IJ(r, j) ];
     }
     diag.perform_checks();
  }

#ifdef NRG_MPI
   if (logletter('t'))
     cout << "Elapsed: " << t.total() << " on node " << myrank << endl;
#else
   if (logletter('t'))
     cout << "Elapsed: " << t.total() << endl;
#endif

  delete[] WORK;
   
  return dim;
}
#endif

#ifdef NRG_REAL
int diagonalise_dsyevr(Matrix &m,
                       Eigen &diag,
                       char jobz = 'V',
                       double ratio = 1.0) // reduction ratio for dsyevr
{
   Timing t(false);

   my_assert(m.size1() >= 1 && m.size1() == m.size2());// must be square

   const unsigned int dim = m.size1();

   // M is the number of the eigenvalues that we will attempt to calculate
   // using dsyevr.
   unsigned int M = dim;

   char RANGE; // 'A'=all, 'V'=interval, 'I'=part
   if (ratio != 1.0) {
      M = static_cast<unsigned int>(ceil(ratio * M)); // round up
      M = CLIP(M, 1, dim); // at least 1, at most dim
      RANGE = 'I';
   } else {
      RANGE = 'A';
   }

#ifdef NRG_MPI
   nrglog('A', "LAPACK, dim=" << dim << " (dsyevr)" << " on node " << myrank);
#else
   nrglog('A', "LAPACK, dim=" << dim << " (dsyevr)");
#endif

   t_matel *ham = bindings::traits::matrix_storage(m);
   t_eigen eigenvalues[dim]; // eigenvalues on exit

   char UPLO = 'L'; // lower triangle of a is stored
   int NN = dim; // the order of the matrix
   int LDA = dim; // the leading dimension of the array a
   int INFO = 0; // 0 on successful exit

   double VL = 0; // value range; not referenced if RANGE != 'V'
   double VU = 0;
   int IL = 1; // index range
   int IU = M;
   double ABSTOL = 0;
   // If ABSTOL=0, EPS*|T| where |T| is the 1-norm of the tridiagonal
   // matrix obtained by reducing m to tridiagonal form.

   int MM; // total number of eigenvalues found
   int LDZ = dim;

   int ISUPPZ[2 * M];
   //  The support of the eigenvectors in Z, i.e., the indices
   //  indicating the nonzero elements in Z.  The i-th eigenvector is
   //  nonzero only in elements ISUPPZ( 2*i-1 ) through ISUPPZ(2*i).

   std::vector<t_matel> Z(LDZ * M); // eigenvectors

   int LWORK0 = -1;
   int LIWORK0 = -1;
   double WORK0[1];
   int IWORK0[1];

   // Step 1: determine optimal LWORK and LIWORK
   dsyevr(&jobz, &RANGE, &UPLO, &NN, ham, &LDA,
	  &VL, &VU, &IL, &IU, &ABSTOL, &MM,
	  (double*)eigenvalues, &Z[0], &LDZ, ISUPPZ,
	  WORK0, &LWORK0, IWORK0, &LIWORK0, &INFO);
   
   my_assert(INFO == 0);

   int LWORK = int(WORK0[0]);
   int LIWORK = IWORK0[0];
   nrglog('d', "LWORK=" << LWORK << " LIWORK=" << LIWORK);
   my_assert(LWORK > 0 && LIWORK > 0);

   const int minLWORK = MAX(1, 26*dim); // cf. LAPACK 3.1 dsyevr.f
   if (LWORK < minLWORK) {
      cerr << "Buggy dsyevr. Fixing LWORK." << endl;
      LWORK = minLWORK;
   }

   const int minLIWORK = MAX(1, 10*dim); // cf. LAPACK 3.1 dsyev.f
   if (LIWORK < minLIWORK) {
      cerr << "Buggy dsyevr. Fixing LIWORK." << endl;
      LIWORK = minLIWORK;
   }

   double * WORK = new double[LWORK];
   if (!WORK)
     my_error("Failed to allocate WORK");

   int * IWORK = new int[LIWORK];
   if (!IWORK)
     my_error("Failed to allocate IWORK");

   // Step 2: perform the diagonalisation
   dsyevr(&jobz, &RANGE, &UPLO, &NN, ham, &LDA,
	  &VL, &VU, &IL, &IU, &ABSTOL, &MM,
	  (double*)eigenvalues, &Z[0], &LDZ, ISUPPZ,
	  WORK, &LWORK, IWORK, &LIWORK, &INFO);

   if (INFO!=0)
     my_error("eigensolver failed. INFO=%i", INFO);

   if (MM != int(M)) {
      cerr << "dsyevr computed " << MM << "/" << M << endl;
      M = MM;
      my_assert(M > 0); // at least one
      my_assert(M <= dim); // huh?
   }

   diag = Eigen(M, dim);
   copy_values(eigenvalues, diag.value, M);

   if (jobz == 'V') {
      for (unsigned int r = 0; r < M; r++) {
#define IJZ(i, j) (dim*(i)+(j))
         for (unsigned int j = 0; j < dim; j++)
            diag.vektor(r, j) = Z[ IJZ(r, j) ];
      }
      diag.perform_checks();
   }

#ifdef NRG_MPI
   if (logletter('t'))
     cout << "Elapsed: " << t.total() << " on node " << myrank << endl;
#else
   if (logletter('t'))
     cout << "Elapsed: " << t.total() << endl;
#endif

   delete[] WORK;
   delete[] IWORK;

   return M;
}
#endif

#ifdef NRG_COMPLEX
int diagonalise_zheev(Matrix &m,
                      Eigen &diag,
                      char jobz = 'V')
{
  Timing t(false);

  my_assert(m.size1() >= 1 && m.size1() == m.size2()); // must be square

  const unsigned int dim = m.size1();

#ifdef NRG_MPI
  nrglog('A', "LAPACK, dim=" << dim << " (zheev)" << " on node " << myrank);
#else
  nrglog('A', "LAPACK, dim=" << dim << " (zheev)");
#endif
   
  t_matel *ham = bindings::traits::matrix_storage(m);
  t_eigen eigenvalues[dim]; // eigenvalues on exit

  char UPLO = 'L'; // lower triangle of a is stored
  int NN = dim; // the order of the matrix
  int LDA = dim; // the leading dimension of the array a
  int INFO = 0; // 0 on successful exit

  int LWORK0 = -1; // length of the WORK array (-1 == query!)
  cmpl WORK0[1];
  
  int RWORKdim = MAX(1, 3*dim-2);
  double RWORK[RWORKdim];

  // Step 1: determine optimal LWORK
  zheev(&jobz, &UPLO, &NN, ham, &LDA, (double*)eigenvalues,
	WORK0, &LWORK0, RWORK, &INFO);

  my_assert(INFO == 0);

  int LWORK = int(WORK0[0].real());
  nrglog('d', "LWORK=" << LWORK);
  my_assert(LWORK > 0);

  const int minLWORK = MAX(1, 2*dim-1); // cf. LAPACK 3.2 zheev.f
  if (LWORK < minLWORK) {
     cerr << "Buggy zheev. Fixing LWORK." << endl;
     LWORK = minLWORK;
  }

  cmpl * WORK = new cmpl[LWORK];
  if (!WORK) 
    my_error("Failed to allocate WORK.");

  // Step 2: perform the diagonalisation
  zheev(&jobz, &UPLO, &NN, ham, &LDA, (double*)eigenvalues,
	WORK, &LWORK, RWORK, &INFO);

  if (INFO != 0)
    my_error("eigensolver failed. INFO=%i", INFO);

  diag = Eigen(dim, dim);
  copy_values(eigenvalues, diag.value, dim);

  if (jobz == 'V') {
     for (unsigned int r = 0; r < dim; r++) {
#define IJ(i, j) (dim*(i)+(j))
        for (unsigned int j = 0; j < dim; j++)
           diag.vektor(r, j) = ham[ IJ(r, j) ];
     }
     diag.perform_checks();
  }

#ifdef NRG_MPI
   if (logletter('t'))
     cout << "Elapsed: " << t.total() << " on node " << myrank << endl;
#else
   if (logletter('t'))
     cout << "Elapsed: " << t.total() << endl;
#endif

  delete[] WORK;

  return dim;
}

int diagonalise_zheevr(Matrix &m,
                      Eigen &diag,
		      char jobz = 'V',
		      double ratio = 1.0)
{
  Timing t(false);

  my_assert(m.size1() >= 1 && m.size1() == m.size2()); // must be square

  const unsigned int dim = m.size1();

  // M is the number of the eigenvalues that we will attempt to calculate
  // using zheevr.
  unsigned int M = dim;

  char RANGE; // 'A'=all, 'V'=interval, 'I'=part
  if (ratio != 1.0) {
    M = static_cast<unsigned int>(ceil(ratio * M)); // round up
    M = CLIP(M, 1, dim); // at least 1, at most dim
    RANGE = 'I';
  } else {
    RANGE = 'A';
  }

#ifdef NRG_MPI
  nrglog('A', "LAPACK, dim=" << dim << " (zheevr)" << " on node " << myrank);
#else
  nrglog('A', "LAPACK, dim=" << dim << " (zheevr)");
#endif
   
  t_matel *ham = bindings::traits::matrix_storage(m);
  t_eigen eigenvalues[dim]; // eigenvalues on exit

  char UPLO = 'L'; // lower triangle of a is stored
  int NN = dim; // the order of the matrix
  int LDA = dim; // the leading dimension of the array a
  int INFO = 0; // 0 on successful exit

  double VL = 0; // value range; not referenced if RANGE != 'V'
  double VU = 0;
  int IL = 1; // index range
  int IU = M;
  double ABSTOL = 0;
  // If ABSTOL=0, EPS*|T| where |T| is the 1-norm of the tridiagonal
  // matrix obtained by reducing m to tridiagonal form.

  int MM; // total number of eigenvalues found
  int LDZ = dim;

  int ISUPPZ[2 * M];
  //  The support of the eigenvectors in Z, i.e., the indices
  //  indicating the nonzero elements in Z.  The i-th eigenvector is
  //  nonzero only in elements ISUPPZ( 2*i-1 ) through ISUPPZ(2*i).

  std::vector<t_matel> Z(LDZ * M); // eigenvectors

  int LWORK0 = -1; // length of the WORK array (-1 == query!)
  cmpl WORK0[1];

  int LRWORK0 = -1; // query
  cmpl RWORK0[1];

  int LIWORK0 = -1; // query
  int IWORK0[1];

  // Step 1: determine optimal LWORK, LRWORK, and LIWORK
  zheevr(&jobz, &RANGE, &UPLO, &NN, ham, &LDA,
	 &VL, &VU, &IL, &IU, &ABSTOL, &MM,
	 (double*)eigenvalues, &Z[0], &LDZ, ISUPPZ,
	 WORK0, &LWORK0, RWORK0, &LRWORK0, IWORK0, &LIWORK0, &INFO);

  my_assert(INFO == 0);

  int LWORK = int(WORK0[0].real());
  nrglog('d', "LWORK=" << LWORK);
  my_assert(LWORK > 0);

  cmpl * WORK = new cmpl[LWORK];
  if (!WORK)
    my_error("Failed to allocate WORK.");

  int LRWORK = int(RWORK0[0].real());
  nrglog('d', "LRWORK=" << LRWORK);
  my_assert(LRWORK > 0);

  cmpl * RWORK = new cmpl[LRWORK];
  if (!RWORK)
    my_error("Failed to allocate RWORK.");

  int LIWORK = IWORK0[0];
  nrglog('d', "LIWORK=" << LIWORK);
  my_assert(LIWORK > 0);

  int * IWORK = new int[LIWORK];
  if (!IWORK)
    my_error("Failed to allocate IWORK.");

  // Step 2: perform the diagonalisation
  zheevr(&jobz, &RANGE, &UPLO, &NN, ham, &LDA,
	 &VL, &VU, &IL, &IU, &ABSTOL, &MM,
	 (double*)eigenvalues, &Z[0], &LDZ, ISUPPZ,
	 WORK, &LWORK, RWORK, &LRWORK, IWORK, &LIWORK, &INFO);

  if (INFO != 0)
    my_error("eigensolver failed. INFO=%i", INFO);

  if (MM != int(M)) {
     cerr << "zheevr computed " << MM << "/" << M << endl;
     M = MM;
     my_assert(M > 0); // at least one
     my_assert(M <= dim); // huh?
  }

  diag = Eigen(M, dim);
  copy_values(eigenvalues, diag.value, M);

  if (jobz == 'V') {
#define IJZ(i, j) (dim*(i)+(j))
     for (unsigned int r = 0; r < M; r++) {
        for (unsigned int j = 0; j < dim; j++)
           diag.vektor(r, j) = Z[ IJZ(r, j) ];
     }
     diag.perform_checks();
  }

#ifdef NRG_MPI
   if (logletter('t'))
     cout << "Elapsed: " << t.total() << " on node " << myrank << endl;
#else
   if (logletter('t'))
     cout << "Elapsed: " << t.total() << endl;
#endif

  delete [] WORK;
  delete [] RWORK;
  delete [] IWORK;

  return M;
}
#endif

const double NORMALIZATION_EPSILON = 1e-12;
const double ORTHOGONALITY_EPSILON = 1e-12;

// Wrapper for the diagonalization of the Hamiltonian matrix
// in the NRG iteration
int diagonalise_h(Matrix &m, Eigen &diag)
{
   /* Note: h is diagonalized and the eigenvalue/eigenvector pairs are
   stored in diag. The number of pairs returned does NOT need to be equal
   to the dimension of the matrix h. All access to diag namely uses STL
   iterators. This opens up the possibility to perform partial
   diagonalizations, which can save a lot of time, especially in
   multi-channel cases. */

  int M = -1;

  // dr is the preferred LAPACK diagonalization routine
  dr_value dr = sP.diagroutine;

#ifdef NRG_REAL
  // Reevaluate situation: for small dim, dsyev appears to be faster.
  const unsigned int dim = m.size1();
  if (dim < sP.dsyevrlimit)
     dr = diagdsyev;

  if (dr == diagdsyev)
     M = diagonalise_dsyev(m, diag, 'V');

  if (dr == diagdsyevr)
     M = diagonalise_dsyevr(m, diag, 'V', sP.diagratio);
#endif

#ifdef NRG_COMPLEX
  const unsigned int dim = m.size1();
  if (dim < sP.zheevrlimit)
      dr = diagzheev;
   
  if (dr == diagzheev) 
     M = diagonalise_zheev(m, diag, 'V');

  if (dr == diagzheevr) 
     M = diagonalise_zheevr(m, diag, 'V', sP.diagratio);
#endif

  my_assert(M >= 0);

  // Dump first P::logenumber eigenvalues
  if (logletter('e')) {
    cout << "eig= ";
    for_each(diag.value.begin(),
             diag.value.begin() + min(int(P::logenumber), M),
             cout << _1 << ' ');
    cout << endl;
  }

  my_assert((unsigned int)M == diag.getnr());

  if (P::checkdiag) {
     const unsigned int dim = diag.getrmax();
     my_assert(diag.matrix0.size1() == M);
     my_assert(diag.matrix0.size2() == dim);
     for (unsigned int r = 0; r < (unsigned int)M; r++) {
        assert_isfinite(diag.value(r));
	double sumabs = 0.0;
        for (unsigned int j = 0; j < dim; j++) {
          assert_isfinite(diag.vektor(r, j));
          sumabs += sqr(abs(diag.vektor(r, j)));
        }
	my_assert(num_equal(sumabs, 1.0, NORMALIZATION_EPSILON));
     }
     // Check orthogonality
     for (unsigned int r1 = 0; r1 < (unsigned int)M; r1++) {
	for (unsigned int r2 = r1+1; r2 < (unsigned int)M; r2++) {
	   t_matel skpdt = 0.0;
	   for (unsigned int j = 0; j < dim; j++) {
	      skpdt += CONJ_ME(diag.vektor(r1, j)) * diag.vektor(r2, j);
	   }
	   my_assert(num_equal(abs(skpdt), 0.0, ORTHOGONALITY_EPSILON));
	}
     }
  }
   
  if (logletter('%')) {
     const t_eigen Emin = diag.value(0);
     const t_eigen Emax = diag.value(M-1);
     cout << "max=" << Emax << " min=" << Emin << endl;
  }   

  ms.check("diag_h");
   
  return M;
}

#endif // _diag_h_
