/* 
 "NRG Ljubljana" - Numerical renormalization group for multiple
 impurities and an arbitrary number of channels

 Copyright (C) 2005-2017 Rok Zitko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

   Contact information:
   Rok Zitko
   F1 - Theoretical physics
   "Jozef Stefan" Institute
   Jamova 39
   SI-1000 Ljubljana
   Slovenia

   rok.zitko@ijs.si
*/

#include "portabil.h"
#include "nrg.h"
#include "debug.h"
#include "misc.h"

#include "param.cc"
#include "outfield.cc"

#ifdef _OPENMP
#include <omp.h>
#endif

// Shared parameters for MPI parallelization.
class sharedparam
{
 public:
   // Parameters which have to be known to the slave processes.
   dr_value diagroutine;
   double diagratio;
   unsigned int dsyevrlimit;
   unsigned int zheevrlimit;
   bool logall;
   string log;

   void init();
     
 private:
   friend class boost::serialization::access;
   
   template<class Archive>
     void  serialize(Archive & ar, const unsigned int version)
       {
	  ar & diagroutine;
	  ar & diagratio;
	  ar & dsyevrlimit;
	  ar & zheevrlimit;
	  ar & logall;
	  ar & log;
       }
};

sharedparam sP;

const int COUT_PRECISION = 16;  // default precision for cout stream

// Timing of various parts of the executing code and memory
// statistics
Timing t;
MemoryStats ms;

// Symmetry type specification
string sym_string = "";

// Quantum number types defined to enforce type checking
typedef int Number;
typedef int Ispin;
typedef int Sspin;
typedef int Tangmom;
typedef int SZspin;

// Invariant subspace abstraction (i.e. the container which holds
// quantum numbers)
#include "nrg-invar-ANY.cc"

// *** Commonly used types ***

#ifdef NRG_REAL
typedef double t_matel; // type for the matrix elements
typedef double t_eigen; // type for the eigenvalues
typedef double t_coef; // type for the Wilson chain coefficients
typedef double t_factor; // type for various prefactors in recalculations
typedef double t_expv; // type for expectation values of operators
// Conjugation of matrix elements: no op
# define CONJ_ME(x) (x)
#endif

#ifdef NRG_COMPLEX
typedef cmpl t_matel;
typedef double t_eigen;
typedef cmpl t_coef;
typedef cmpl t_factor;
typedef cmpl t_expv; // we allow the calculation of expectation values of
                     // non-Hermitian operators!
#define CONJ_ME(x) (conj(x))
#endif

typedef t_matel t_weight; // type for spectral weight accumulators

// Type for arrays of eigenvalues
typedef ublas::vector<t_eigen> EVEC;
typedef std::vector<t_eigen> STDEVEC;

// Type for arrays of coefficients for Wilson chains
typedef ublas::vector<t_coef> CVEC;

typedef ublas::vector<double> DVEC;
typedef ublas::vector<unsigned int> IVEC;

/* NOTE: Row major is the C array format: A[0][0], A[0][1], A[0][2],
A[1][0], A[1][1], etc. The default in UBLAS is row major, while LAPACK
routines expect column major matrices. Of course, this is of no concern for
symmetric matrices. Default storage type is unbounded_array<T>. 
 
Thus, as always:
  first index - row
  second index - column
 
 when accessing columns - stride=m.size2()
 when accessing rows - stide=1

 Optimization rule: use stride 1 sequential access where possible.
 ublas default matrix storage is row major (i.e. C-like). The rule is
 "right index the same as inner loop variable".
*/

typedef matrix<t_matel, ublas::row_major> Matrix;

#include "numerics.h"

typedef std::vector<Invar> InvarVec; // vector of Invars

typedef pair<Invar, Invar> Twoinvar;
typedef map<Twoinvar, Matrix> MatrixElements;
typedef map<Invar, Matrix> DensMatElements;

// Dump irreducible matrix elements between two subspaces
void dump_matrix_elements(const MatrixElements &m, ostream &fout = cout)
{
  const double CHOPSMALL = 1e-8;
  const unsigned int MAXDUMP = 10;

  FOREACH(const MatrixElements::value_type &i, m) {
    fout << "----" << i.first << "----" << endl;
    for (unsigned int r1 = 0; r1 < MIN(i.second.size1(), MAXDUMP); r1++) {
      for (unsigned int r2 = 0; r2 < MIN(i.second.size2(), MAXDUMP); r2++)  
        fout << chop( (i.second)(r1,r2), CHOPSMALL ) << " ";
      fout << endl;
    }
  }
}

// (Reduced) density matrix
DensMatElements rho;
DensMatElements rhoFDM;

template <typename T>
inline pair<T, T> reverse_pair(const pair<T, T> &i)
{
   return make_pair(i.second, i.first);
}

// Map of operators matrices
typedef map<string, MatrixElements> CustomOp;

// Name of the operator corresponding to a CustomOp map element
string NAME(const CustomOp::value_type &i) { 
   return i.first; 
}

// Vector containing irreducible matrix elements of f operators.
typedef std::vector<MatrixElements> OpchChannel;
// Each channel contains P::perchannel OpchChannel matrices.
typedef std::vector<OpchChannel> Opch;

// Object of class IterInfo cotains full information when entering
// stage N of the NRG iteration.

class IterInfo {
public:
  Opch opch; // f operators (channels)

  CustomOp ops; // singlet operators (even parity)
  CustomOp opsp; // singlet operators (odd parity)
  CustomOp opsg; // singlet operators [global op]
  CustomOp opd; // doublet operators (spectral functions)
  CustomOp opt; // triplet operators (dynamical spin susceptibility)
  CustomOp opq; // quadruplet operators (spectral functions for J=3/2)
  CustomOp opot; // orbital triplet operators

  // For debugging purposes we also store the information about the
  // ancestors of each invariant subspace. Ancestors are those invariant
  // subspaces which are combined with the states on the newly added
  // lattice site(s) to give each new invariant subspace.
  
  map<Invar, InvarVec> ancestors;
   
  // Erases all information in the structures of IterInfo object. This is
  // called at the start-up and before the DM-NRG run, so that the data
  // structures are properly reset. opch is erased in read_ireducf().
  void cleanup() {
     ops.clear();
     opsp.clear();
     opsg.clear();
     opd.clear();
     opt.clear();
     opq.clear();
     opot.clear();
   
     ancestors.clear();
  }   
};

IterInfo a; // NOTE: global object!

// We need to store the dimensions of the invariant subspaces |r,1>,
// |r,2>, |r,3>, etc. The name "rmax" comes from the maximal value of
// the index "r" which ranges from 1 through rmax.

class Rmaxvals {
 private:
   IVEC values;
   
   friend ostream& operator<<(ostream &os, const Rmaxvals &rmax);
   void store(IVEC &rmx);

public:
   Rmaxvals () {};
   Rmaxvals(const Rmaxvals &v);
   unsigned int rmax(unsigned int i) const;
   unsigned int offset(unsigned int i) const;
   unsigned int operator[](int i) const { return rmax(i); }
   
   // The only way to set up the values in Rmaxvals is by calling
   // determine_ranges(), or by using the copy constructor.
   void determine_ranges(const Invar &I, const InvarVec &In);

   // total() returns the total number of states. There is therefore
   // no need to store this value separately.
   unsigned int total() {
      return accumulate(values.begin(), values.end(), 0);
   }

 private:
   friend class boost::serialization::access;
   
   template<class Archive>
     void serialize(Archive & ar, const unsigned int version) {
	ar & values;
     }
};

// Information about the number of states, kept and discarded, rmax,
// and eigenenergies.
class DimSub {
  public:
    unsigned int kept;
    unsigned int discarded;
    unsigned int total;
    Rmaxvals rmax; // substructure of vectors omega
   
    EVEC eigenvalue; // all eigenvalues
    EVEC absenergy; // absolute energies (for FDM)
   
    DimSub () {
      kept = discarded = total = 0;
    };
    DimSub(unsigned int _kept, unsigned int _total) {
      kept = _kept;
      total = _total;
      my_assert(kept <= total);
      discarded = total-kept;
    }
    DimSub(const DimSub &ds) {
      kept = ds.kept;
      discarded = ds.discarded;
      total = ds.total;
    }
};

// Full information about the number of states and matrix dimensions
// Example: dm[N].rmax[I] etc.
typedef map<Invar, DimSub> Subs;
typedef std::vector<Subs> AllSteps;
AllSteps dm; // was: Dims dims

// Result of a diagonalisation: eigenvalues and eigenvectors
class Eigen {
 public:
   unsigned int nr; // number of eigenpairs (currently stored)
   unsigned int rmax; // dimensionality of the matrix space
   unsigned int nrpost; // number of eigenpairs after truncation

   EVEC value; // eigenvalues (shifted by Egs)
   EVEC value_raw; // eigenvalues (as obtained in diagonalization, never shifted)
   EVEC absenergy; // absolute energies (0 is the absolute ground state of the system)
   Matrix matrix0; // eigenvectors in matrix form

   // 'blocks' contains eigenvectors separated according to the invariant
   // subspace from which they originate. This separation is required for
   // using the efficient BLAS routines when performing recalculations of
   // the matrix elements.
   std::vector<Matrix> blocks;

   // Various assertion checks; to be called after the
   // diagonalisation routine or reading Eigen objects through MPI or
   // from disk.
   void perform_checks() const;

   // Copy constructor
   Eigen(const Eigen & t) 
      : nr(t.nr), rmax(t.rmax), nrpost(t.nrpost),
        value(t.value), value_raw(t.value_raw), absenergy(t.absenergy),
        matrix0(t.matrix0) { 
	   perform_checks();
	}
   
   // nr - number of eigenpairs, rmax - dimensionality of the matrix space
   Eigen(int nr1, int rmax1) 
     : nr(nr1), rmax(rmax1) {
      nrpost = 0;
      my_assert(nr > 0 && rmax > 0 && rmax >= nr);
      value.resize(nr);
      value_raw.resize(nr);
      absenergy.resize(nr);
      matrix0.resize(nr, rmax);
      perform_checks();
   }
   
   Eigen() { 
      nr = 0;
      rmax = 0;
      nrpost = 0;
      value.resize(nr);
      value_raw.resize(nr);
      matrix0.resize(nr, nr);
   };
   
   // Accessor routine for j-th element of i-th eigenvector.
   inline t_matel & vektor(int i, int j) {
      return matrix0(i, j); 
   }
   
   // Returns the number of eigenpairs CURRENTLY STORED.
   unsigned int getnr() const {
      return nr; 
   }
   
   // Returns the dimensionality of a subspace, i.e. the number of
   // components of each eigenvector.
   unsigned int getrmax() const { 
      return rmax; 
   }
   unsigned int getdim() const {
      return rmax;
   }

   // Returns the number of eigenpairs after truncation.
   unsigned int getnrpost() const { 
      return nrpost; 
   }

   // Truncate to nrpost states.
   void truncate_prepare(unsigned int nrpost1) {
      nrpost = nrpost1;
      my_assert(nrpost <= nr);
   }

   void truncate_perform();

   // Initialize the data structures with eigenvalues 'v'. The eigenvectors
   // form an identity matrix. This is used to represent the spectral
   // decomposition in the eigenbasis itself.
   void diagonal(EVEC &v) {
      nr = rmax = v.size();
      value = v;
      value_raw = v;
      matrix0 = identity_matrix<t_eigen>(nr);
   }
   
   // Pad eigenvectors with zeros to size new_size.
   void pad_eigenvectors(unsigned int new_size) {
      my_assert(new_size >= rmax);
      if (new_size == rmax)
	return;
      my_assert(nr > 0);
      Matrix new_matrix0(nr, new_size);
      new_matrix0.clear();
      matrix_range<Matrix> view(new_matrix0, 
				range(0, nr),
				range(0, rmax));
      view = matrix0;
      matrix0.swap(new_matrix0);
      rmax = new_size;
      my_assert(matrix0.size1() == nr && matrix0.size2() == rmax);
   }
   
 private:
   friend class boost::serialization::access;
   
   template<class Archive>
     void serialize(Archive & ar, const unsigned int version) {
	ar & nr;
	ar & rmax;
	ar & value;
	ar & value_raw;
	ar & absenergy;
	ar & matrix0;
     }
};

void Eigen::truncate_perform()
{
   // Note that the vector 'value' is not truncated: it contains **all**
   // the calculated eigenvalues.
   const unsigned int nrblocks = blocks.size();
   my_assert(nrblocks > 0);
   for (unsigned int bl = 0; bl < nrblocks; bl++) {
      my_assert(nrpost <= blocks[bl].size1());
      blocks[bl].resize(nrpost, blocks[bl].size2());
   }
   nr = nrpost;
}

void Eigen::perform_checks() const
{
   my_assert(value.size() == matrix0.size1());
   my_assert(matrix0.size1() <= matrix0.size2());
   my_assert(nr == matrix0.size1());
   my_assert(rmax == matrix0.size2());
}

// Full information after diagonalizations.
typedef map<Invar, Eigen> DiagInfo;

#define LOOP(diag, var) FOREACH(DiagInfo::value_type &var, diag)
#define LOOP_const(diag, var) FOREACH(const DiagInfo::value_type &var, diag)

Invar INVAR(const DiagInfo::value_type &i) { return i.first; }
Eigen & EIGEN(DiagInfo::value_type &i) { return i.second; }
Eigen const & EIGEN(const DiagInfo::value_type &i) { return i.second; }

// Number of calculated states
unsigned int NRSTATES(const DiagInfo::value_type &i) {
   return i.second.getnr(); 
}

// Dimensionality of the subspace (dim)
unsigned int RMAX(const DiagInfo::value_type &i) { 
   return i.second.getrmax(); 
}

ostream & operator<<(ostream &os, const Twoinvar &p)
{
  return os << "(" << p.first << ") (" << p.second << ")";
}

template <typename T>
ostream & operator<<(ostream &os, const ublas::vector<T> &v)
{
   for_each(v.begin(), v.end(), os << _1 << " ");
   return os;
}

typedef map<Invar, Rmaxvals> QSrmax;

/*
  The following computational approaches are implemented:

 - SPEC_FT = Zlatic, Costi, Hewson: the traditional approach

 - SPEC_DMNRG = Hoffstetter's DM-NRG approach

 - SPEC_CFS = Complete Fock Space approach (Peters, Pruschke, Anders)
 - SPEC_CFSgt = CFS, greater spectral function, FT[ G^>_{AB}(t) = -i theta(t) <A(t)B> ]
                (with proper theta function, so this is a retarded GF)
 - SPEC_CFSls = CFS, lesser spectral function, FT[ G^<_{AB}(t) = -i theta(t) epsilon <BA(t)> ]
                Convention: CFSgt+CFSls=CFS

 - SPEC_FDM = full-density-matrix (Weichselbaum, von Delft)
 - SPEC_FDMgt = full-density-matrix, greater spectral function, FT[ G^>_{AB}(t) = -i theta(t) <A(t)B> ]
                (with proper theta function, so this is a retarded GF)
 - SPEC_FDMls = full-density-matrix, lesser spectral function, FT[ G^<_{AB}(t) = -i theta(t) epsilon <BA(t)> ]
                Convention: FDMgt+FDMls=FDM

 - SPEC_GT = conductance G(T), masquerading as a spectral function
 - SPEC_I1T = first moment of A(w)(-df/dw)
 - SPEC_I2T = second moment of A(w)(-df/dw)

 - SPEC_CHIT = susceptibility chi(T), masquerading as a spectral function
*/

enum SPECTYPE { SPEC_FT,
                SPEC_DMNRG, 
                SPEC_CFSold, SPEC_CFS, SPEC_CFSgt, SPEC_CFSls, 
                SPEC_FDMold, SPEC_FDM, SPEC_FDMgt, SPEC_FDMls,
                SPEC_GT, SPEC_I1T, SPEC_I2T,
                SPEC_CHIT };

string spectypestring(SPECTYPE spectype)
{
   switch (spectype) {
    case SPEC_FT: return "FT";

    case SPEC_DMNRG: return "DMNRG";

    case SPEC_CFSold: return "CFSold";
    case SPEC_CFS: return "CFS";
    case SPEC_CFSgt: return "CFSgt";
    case SPEC_CFSls: return "CFSls";

    case SPEC_FDMold: return "FDMold";
    case SPEC_FDM: return "FDM";
    case SPEC_FDMgt: return "FDMgt";
    case SPEC_FDMls: return "FDMls";

    case SPEC_GT: return "GT";
    case SPEC_I1T: return "I1T";
    case SPEC_I2T: return "I2T";

    case SPEC_CHIT: return "CHIT";

    default: return "undefined";
   }   
}

// spectype represents a temperature-dependent quantity such as G(T)
bool temp_spectrum(SPECTYPE spectype)
{
  switch (spectype) {
  case SPEC_GT: // Conductance G(T)
  case SPEC_I1T: // First-moment transport integral
  case SPEC_I2T: // Second-moment transport integral
  case SPEC_CHIT: // Susceptibility (calculated via spectral representation)
    return true;
  default:
    return false;
  }
}  

// Returns 'true' if spectype corresponds to a true spectral function
// represented as a collection of delta functions.
// Returns 'false' if spectype is a temperature-dependent quantity,
// masquerading as a spectral function [this affects patching,
// trimming, broadening, and output].
bool real_spectrum(SPECTYPE spectype)
{
  if (temp_spectrum(spectype))
    return false;
  
  return true;
}

ostream & operator<<(ostream &os, const SPECTYPE spectype)
{
   return os << spectypestring(spectype);
}

// In namespace NRG we store run-time information about the calculation.
namespace NRG 
{
   // Flag to signal an insufficient number of states computed.
   bool notenough;

   // Diagratio for the following diagonalisation call
   double diagratio;
   
   // Invariant subspaces to be computed.
   std::vector<Invar> tasks;
};

// Includes which require P:: parameters.
#include "spectral.h"

// Setup output fields that will appear in the file "td".
// Additional elements are defined in symmetry.cc.
namespace TD {
  outfield T("T");
  // insert others here (starting with pos=1, i.e. after the 1st element)
  outfield E("<E>");
  outfield E2("<E^2>");
  outfield C("C");
  outfield F("F");
  outfield S("S");

   void save_TD_quantities(ostream &F) {
      F << " ";
      FOREACH(outfield *i, allfields) 
	i->put(F);
      F << endl;
   }

   void save_header(ostream &F) {
      F << "#";
      FOREACH(outfield *i, allfields) 
	i -> putheader(F);
      F << endl;
   }   
};

// The factor that multiplies the eigenvalues of the length-N Wilson
// chain Hamiltonian in order to obtain the energies on the original
// scale. Also named the "reduced bandwidth". Note that STAT::scale =
// SCALE(STAT::N+1) [see function set_N], thus for Ninit=0 and calc0,
// setting N=-1 results in a call SCALE(0). In the actual NRG run,
// the scale is at least SCALE(1). This is important for correct
// handling of rescaling for substeps==true.
double SCALE(int N) 
{
   double scale = 0.0;

   if (string(P::discretization) == "Y")
     // Yoshida,Whitaker,Oliveira PRB 41 9403 Eq. (39)
     scale = 0.5 * (1. + 1./P::Lambda);

   if (string(P::discretization) == "C" || string(P::discretization) == "Z")
     // Campo, Oliveira PRB 72 104432, Eq. (46) [+ Lanczos]
     scale = (1.0 - 1./P::Lambda)/log(P::Lambda);

   if (!P::substeps) {
      scale *= pow(P::Lambda, -(N-1)/2. + 1 - P::z);
   } else {
      scale *= pow(P::Lambda, -N/(2.*P::channels) + 3/2. - P::z);
   }
   
   my_assert(scale != 0.0); // yes, != is intentional here.
      
   scale = scale * P::bandrescale; // RESCALE
   
   return scale;
}

typedef pair<int, int> pii;
// M ranges 0..channels-1
pii get_Ntrue_M(int N)
{ 
   int i = N/P::channels;
   return pii(i, N-i*P::channels);
}

// Compensate for different definition of SCALE in initial.m and C++
// code in case of substeps==true.
double scale_fix(int N)
{
   int Ntrue, M;
   tie(Ntrue, M) = get_Ntrue_M(N);
   my_assert(N == Ntrue*P::channels + M);
   
   int N_at_end_of_full_step = Ntrue*P::channels + P::channels-1; // M=0,...,channels-1
   double scale_now = SCALE(N+1);
   double scale_at_end_of_full_step = SCALE(N_at_end_of_full_step+1);
   return scale_now/scale_at_end_of_full_step;
}

enum RUNTYPE { RUN_NORMALNRG, RUN_DMNRG };

// Map from double to double. This is used for the "fixeps" trick as a data
// structure which holds the transformation rules from original eigenvalues
// to shifted eigenvalues.

typedef MY_UNORDERED_MAP<t_eigen, t_eigen> mapdd;

// Pair of energy and multiplicity. This is used to compute the true
// grand-canonical partition function STAT::ZZ.
typedef pair<t_eigen, int> energy_mult_type;
typedef list<energy_mult_type> excitation_list;

// Namespace for storing various statistical quantities calculated
// during iteration.
namespace STAT {
  RUNTYPE runtype; // NRG vs. DM-NRG run

  int N; // iteration step (-1 indicates the step before 1st iter, cf. calc0)
  double scale; // scale factor
  double Teff; // effective temperature
  double scT; // scT = scale/P::T, this combination appears in the
              // exponents (Boltzmann weights)

  /* After the diagonalization, the energy scale is reduced by a factor of
   \Lambda^{1/2}. True energies are now eigenvalues multiplied by this new
   scale. The effective temperature is T_(N+1)=scale/P::betabar. Note that
   E/T = E_N * scale / (scale / betabar) = betabar E_N. P::betabar
   determines the factor c in the expression T_N = c Lambda^-(N-1)/2. */

  void set_N(int newN) {
     N = newN;
     scale = SCALE(newN + 1); // Current energy scale in units of bandwidth D
     Teff = scale/P::betabar;
     scT = scale/P::T; // Boltzmann weight
  }
   
  // Ground-state subspace and energy at the current NRG iteration
  Invar ground;
  t_eigen Egs;

  // All energies, sorted in increasing order
  STDEVEC energies;

  // Mapping old_energy -> new_energy for curing small splittings due to
  // the floating point roundoff errors.
  mapdd cluster_mapping;

  double Zft; // grand-canonical partition function (at shell n)
  double Zgt; // grand-canonical partition function for computing G(T)
  double Zchit; // grand-canonical partition function for computing chi(T)

  double E; // energy times beta
  double E2; // (energy times beta) squared
  double C; // heat capcity (in units of 1/k_B)
  double F; // free energy times beta
  double S; // entropy: (beta E)-(beta F)

  // Expectation values 
  typedef map<string, t_expv> t_mapexpv;
  t_mapexpv expv; // expectation values of custom operators
  t_mapexpv fdmexpv;   // Expectation values computed using the FDM algorithm

   // Total energy of the ground state. This is the sum of all the zero state
  // energies for all the iteration steps. 
  t_eigen totalenergy;
   
  // This is the value of the variable "totalenergy" at the end of the
  // iteration. This is different from 'Egs'.
  t_eigen GSenergy;

  // Containers related to the FDM-NRG approach.
  // Consult A. Weichselbaum, J. von Delft, PRL 99, 076402 (2007).
  std::vector<double> ZnD; // Z_n^D=\sum_s^D exp(-beta E^n_s),
                           // sum over **discarded** states at shell n
                           // cf. Eq. (8) and below in WvD paper
  std::vector<double> wn; // Weights w_n. They sum to 1.
  std::vector<double> wnfactor; // wn/ZnD

  double ZZ = 0; // grand-canonical partition function (full-shell),
                 // evaluated at the temperature P::T
};

// Return true if this is the first step of the NRG iteration
bool FIRST_ITERATION(unsigned int N = STAT::N) 
{
  return N == P::Ninit;
}

// Return true if this is the last step of the NRG iteration
bool LAST_ITERATION(unsigned int N = STAT::N)
{
  return (N+1) == P::Nmax;
}

// Test for first NRG run [diagonalisations] with (nrgrun)
class runtypenrg 
{
 public:
   operator bool () { 
      return STAT::runtype == RUN_NORMALNRG;
   }
} nrgrun;

// Test for second NRG run [density matrix computations] with (dmnrgrun)
class runtypedmnrg
{
 public:
   operator bool () { 
      return STAT::runtype == RUN_DMNRG; 
   }
} dmnrgrun;

Rmaxvals::Rmaxvals(const Rmaxvals &v) { 
   values = v.values; 
}

void Rmaxvals::store(IVEC &rmx) {
   my_assert(rmx.size() == P::combs);
   values = rmx;
}

// Returns 0 if no values are stored!
unsigned int Rmaxvals::rmax(unsigned int i) const {
   allowed_block_index(i);
   return values[i-1]; // FOR COMPATIBILITY OFFSET 1!
}

unsigned int Rmaxvals::offset(unsigned int i) const {
   allowed_block_index(i);
   return accumulate(values.begin(), values.begin() + (i-1), 0);
}

ostream& operator<<(ostream &os, const Rmaxvals &rmax) 
{
  for_each(rmax.values.begin(), rmax.values.end(), os << _1 << ' ');
  return os;
}

// Data structures containing NRG eigenvalues and eigenvectors

DiagInfo diag; // holds eigenvalues and unitary matrices for all subspaces
DiagInfo diagprev; // holds eigenenegies (from previous/current iteration and initial ones)
QSrmax qsrmax; // holds the number of states for |r,1>, |r,2>,...

// Returns true if option 'c' is selected for logging
bool logletter(char c)
{
   return (sP.logall ? true : sP.log.find(c) != string::npos);
}

// Index 'n' of the last site in the existing chain, f_n (at iteration 'N').
// The site being added is f_{n+1}.
// This is the value that we use in building the matrix,
// cf. nrg-makematrix-ISO.cc
int getnn()
{
   return STAT::N;
}

// Energy scale at the last NRG iteration
double LAST_STEP_SCALE()
{
  return SCALE(P::Nmax);
}

void validateparameters();

map<string, string> parsed_params;

// Parse command line and parameter file
// Warning: there's no error checking for parameters. 
// Check the parameter dump to see if everything is OK.
void parse_cmnd_line_read_parameters(int argc, char *argv[])
{
  if (argc == 3) {
     if (strcmp(argv[1], "-w") == 0) {
	P::workdir = argv[2];
     }
  }
  cout << "workdir=" << P::workdir << endl;
   
  parser(parsed_params, "param");

  FOREACH(parambase *i, allparams) {
    const string keyword = i->getkeyword();
    if (parsed_params.count(keyword) == 1) {
      i->setvalue_str(parsed_params[keyword]);
    }
  }
};

void dump_parameters()
{
  FOREACH(parambase *i, allparams) 
     i->dump();
}

// This class holds table of generalized xi/zeta/etc. coefficients
class coef_table {
private:
  CVEC t;

public:
  // Read values from a stream f
  void read_values(ifstream &f, bool update = true) {
    unsigned int len;
    f >> len; // get length (= last index n still included)
    read_vector(f, t, len+1, update);
    if (update)
       my_assert(t.size() == len+1);
  }

  t_coef coef(unsigned int n) const {
    my_assert(n < t.size());
    return t[n];
  }
   
  // Returns the index of the last coefficient still included in the table.
  unsigned int max() const {
     my_assert(t.size() >= 1);
     return t.size()-1;
  }
   
   void setvalue(unsigned int n, t_coef val) {
     if (n+1 > t.size()) 
       t.resize(n+1);
     t[n] = val;
  }   
};

// NOTE: One table of discretization coefficients for each channel
class set_of_tables
{
 private:
   std::vector<coef_table> tabs;
   
 public:
   unsigned int nr_tabs() const { return tabs.size(); }

   void read(ifstream &fdata, bool update = true) {
      tabs.resize(P::coefchannels);
      for (unsigned int alpha = 0; alpha < P::coefchannels; alpha++)
	tabs[alpha].read_values(fdata, update);
   }

   t_coef operator()(int N, unsigned int alpha) const {
      allowed_coefchannel(alpha);
      my_assert(alpha < tabs.size());
      return tabs[alpha].coef(N);
   }
   
   unsigned int max(unsigned int alpha) const {
      allowed_coefchannel(alpha);
      my_assert(alpha < tabs.size());
      return tabs[alpha].max();
   }
   
   void setvalue(unsigned int N, unsigned int alpha, t_coef val) {
      allowed_coefchannel(alpha);
      my_assert(alpha < tabs.size() && N <= P::Nmax);
      tabs[alpha].setvalue(N, val);
   }
};

set_of_tables xi; // f^dag_N f_N+1 terms
set_of_tables zeta; // f^dag_N f_N terms

// Support for spin-polarized conduction bands. See also P::polarized.
// Hack: the total number of channels is doubled, the index runs from
// 0...2*P::channels-1. Numbers 0...P::channels-1 correspond to spin up,
// while P::channels...2*P::channels-1 correspond to spin down.
// Compare P::channels and P::coefchannels (which reflects the same
// convention in initial.m, i.e. CHANNELS vs. COEFCHANNELS).
#define xiUP(N, ch)     xi((N),   (ch))
#define xiDOWN(N, ch)   xi((N),   (ch)+P::channels)
#define zetaUP(N, ch)   zeta((N), (ch))
#define zetaDOWN(N, ch) zeta((N), (ch)+P::channels)

// Support for conduction bands with full 2x2 matrix structure, a
// generalization of P::polarized. The total number of "channels" is
// here multiplied by 4, i.e., the index runs from 0 to
// 4*P::channels-1. Numbers 2*P::channels...3*P::channels-1 correspond
// to UP/DO, 3*P::channels...4*P::channels-1 correspond to DO/UP.
#define xiUPDO(N, ch)   xi((N),   (ch)+2*P::channels)
#define xiDOUP(N, ch)   xi((N),   (ch)+3*P::channels)
#define zetaUPDO(N, ch) zeta((N), (ch)+2*P::channels)
#define zetaDOUP(N, ch) zeta((N), (ch)+3*P::channels)

// Support for channel-mixing Wilson chains
set_of_tables xiR; 
set_of_tables zetaR;

// Support for superconducting bands.
set_of_tables delta; // f^dag_up,N f^dag_down,N terms
set_of_tables kappa; // f^dag_N f^dag_down,N+1 terms

set_of_tables ep, em; // e_n coefficients
set_of_tables u0p, u0m; // u_{0,m} coefficients

#ifdef TRIDIAG
#include "tridiag.h"
#endif // TRIDIAG

#include "diag.h"

#include "symmetry.cc"
#include "matrix.cc"
#include "recalc.cc"

// Select which symmetries to compile in. For exotic types,m the
// coefficient tables are long and the compilation becomes quite
// long.
#define STANDARD_SYMS
#define MORE_SYMS
#define EXOTIC_SYMS

#ifdef STANDARD_SYMS
#include "sym-QS.cc"
#include "sym-QSZ.cc"
#endif

#ifdef MORE_SYMS
#include "sym-ISO.cc"
#include "sym-ISOSZ.cc"
#include "sym-SPSU2.cc"
#include "sym-SPU1.cc"
#endif

#ifdef EXOTIC_SYMS
#include "sym-ANYJ.cc"
#include "sym-DBLSU2.cc"
#include "sym-DBLISOSZ.cc"
#include "sym-ISOLR.cc"
#include "sym-ISOSZLR.cc"
#include "sym-NONE.cc"
#include "sym-P.cc"
#include "sym-PP.cc"
#include "sym-SL.cc"
#include "sym-SL3.cc"
#include "sym-SPSU2LR.cc"
#include "sym-SPSU2T.cc"
#include "sym-SPSU2C3.cc"
#include "sym-SPU1LR.cc"
#include "sym-SU2.cc"
#include "sym-QSLR.cc"
#include "sym-QSC3.cc"
#include "sym-QST.cc"
#include "sym-QSTZ.cc"
#include "sym-QSZTZ.cc"
#include "sym-QSZLR.cc"
#include "sym-QJ.cc"
#include "sym-U1.cc"
#endif

#include "read-input.cc"

/**** Calculation of traces ****/

// Used in calculate_TD().
CONSTFNC double calculate_Z(const DiagInfo::value_type &is, double rescale_factor)
{
     double sumZ = 0;
     unsigned int dim = EIGEN(is).value.size();
     for (unsigned int r = 0; r < dim; r++)
           sumZ += exp(-rescale_factor * EIGEN(is).value(r));
   /*
  double sumZ = accumulate(EIGEN(is).value.begin(),
			   EIGEN(is).value.end(),
			   0.0,
			   _1 + bind(dblexp, -rescale_factor * _2) );
    */
  sumZ *= mult(INVAR(is));
  return assert_isfinite(sumZ);
}

// Formated output for the expectation values. 

template <typename T> void output_val(ostream &F, T x)
{
   F << x;
}

template <typename T> void formatted_output(ostream &F, T x) {
   // Important: setw first, setprecision second
   F << setw(P::width_custom) 
     << setprecision(P::prec_custom) << x << " ";
}

// Specialization for complex values, the output format is X+IY or
// X-IY, where X and Y are real and imaginary part, respectively. The
// imaginary part is only shown where its value relative to the real
// part is sufficiently large. No space is used in the outputted
// string in order to simplify parsing.
// This behavior can be turned off using P::noimag.
// The default is now P::noimag=true.
#ifdef NRG_COMPLEX
const double OUTPUT_IMAG_EPS=1.0e-13;

void output_val(ostream &F, cmpl val)
{
   if (P::noimag || 
       abs(val.imag()) < abs(val.real()) * OUTPUT_IMAG_EPS) {
      F << val.real();
   } else {
      F << val.real();
      if (val.imag() > 0.0) {
	 F << "+I" << val.imag();
      } else {
	 F << "-I" << -val.imag();
      }
   }
}

void formatted_output(ostream &F, cmpl val) 
{
   ostringstream str;

   // This sets precision for both real and imaginary parts.
   str << setprecision(P::prec_custom);

   if (P::noimag ||
       abs(val.imag()) < abs(val.real()) * OUTPUT_IMAG_EPS) {
      str << val.real();
   } else {
      str << val.real();
      if (val.imag() > 0.0) {
	 str << "+I" << val.imag();
      } else {
	 str << "-I" << -val.imag();
      }
   }

   // The width for the whole X+IY string.
   F << setw(P::width_custom) << str.str() << " ";
}
#endif // NRG_COMPLEX

/* "Trace" of a singlet operator: actually this is the statistical average
 with respect to exp(-beta H), but without the 1/Z factor. I.e.
 \sum_n <n|exp(-beta H) O|n>, for a given singlet operator 'O'.  As a
 side effect, dump the matrix elements to stream F if so requested
 (parameter "dumpdiagonal").  */
CONSTFNC t_expv calc_trace_singlet(const DiagInfo &diag,
				   const MatrixElements &n,
				   ostream &F = cout)
{
  weight_bucket tr;

  LOOP_const(diag, is) {
    const Twoinvar I = make_pair(INVAR(is), INVAR(is));

    if (n.count(I) != 1) {
       cout << "ERROR: calc_trace_singlet() I=(" << I << ") cnt=" << n.count(I) << endl;
       exit(1);
    }
     
    const Matrix & nI = n.find(I)->second;

    const unsigned int dim = NRSTATES(is);
    my_assert( dim == nI.size2() );

    weight_bucket sum;
    for (unsigned int r = 0; r < dim; r++) {
       sum += exp(-P::betabar * EIGEN(is).value(r) ) * nI(r, r);
    }
     
    if (P::dumpdiagonal != 0) {
       F << INVAR(is) << ": ";
       for (unsigned int r = 0; r < dim; r++) {
	  if (r < P::dumpdiagonal) {
	     F << nI(r, r) << " ";
	  }
       }
       F << endl;
    }
     
    tr += t_weight(mult(INVAR(is))) * t_weight(sum);
  }

  return tr;
}

CONSTFNC t_expv calc_trace_fdm_discarded(const DiagInfo &diag,
					 const MatrixElements &n)
{
   const double T = P::T;
   weight_bucket tr;
   
   LOOP_const(diag, is) {
      const Invar I = INVAR(is);
      const Twoinvar II = make_pair(INVAR(is), INVAR(is));
      
      my_assert(n.count(II) == 1);
      const Matrix & nI = n.find(II)->second;
      
      const unsigned int dim = dm[STAT::N][I].total;
      unsigned int ret = dm[STAT::N][I].kept;
      
      if (LAST_ITERATION()) {
	 ret = 0; // All states are considered discarded in the last iteration
      }
      
      const Eigen & diagI = EIGEN(is);
      
      my_assert( dim == nI.size2() );
      
      weight_bucket sum;
      for (unsigned int i = ret; i < dim; i++) { // over discarded states ONLY
	 sum += exp(-diagI.absenergy(i)/T) * nI(i, i);
      }
   
      tr += t_weight(mult(INVAR(is))) * t_weight(sum);
   }

   return tr;
}

CONSTFNC t_expv calc_trace_fdm_kept(const DiagInfo &diag,
				    const MatrixElements &n)
{
   weight_bucket tr;
   
   LOOP_const(diag, is) {
      const Invar I = INVAR(is);
      const Twoinvar II = make_pair(INVAR(is), INVAR(is));
      
      my_assert(n.count(II) == 1);
      const Matrix & nI = n.find(II)->second;
      
      const unsigned int dim = dm[STAT::N][I].total;
      const unsigned int ret = dm[STAT::N][I].kept;
      
      const Matrix & rho = rhoFDM[I];
      
      weight_bucket sum;
      for (unsigned int i = 0; i < ret; i++) // over kept states ONLY
	 for (unsigned int j = 0; j < ret; j++) 
	    sum += rho(i, j) * nI(j, i);
   
      tr += t_weight(mult(INVAR(is))) * t_weight(sum);
   }

   return tr;
}

#include "bins.h"

// Return true if the spectrum of the given type is computed using the
// density matrix.
bool dm_spec(SPECTYPE spectype)
{
  switch (spectype) {
    case SPEC_DMNRG:

    case SPEC_CFSold:
    case SPEC_CFS:
    case SPEC_CFSgt:
    case SPEC_CFSls:

    case SPEC_FDMold:
    case SPEC_FDM:
    case SPEC_FDMgt:
    case SPEC_FDMls: 
      return true;
      
    default:
      return false;
  }
}

/* Object of class 'ChainSpectrum' will contain information about the
 spectral density calculated at a given stage of the NRG run, i.e. for a
 finite Wilson chain.  We then merge it in an object of class 'Spectrum'
 which holds the spectral information for the entire run (i.e. the physical
 spectral density). */

class ChainSpectrum {
public:
  // Spectral information
  Bins spos, sneg;

  ChainSpectrum() {};

  virtual void add(double energy, t_weight weight) {
    my_error("bug: add() in base class should not be called.");
  };
   
  virtual ~ChainSpectrum() {};
   
  // Total weight of the spectrum
  virtual t_weight total_weight() { 
    my_error("should not be reached"); 
  };
};

class ChainSpectrumBinning : public ChainSpectrum {
public:
  ChainSpectrumBinning() {};

  virtual ~ChainSpectrumBinning() {
     assert_isfinite(sum_weights(spos)); // Bug trap
     assert_isfinite(sum_weights(sneg));
  }
   
  t_weight total_weight() {
     return sum_weights(spos) + sum_weights(sneg);
  }
   
  void add(double energy, t_weight weight);
};

class ChainSpectrumTemp: public ChainSpectrum {
public:
  ChainSpectrumTemp() {
    spos.bins.clear();
  };

  virtual ~ChainSpectrumTemp() {
     assert_isfinite(sum_weights(spos)); // Bug trap
  }
  
  void add(double T, t_weight value);
   
  t_weight total_weight() {
     return sum_weights(spos);
  }
};   

// energy can be of either sign!
void ChainSpectrumBinning::add(double energy, t_weight weight)
{
   assert_isfinite(energy);
   assert_isfinite(weight);
   if (energy >= 0.0)
     spos.add(energy, weight);
   else 
     sneg.add(-energy, weight);
}

void ChainSpectrumTemp::add(double T, t_weight value)
{
   my_assert(T >= 0.0);
   spos.add_value(T, value); // NOTE: add_value(), not add()
}

// Object of class spectrum will contain everything that we know about a
// spectral density.
class Spectrum {
 public:
  string opname, filename;
  SPECTYPE spectype;

  Spectrum(string _opname, string _filename, SPECTYPE _spectype) 
     : opname(_opname), filename(_filename), spectype(_spectype) 
       {};
  
  virtual void merge(ChainSpectrum *cs) {
    my_error("bug: Specutrm::merge() should not never be called");
  }
  
  virtual void continuous() {
    my_error("should not be reached");
  };
  virtual void savebins() { // save raw bins
    my_error("should not be reached");
  };
  virtual void postprocess() {
    my_error("should not be reached");
  };
  string name() { return opname; }
};

// Real-frequency spectral function
class SpectrumRealFreq : public Spectrum
{
private:
  // Full spectral information 
  Bins fspos;
  Bins fsneg;

  void continuous_old(DENSFUNC &, DENSFUNC &);
  void continuous_old1(DENSFUNC &, DENSFUNC &, 
		       const XYFUNC &, const Spikes &,
		       unsigned int, unsigned int); // helper func
  void continuous_wvd(DENSFUNC &, DENSFUNC &);
  void continuous_new(DENSFUNC &, DENSFUNC &);
  void continuous_newsc(DENSFUNC &, DENSFUNC &);
  void continuous_lorentz(DENSFUNC &densitypos,
			  DENSFUNC &densityneg);
  t_weight continuous_lorentz1(double E);
  void mergeNN2(ChainSpectrum *cs);
  void mergeCFS(ChainSpectrum *cs);

  void weight_report(void);
  void trim();
  void savebins();

public:
  SpectrumRealFreq(string _opname, string _filename, SPECTYPE _spectype) 
     : Spectrum(_opname, _filename, _spectype) {};

  void merge(ChainSpectrum *cs);
  void continuous();
  
  void postprocess();
};

// G(T) type of results, i.e. not a real spectrum
class SpectrumTemp : public Spectrum
{
 private:
  // Temperature dependence (maskerading as a spectrum)
  Bins fspos;

 public:
  SpectrumTemp(string _opname, string _filename, SPECTYPE _spectype) 
     : Spectrum(_opname, _filename, _spectype) {
	     fspos.bins.clear();
       };

  void merge(ChainSpectrum *cs);
  void postprocess();
};

void SpectrumRealFreq::postprocess()
{
  weight_report();
  trim();
  savebins();
  continuous();
}

/* Merge the spectrum for a finite Wilson chain into the "true" NRG
 spectrum. For complete Fock space NRG calculation, the merging tricks are
 not necessary: we just collect all the delta peaks from all iterations. */
void SpectrumRealFreq::merge(ChainSpectrum *cs) 
{
   if (logletter('*')) {
      cout << "weight=" << cs->total_weight() << endl;
   }   
   
   switch (spectype) {
    case SPEC_CFSold:
    case SPEC_CFS:
    case SPEC_CFSgt:
    case SPEC_CFSls:
    case SPEC_FDMold:
    case SPEC_FDM:
    case SPEC_FDMgt:
    case SPEC_FDMls:
     mergeCFS(cs);
     break;
     
    case SPEC_FT:
    case SPEC_DMNRG:
     mergeNN2(cs);
     break;
     
    default:
     my_error("should not be reached");
   }
   
  if (logletter('*')) {
     t_weight sumpos = fspos.total_weight();
     t_weight sumneg = fsneg.total_weight();
     t_weight total = sumpos + sumneg;
     cout << "Cummulative weight(pos)=" << sumpos
          << " weight(neg)=" << sumneg
          << " total=" << total << endl;
  }   
}

void SpectrumTemp::merge(ChainSpectrum *cs)
{
   copy(cs->spos.bins.begin(), cs->spos.bins.end(), back_inserter(fspos.bins));
}

// Spectrum merging for complete Fock space calculation.
void SpectrumRealFreq::mergeCFS(ChainSpectrum *cs)
{
  fspos.merge(cs->spos);
  fsneg.merge(cs->sneg); 
}

// energy scale factor that is exponentiated
// N/N+1 patching is recommended for larger values of Lambda,
// so as to reduce the upper limit of the patching window to
// omega_N*goodE*Lambda, i.e., by a factor of Lambda
// compared to the N/N+2 patching where the upper limit is
// omega_N*goodE*Lambda^2.
double getfactor()
{
   double factor;
   if (P::channels == 1 && !P::NN1) {
      factor = P::Lambda; // N/N+2 patching
   } else {
      factor = sqrt(P::Lambda); // N/N+1 patching
   }
   return factor;
}

// sets the scale - in relative units of STAT::scale !!
double getE0()
{
   return P::goodE; // (patching parameter p)
}

// Note: in the current iteration, spectral peaks in the [0:Emin] are
// discarded.
double getEmin()
{
   return getE0();
}

// The "peak" energy of the "window function" in the patching procedure.
double getEx()
{
   return getE0() * getfactor();
}

double getEmax()
{
   return getE0() * sqr( getfactor() );
}

// Window function: f(0)=0, f(1)=1.
inline double fnc_linear(double x)
{
   return x;
}

inline double fnc_tanh_0(double x)
{
   return tanh(P::NNtanh * (x-0.5));
}

inline double fnc_tanh(double x)
{
   const double f0 = fnc_tanh_0(0.0);
   const double fx = fnc_tanh_0(x);
   const double f1 = fnc_tanh_0(1.0);
   
   return (fx-f0)/(f1-f0);
}

inline double fnc(double x)
{
   if (P::NNtanh > 0.0) {
      return fnc_tanh(x);
   } else {
      return fnc_linear(x);
   }
}   

inline double windowfunction(double E, double Emin, double Ex, double Emax)
{
   // Exception 1
   if (E <= Ex && LAST_ITERATION())
     return 1.0;
   
   // Exception 2
   if (E >= Ex && FIRST_ITERATION())
     return 1.0;

   if (E <= Emin || E >= Emax)
     return 0.0;
   
   if (Emin < E && E <= Ex) 
     return fnc( (E-Emin)/(Ex-Emin) );
   
   if (Ex < E && E < Emax)
     return 1.0-fnc( (E-Ex)/(Emax-Ex) );
   
   my_assert_not_reached();
}

// Here we perform the actual merging of data using the N/N+2 scheme.
// Note that we use a windowfunction (see above) to accomplish the
// smooth combining of data.
void mergeNN2half(Bins &fullspec, const Bins &cs)
{  
   double Emin = STAT::scale * getEmin(); // p
   double Ex =   STAT::scale * getEx(); // p Lambda
   double Emax = STAT::scale * getEmax(); // p Lambda^2
   
   const unsigned int len = fullspec.bins.size();
   my_assert(len == cs.bins.size()); // We require equivalent bin sets!!

   for (unsigned int i = 0; i < len; i++) {
      const double energy = cs.bins[i].first;
      const t_weight weight = cs.bins[i].second;
      if (Emin < energy && energy < Emax && weight != 0.0) {
	 const double factor = (P::NN2avg ? 0.5 : 1.0);
	 fullspec.bins[i].second += 
	   factor * weight * windowfunction(energy, Emin, Ex, Emax);
      }
   }
}

// Return true if the spectral-function merging is to be performed at the
// current NRG iteration.
bool N_for_merging(int N)
{
  if (P::NN1) 
     return true;

  if (P::NN2avg)
     return true;

  const bool even = IS_EVEN(N);
  return (P::NN2even ? even : !even);
}

// Spectrum merging using the N/N+n patching. One has to be careful about
// the specifics, in particular the values of Emin, Ex and Emax. The
// current choice seems to be working quite all right.
// See R. Bulla, T. A. Costi, D. Vollhardt, Phys. Rev. B 64, 045103 (2001).

void SpectrumRealFreq::mergeNN2(ChainSpectrum *cs)
{
  if (!N_for_merging(STAT::N))
    return;
     
  mergeNN2half(fspos, cs->spos);
  mergeNN2half(fsneg, cs->sneg);
}

void SpectrumRealFreq::weight_report(void)
{
  const t_weight twneg = sum_weights(fsneg);
  const t_weight twpos = sum_weights(fspos);

  cout << "[" << opname << "]" << " pos=" << twpos << " neg=" << twneg 
     << " sum= " << twpos+twneg << endl;
   
  // Spectral moments from delta-peaks
  const t_weight mom1 = moment(fsneg, fspos, 1);
  const t_weight mom2 = moment(fsneg, fspos, 2);
  const t_weight mom3 = moment(fsneg, fspos, 3);
  const t_weight mom4 = moment(fsneg, fspos, 4);
  const t_weight f = fd_fermi(fsneg, fspos, P::T);
  const t_weight b = fd_bose(fsneg, fspos, P::T);
   
  cout <<  "mu1=" << mom1 << " mu2=" << mom2 
       << " mu3=" << mom3 << " mu4=" << mom4 << endl;
  cout << "f=" << f << " b=" << b << endl;
}

/* Here we set the lowest frequency at which we will evaluate the spectral
 density. If the value is not predefined in the parameters file, use the
 smallest scale from the calculation multiplied by P::broaden_min_ratio. */

double get_broaden_min()
{
   return (P::broaden_min <= 0.0 
	   ? P::broaden_min_ratio * LAST_STEP_SCALE() 
	   : P::broaden_min);
}

// Called by continuous_old().
// OPTIMIZATIONS: ElogE is an array holding (positive) energies as X,
// and logarithms of energies divided by loggauss_b as Y column.
void SpectrumRealFreq::continuous_old1(DENSFUNC &density,
			       DENSFUNC &density_2,
			       const XYFUNC &ElogE,
			       const Spikes &fs,
			       unsigned int nr,
			       unsigned int nrloggauss)
{
   const double gauss_b = P::gauss_b;
   const double loggauss_b = P::loggauss_b;
 
   // We loop over raw spectral peaks [actually over bins]
   FOREACH(const Spikes::value_type &i, fs) {
      const t_weight w = assert_isfinite(WEIGHT(i));
      const double e = assert_isfinite(ENERGY(i));

      if (e != 0.0) {
	 const t_weight factor = w * BR_LOG_GAUSS_factor(e, loggauss_b);
	 const double loge = log(e)/loggauss_b; // !!
	 
	 /* In the part of the broadened spectrum with energies above the
	  limit set by limitLL, we add a contribution obtained by
	  log-gaussian broadening. A peak with an energy of a given sign
	  only contributes to the broadened spectrum with energy of the
	  *same* sign. */
	 for (unsigned int cnt = 0; cnt < nrloggauss; cnt++) {
	    const double logE = ElogE[cnt].second;
	    density[cnt].second += 
	      factor * BR_LOG_GAUSS_exp(logE, loge); // optimized!
	 }
      }

      // The width of the Gaussian is proportional to the (absolute value
      // of) the peak energy.
      const double gauss_b_sc = gauss_b * abs(e);
      const double invsigma = 1.0/gauss_b_sc;
      const t_weight wfactor = w * BR_GAUSSIAN_factor(invsigma);
      
      /* In the part of the broadened spectrum with energies below the
       limit set by limitLL, we add contributions obtained by Gaussian
       broadening. Note that spectral peaks of all energies contribute! */
      for (unsigned int cnt = nrloggauss; cnt < nr; cnt++) {
	 const double E = ElogE[cnt].first;
	 density[cnt].second   += wfactor * BR_GAUSSIAN_exp(E, e, invsigma);
	 density_2[cnt].second += wfactor * BR_GAUSSIAN_exp(-E, e, invsigma);
      }
   }
}

/* For finite temperature calculations, Gaussian broadening will be used
 near omega=0 and log gaussing otherwise. For zero temperature
 calculations, only log gaussing broadening will be used. */
void SpectrumRealFreq::continuous_old(DENSFUNC &densitypos,
			      DENSFUNC &densityneg)
{
  XYFUNC ElogE; // Look-up table of energies and logs of energies.
  unsigned int nr = 0; // All frequency points
  unsigned int nrloggauss = 0; // Frequency points where loggaussian broadening is used
  const double broaden_min = get_broaden_min();
  for (double E = P::broaden_max; E > broaden_min; E /= P::broaden_ratio) {
    densitypos.push_back(make_pair(E, 0));
    densityneg.push_back(make_pair(-E, 0));
    ElogE.push_back(make_pair(E, log(E)/P::loggauss_b)); // !!
    // below energy limitLL, simple Gaussian broadening will be used
    if (E > P::limitLL)
       nrloggauss++;
    nr++;
  }

  continuous_old1(densitypos, densityneg, ElogE, fspos, nr, nrloggauss);
  continuous_old1(densityneg, densitypos, ElogE, fsneg, nr, nrloggauss);
}

t_weight SpectrumRealFreq::continuous_lorentz1(double E)
{
   const double eta = P::eta;
   
   weight_bucket sum;
   
   FOREACH(const Spikes::value_type &i, fspos) {
      const t_weight weight = assert_isfinite(WEIGHT(i));
      const double energy = assert_isfinite(ENERGY(i));
      
      const t_weight A = weight * BR_LORENTZIAN(eta, E, energy);
      sum += A;
   }

   FOREACH(const Spikes::value_type &i, fsneg) {
      const t_weight weight = assert_isfinite(WEIGHT(i));
      const double energy = -assert_isfinite(ENERGY(i)); // sign!
      
      const t_weight A = weight * BR_LORENTZIAN(eta, E, energy);
      sum += A;
   }

   return sum;
}

void SpectrumRealFreq::continuous_lorentz(DENSFUNC &densitypos,
				  DENSFUNC &densityneg)
{
   const double broaden_min = get_broaden_min();
   for (double E = P::broaden_max; E > broaden_min; E /= P::broaden_ratio) {
      const t_weight A_plus = continuous_lorentz1(+E);
      densitypos.push_back(make_pair(+E, A_plus));
      const t_weight A_minus = continuous_lorentz1(-E);
      densityneg.push_back(make_pair(-E, A_minus));
   }
}

// Note: omega0_ratio is defined similarly as gauss_b as a prefactor of the
// temperature P::T. [omega0_ratio = sqrt(2) gauss_b]
// WARNING: Weichselbaum and von Delft propose to choose omega_0 to be roughly
// a factor of 2 smaller than the smallest energy scale in the problem,
// including the Kondo temperature T_K ("the transition to regular Gaussian
// sets in below omega_0"). cf. cond-mat:0607497
double get_omega0()
{
   return (P::omega0 < 0.0
	   ? P::omega0_ratio * P::T
	   : P::omega0);
}

// Energy mesh for spectral functions
std::vector<double> make_mesh()
{
  const double broaden_min = get_broaden_min();
   
  std::vector<double> vecE; // Energies on the mesh
  for (double E = P::broaden_max; E > broaden_min; E /= P::broaden_ratio) 
    vecE.push_back(E);
    
  return vecE;
}

// New approach to spectral function broadening. cf. A. Weichselbaum and J.
// von Delft, cond-mat/0607497

void SpectrumRealFreq::continuous_wvd(DENSFUNC &densitypos,
			      DENSFUNC &densityneg)
{
  alpha = P::alpha; // GLOBAL!
  omega0 = get_omega0(); // GLOBAL!
   
  std::vector<double> vecE = make_mesh(); // Energies on the mesh

  for (unsigned int j = 0; j < vecE.size(); j++) {
    const double E = vecE[j];

    weight_bucket valpos, valneg;
     
    FOREACH(const Spikes::value_type &i, fspos) {
      const t_weight w = WEIGHT(i);
      const double e = ENERGY(i);
      my_assert(e > 0.0);

      valpos += w * BR_WVD(E, e);
      valneg += w * BR_WVD(-E, e);
    }

    FOREACH(const Spikes::value_type &i, fsneg) {
      const t_weight w = WEIGHT(i);
      const double e = ENERGY(i);
      my_assert(e > 0.0); // attention!

      valneg += w * BR_WVD(-E, -e);
      valpos += w * BR_WVD(E, -e);
    }

    densitypos.push_back(make_pair(E, valpos));
    densityneg.push_back(make_pair(-E, valneg));
   }
}

void SpectrumRealFreq::continuous_new(DENSFUNC &densitypos,
			      DENSFUNC &densityneg)
{
  alpha = P::alpha; // GLOBAL!
  omega0 = get_omega0(); // GLOBAL!
   
  std::vector<double> vecE = make_mesh(); // Energies on the mesh

  for (unsigned int j = 0; j < vecE.size(); j++) {
    const double E = vecE[j];

    weight_bucket valpos, valneg;
     
    FOREACH(const Spikes::value_type &i, fspos) {
      const t_weight w = WEIGHT(i);
      const double e = ENERGY(i);
      my_assert(e > 0.0);

      valpos += w * BR_NEW(E, e);
      valneg += w * BR_NEW(-E, e);
    }

    FOREACH(const Spikes::value_type &i, fsneg) {
      const t_weight w = WEIGHT(i);
      const double e = ENERGY(i);
      my_assert(e > 0.0); // attention!

      valneg += w * BR_NEW(-E, -e);
      valpos += w * BR_NEW(E, -e);
    }

    densitypos.push_back(make_pair(E, valpos));
    densityneg.push_back(make_pair(-E, valneg));
   }
}

// As above, but adapted for problems with a superconducting gap.
// omega0 controls the width of the Gaussians used to broaden
// spectral features in the range [-brcut;+brcut]. All spectral data
// in the intervals [-D:-brcut]U[+brcut;+D] is broadened with
// the modified log-Gaussian kernel.
void SpectrumRealFreq::continuous_newsc(DENSFUNC &densitypos,
			      DENSFUNC &densityneg)
{
  alpha = P::alpha; // GLOBAL!
  brcut = P::brcut; // GLOBAL!
  omega0 = get_omega0(); // GLOBAL!
   
  std::vector<double> vecE = make_mesh(); // Energies on the mesh

  for (unsigned int j = 0; j < vecE.size(); j++) {
    const double E = vecE[j];

    weight_bucket valpos, valneg;
     
    FOREACH(const Spikes::value_type &i, fspos) {
      const t_weight w = WEIGHT(i);
      const double e = ENERGY(i);
      my_assert(e > 0.0);

      valpos += w * BR_NEWSC(E, e);
      valneg += w * BR_NEWSC(-E, e);
    }

    FOREACH(const Spikes::value_type &i, fsneg) {
      const t_weight w = WEIGHT(i);
      const double e = ENERGY(i);
      my_assert(e > 0.0); // attention!

      valneg += w * BR_NEWSC(-E, -e);
      valpos += w * BR_NEWSC(E, -e);
    }

    densitypos.push_back(make_pair(E, valpos));
    densityneg.push_back(make_pair(-E, valneg));
   }
}

// Save binary raw (binned) spectral function. If using complex
// numbers and P::reim==true, we save triplets (energy,real part,imag
// part). To obtain the true spectral function, one then needs to
// take the "imag part" column, perform a Kramers Kronig transform of
// this data, and add it to the "real part" column. 
// 
// XXX: would it make more sense to change the nomenclature and use
// Green's functions as fundamental objects rather than spectral
// functions??
void SpectrumRealFreq::savebins()
{
   if (!P::savebins)
     return;

   ofstream Fbins;
   safeopen(Fbins, filename + ".bin", true); // true=binary!

   FOREACH(const Spikes::value_type &i, fspos) {
      const double w = real_part(WEIGHT(i));
      const double e = ENERGY(i);
      my_assert(e > 0.0);
      Fbins.write((char*)&e, sizeof(double));
      Fbins.write((char*)&w, sizeof(double));
      if (P::reim) {
	 const double wi = imag_part(WEIGHT(i));
	 Fbins.write((char*)&wi, sizeof(double));
      }
   }

   FOREACH(const Spikes::value_type &i, fsneg) {
      const double w = real_part(WEIGHT(i));
      const double e = -ENERGY(i);
      my_assert(e < 0.0); // attention!
      Fbins.write((char*)&e, sizeof(double));
      Fbins.write((char*)&w, sizeof(double));
      if (P::reim) {
	 const double wi = imag_part(WEIGHT(i));
	 Fbins.write((char*)&wi, sizeof(double));
      }
   }
}

void SpectrumRealFreq::trim()
{
  fspos.trim();
  fsneg.trim();
}

void SpectrumTemp::postprocess()
{
   // Save the results
   
   DENSFUNC d;

   // Not really a spectrum: just save the raw results
   for(Spikes::iterator i = fspos.begin(); i != fspos.end(); i++) {
	DENSPOINT xy = make_pair(i->first, i->second);
	d.push_back(xy);
   }  

  sort(d.begin(), d.end(), sortfirst());

  ofstream Fd;
  safeopen(Fd, filename + ".dat");
  save_densfunc(Fd, d, P::reim);
}

void SpectrumRealFreq::continuous()
{
  DENSFUNC densitypos, densityneg;

  // Spectrum is true spectral density: we broaden it, if so requested.
  if (!P::broaden)
     return;
     
  if (string(P::smooth) == "old")
    continuous_old(densitypos, densityneg);
  if (string(P::smooth) == "wvd")
    continuous_wvd(densitypos, densityneg);
  if (string(P::smooth) == "new")
    continuous_new(densitypos, densityneg);
  if (string(P::smooth) == "newsc")
    continuous_newsc(densitypos, densityneg);
  if (string(P::smooth) == "lorentz")
    continuous_lorentz(densitypos, densityneg);
      
  sort(densityneg.begin(), densityneg.end(), sortfirst());
  sort(densitypos.begin(), densitypos.end(), sortfirst());

  ofstream Fdensity;
  safeopen(Fdensity, filename + ".dat");
  save_densfunc(Fdensity, densityneg, P::reim);
  save_densfunc(Fdensity, densitypos, P::reim);
}

// This is mathematical trace, i.e. the sum of the diagonal elements.
CONSTFNC double trace(const DensMatElements &m)
{
  double tr = 0.0;
  FOREACH(const DensMatElements::value_type &i, m) 
    tr += mult(i.first) * trace_real_nochecks(i.second);
  return tr;
}

// Check if the trace of the density matrix equals 'ref_value'.
void check_trace_rho(const DensMatElements &m, double ref_value = 1.0)
{
   const double tr = trace(m);
   if (!num_equal(trace(m), ref_value)) {
      cerr << "check_trace_rho() failed: " << tr << " instead of " 
	<< ref_value << endl;
      exit(1);
   }
}

/* class BaseSpectrum contains all information about calculating the
 spectrum: pointers to the operator data and miscelaneous data, such
 as the spectrum type. Functions calc_specdens() et al. receive an
 object of this type as input. */

typedef boost::shared_ptr<Spectrum> SpectrumPtr;

class BaseSpectrum {
public:
  string name;
  MatrixElements *op1;
  MatrixElements *op2;
  SpectrumPtr spec;
  SPECTYPE spectype;
  int spin;

  void set(SpectrumPtr s, SPECTYPE st) {
    spec = s; 
    spectype = st;
  }
   
  // Compute the continuous representation by delta-peak broadening.
  // TO DO: should this be part of the BaseSpectrum destructor???
  void fn_continuous() {
     cout << "Spectrum: " << name << " " << spectype << endl;
     spec->postprocess();
     cout << endl;
  }
   
  void about() {
     cout << "Spectrum: " << name << " " << spectype << endl;
  }

};

class speclist;
typedef list<speclist*> lsl;
lsl allspectra; // list of list of spectra

class speclist
{
 public:
   typedef list<BaseSpectrum> lBS;
   lBS spectra;
   
   speclist() { 
      allspectra.push_back(this); 
   }
   operator lBS &() { 
      return spectra; 
   }
   typedef lBS::iterator iterator;
   typedef lBS::const_iterator const_iterator;
   typedef lBS::value_type value_type;
   iterator begin() { 
      return spectra.begin(); 
   }
   iterator end() { 
      return spectra.end(); 
   }
   void push_back(BaseSpectrum &bs) { 
      spectra.push_back(bs); 
   }
   
   // Broaden spectra, close spectral files and deallocate all data storage.
   void spec_continuous_clear() {
      FOREACH(lBS::value_type &i, spectra)
	  i.fn_continuous();
      spectra.clear();
   }
   
   void about() {
      FOREACH(lBS::value_type &i, spectra)
	  i.about();
   } 
};
    
speclist spectraD, spectraS, spectraT, spectraQ, spectraGT, 
  spectraI1T, spectraI2T, spectraK, spectraCHIT, spectraC, spectraOT;
   
// Determine the "effective" dimension of the subspace, i.e. the range
// of indexes over which we loop [kept states!] in calculating
// spectral functions.
// XXX: remove this function!
int get_loop_dim(SPECTYPE spectype, const Eigen &diagI, const Invar &I)
{
  // Number of states in the subspace, the actual numbers depends on
  // whether strategy=all or strategy=kept!
  unsigned int dim = diagI.getnr();
  
  if (dm_spec(spectype)) {
    // Reduced density matrix is defined over a limited range of states due
    // to the truncation (i.e. only over kept states!).
    // TO DO: get this information from elsewhere (i.e. write a global object 
    // for all matrix dimensions for all subspaces and all iterations).
    const unsigned int dim_red = rho[I].size2();
    if (!LAST_ITERATION()) { // exception.. see P::lastall parameter.
       my_assert(dim_red <= dim);
    }
    // TO DO: CLEANUP!
    const unsigned int dim_red2 = dm[STAT::N][I].kept;
    if (!LAST_ITERATION())
      my_assert(dim_red == dim_red2);
    
    dim = min(dim, dim_red);
  }
   
  return dim;
}

/**** CALCULATION OF SPECTRAL FUNCTIONS ****/

class CorrelatorFactorFnc 
{
 public:
   t_factor operator() (const Invar &Ip, const Invar &I1)
     {
	// For odd-parity operators, the quantum numbers may change, but
	// the multiplicity should be the same for bra and ket subspaces.
	my_assert(mult(I1) == mult(Ip));
	return mult(I1); // S_z unchanged by singlet operators
     }
};

class trivialCheckSpinFnc
{
 public:
   bool operator() (const Invar &, const Invar &, int) { return true; }
};

class SpecdensFactorFnc 
{
 public:
   t_factor operator() (const Invar &Ip, const Invar &I1) {
     t_factor factor = Sym->specdens_factor(Ip, I1);
     return assert_isfinite(factor);
   }
};

class SpecdensquadFactorFnc 
{
 public:
   t_factor operator() (const Invar &Ip, const Invar &I1) {
     t_factor factor = Sym->specdensquad_factor(Ip, I1);
     return assert_isfinite(factor);
   }
};

class SpecdensCheckSpinFnc
{
 public:
   bool operator() (const Invar &I1, const Invar &Ip, int SPIN) { 
     return Sym->check_SPIN(I1, Ip, SPIN); 
   }
};

class SpinSuscFactorFnc
{
 public:
   t_factor operator() (const Invar &Ip, const Invar &I1) { 
      t_factor factor = Sym->dynamicsusceptibility_factor(Ip, I1);
      return assert_isfinite(factor);
   }
};					

class OrbSuscFactorFnc
{
 public:
   t_factor operator() (const Invar &Ip, const Invar &I1) { 
      t_factor factor = Sym->dynamic_orb_susceptibility_factor(Ip, I1);
      return assert_isfinite(factor);
   }
};					

void doublet_check_norm(CustomOp::value_type &op,
			const DiagInfo &diag, 
			int SPIN)
{
   weight_bucket sum;
   
   LOOP_const(diag, isp) {
      const Invar Ip = INVAR(isp);
    
      LOOP_const(diag, is1) {
	 const Invar I1 = INVAR(is1);

	 const Twoinvar II = make_pair(I1, Ip);
	 if (op.second.count(II)) {
	    Invar I1, Ip;
	    tie(I1, Ip) = II;
	    
	    if (!Sym->check_SPIN(I1, Ip, SPIN))
	      continue;

	    t_factor spinfactor = Sym->specdens_factor(Ip, I1);
	    
	    Matrix & mat = op.second[II];
	    const unsigned int dim1 = mat.size1();
	    const unsigned int dimp = mat.size2();
	    
	    for (unsigned int r1 = 0; r1 < dim1; r1++) 
	       for (unsigned int rp = 0; rp < dimp; rp++) 
		  sum += spinfactor * sqr(abs(mat(r1, rp)));
	    // TO DO: generalize this for complex elements!
	 }
      }
   }

   const double result = 2.0 * real_part(sum);
   
   // Factor 2: Tr[d d^\dag + d^\dag d] = 2 \sum_{i,j} A_{i,j}^2 !!
   cout << "check_norm[" << NAME(op) << "]=" << result << endl;
}

void quadruplet_check_norm(CustomOp::value_type &op,
			const DiagInfo &diag, 
			int SPIN)
{
   weight_bucket sum;
   
   LOOP_const(diag, isp) {
      const Invar Ip = INVAR(isp);
    
      LOOP_const(diag, is1) {
	 const Invar I1 = INVAR(is1);

	 const Twoinvar II = make_pair(I1, Ip);
	 if (op.second.count(II)) {
	    Invar I1, Ip;
	    tie(I1, Ip) = II;
	    
	    if (!Sym->check_SPIN(I1, Ip, SPIN))
	      continue;

	    t_factor spinfactor = Sym->specdensquad_factor(Ip, I1);
	    
	    Matrix & mat = op.second[II];
	    const unsigned int dim1 = mat.size1();
	    const unsigned int dimp = mat.size2();
	    
	    for (unsigned int r1 = 0; r1 < dim1; r1++) 
	       for (unsigned int rp = 0; rp < dimp; rp++) 
		  sum += spinfactor * sqr(abs(mat(r1, rp)));
	    // TO DO: generalize this for complex elements!
	 }
      }
   }

   const double result = 2.0 * real_part(sum);
   
   // Factor 2: Tr[d d^\dag + d^\dag d] = 2 \sum_{i,j} A_{i,j}^2 !!
   cout << "check_norm[" << NAME(op) << "]=" << result << endl;
}

// GENERAL NOTE: CONJ_ME() and reversal of indexes is appropriate for
// spectral functions of fermionic operators, since we need to
// distinguish between d and d^\dag. For generic G_{AB}, this is an
// unnecessary complication and confusing. For real symmetric
// operators, the result is not affected. For complex Hermitian, the
// same is true. But it is ugly.

// For argument 'sign' in calc_generic_* functions. This is the only
// difference between the GFs for bosonic and fermionic operators.
const double S_FERMIONIC = -1.0;
const double S_BOSONIC = 1.0;
	     
#include "spec_FT.cc"
#include "spec_DMNRG.cc"
#include "spec_CFSold.cc"
#include "spec_CFSnew.cc"
#include "spec_FDMold.cc"
#include "spec_FDMnew.cc"

#include "spec.cc"

InvarVec input_subspaces()
{
   return In;
}

void init_rho_FDM(DensMatElements &rho, int N);

#include "dmnrg.h"
   
// **** Helper functions for the NRG RUN ****

void copy_sort_energies(const DiagInfo &diag, STDEVEC &energies)
{
  energies.clear();
   
  LOOP_const(diag, is) 
     energies.insert(energies.end(),
		     EIGEN(is).value.begin(), 
		     EIGEN(is).value.end());

  sort(energies.begin(), energies.end());
}

#include "splitting.cc"

// Number of eigenstates in subspace I at the previous (!) iteration.
int size_subspace_prev(const Invar &I)
{
  return diagprev[I].getnr();
}

// Generate an info string for headers
void infostring()
{
  string info = " ***** [" + (string)(nrgrun ? "NRG" : "DM") + "] " +
    "Iteration " + tostring(STAT::N+1) + "/" + tostring(P::Nmax) +
    " (scale " + tostring(STAT::scale) + ")" + " ***** ";

  if (P::substeps) {
     int Ntrue, M;
     tie(Ntrue, M) = get_Ntrue_M(STAT::N);
     info += " step " + tostring(Ntrue+1) + " substep " + tostring(M+1);
  }

  cout << endl << info << endl;
}

// TODO: generalize to all symmetry types! (i.e. add additional
// coefficients of interest). This should ideally be a call to a
// method in class Symmetry.
void show_coefficients()
{
   if (!P::substeps) {
      using namespace STAT;
      for (unsigned int i = 0; i < P::coefchannels; i++) {
	 cout << "[" << i+1 << "]"
	      << " xi(" << N << ")=" << xi(N, i)
	      << " xi_scaled(" << N << ")=" << xi(N, i)/SCALE(N+1)
	      << " zeta(" << N+1 << ")=" << zeta(N+1, i)
	      << endl;
      }
   } else {
     int Ntrue, M;
     tie(Ntrue, M) = get_Ntrue_M(STAT::N);

     for (unsigned int i = 0; i < P::coeffactor; i++) {
	int index = M + P::channels * i;
	cout << "[" << index << "]" 
	     << " xi(" << Ntrue << ")=" << xi(Ntrue, index)
	     << " zeta(" << Ntrue+1 << ")=" << zeta(Ntrue+1, index)
	     << endl;
     }
   }
   Sym->show_coefficients();
}

// Information about ancestor subspaces
void logancestors(const Invar &I, const InvarVec &In, const Rmaxvals &rmx)
{
  if (logletter('s')) { 
    cout << "Ancestors of (" << I << "): ";
    for (unsigned int i = 1; i <= P::combs; i++) {
      if (rmx[i] > 0) 
	 cout << "(" << In[i] << ", dim=" << rmx[i] << ") ";
    }
    cout << endl;
  }
}

// To be called before functions find_groundstate() and
// subtract_groundstate_energy(), preferably right after the
// diagonalizations, so that we retain information about the
// unshifted values.
void copy_unshifted_eigenvalues(DiagInfo &diag)
{
   LOOP(diag, is)
      EIGEN(is).value_raw = EVEC( EIGEN(is).value );
}

// Find the ground state in the current NRG shell. We use value_raw
// vector for this purpose, because it contains the unshifted values.
void find_groundstate(const DiagInfo &diag)
{
  STAT::Egs = DBL_MAX;

  LOOP_const(diag, is) {
    my_assert(EIGEN(is).value_raw.size() > 0);
     
    const t_eigen Emin = EIGEN(is).value_raw(0); // Eigenvalues are sorted
    if (Emin < STAT::Egs) { 
      STAT::Egs = Emin;
      STAT::ground = INVAR(is);
    }
  }
}

// Dump all energies in diag to a file
void dumptofile(const DiagInfo &diag, ostream &file)
{
  file << endl << "===== Iteration number: " << STAT::N << endl;

  LOOP_const(diag, is) {
    file << "Subspace: " << is.first << endl;
    for_each(EIGEN(is).value.begin(), EIGEN(is).value.end(), file << _1 << ' ');
    file << endl;
  }
}

/*** Subtract ground state energy ***/
void subtract_groundstate_energy(DiagInfo &diag)
{
   LOOP(diag, is) {
      my_assert(EIGEN(is).value_raw.size() != 0);
      my_assert(EIGEN(is).value_raw.size() == EIGEN(is).value.size());
      transform(EIGEN(is).value_raw.begin(), EIGEN(is).value_raw.end(), 
		EIGEN(is).value.begin(), _1 - STAT::Egs);
      my_assert(EIGEN(is).value[0] >= 0);
   }
}

// Calculate the (on-shell) statistical sum using the excitation
// energies relative to the current step energy scale.
// Used in nrg_measure_singlet().
double calc_Z(const DiagInfo &diag)
{
  bucket Z;
  LOOP_const(diag, is)
    for (unsigned int r = 0; r < NRSTATES(is); r++) 
      Z += mult(INVAR(is)) * exp(-P::betabar * EIGEN(is).value(r) );
  return Z;
}

/* Function newcombination_allowed() checks if states |q,ss,i>_{N+1} can be
 formed for a given spin. Some combinations might be forbidden even when
 the corresponding |q',ss'> subspaces exist at iteration step N (Invar In).
 */
bool newcombination_allowed(int i, const Invar &I, const Invar &In)
{
   return Sym->triangle_inequality(I, In, QN[i]);
};

// Determine the ranges of index r
void Rmaxvals::determine_ranges(const Invar &I, const InvarVec &In)
{
  IVEC rmx(P::combs);
  rmx.clear();
  for (unsigned int i = 0; i < P::combs; i++) 
    if (newcombination_allowed(i+1, I, In[i+1]))
      rmx[i] = size_subspace_prev(In[i+1]);
  store(rmx);
}

// *********************************** NRG RUN **********************************

ofstream F; // all energies (different file for NRG and for DMNRG)
ofstream Ftd; // magnetic and charge susceptibility
ofstream Fcustom; // expectation values
ofstream Fcustomfdm; // expectation values at T (FDM algorithm)
ofstream Fannotated; // annotated eigenvalue spectrum

// Construct the suffix of the filename for spectral density files: 'A_?-A_?'.
// If SPIN == 1 or SPIN == -1, '-u' or '-d' is appended to the string.
string SDNAME(string a, string b, int SPIN = 0)
{
  string name = a + "-" + b;
  if (SPIN ==  1) name = name + "-u";
  if (SPIN == -1) name = name + "-d";
  return name;
}
 
// Formatted output of the computed expectation values
class ExpvOutput
{
 private:
   ostream &F; // output stream
   map<string, t_expv> &m; // reference to the name->value mapping
   list<string> fields; // list of fields to be output (may be a subset
                        // of the fields actually present in m)
  
   // Consecutive numbers for the columns
   void field_numbers() {
      F << "#";
      formatted_output(F, 1);
      for(unsigned int ctr = 1; ctr <= fields.size(); ctr++)
	 formatted_output(F, 1 + ctr);
      F << endl;
   }   
   
   // Label and field names. Label is the first column (typically the
   // temperature).
   void field_names(string labelname = "T") {
      F << "#";
      formatted_output(F, labelname);
      FOREACH(list<string>::value_type &op, fields)
	formatted_output(F, op);
      F << endl;
   }
   
 public:
   // Output the current values for the label and for all the fields
   void field_values(double labelvalue = STAT::Teff) {
      F << " ";
      formatted_output(F, labelvalue);
      FOREACH(list<string>::value_type &op, fields)
	formatted_output(F, m[op]);
      F << endl;
   }

   ExpvOutput(ostream &_F, 
	      map<string, t_expv> &_m,
	      list<string> &_fields) 
     : F(_F), m(_m), fields(_fields) {
      field_numbers();
      field_names();
   }
};

ExpvOutput *custom, *customfdm;

// Prepare "td" for output: write header with parameters and
// a line with field names.
void open_Ftd(ofstream &Ftd)
{
  Ftd.open(FN_TD);
  outfield::width = P::width_td;
  outfield::prec = P::prec_td;
  TD::save_header(Ftd);
}

string get_specfile(string prefix, SPECTYPE spectype, string name)
{
  // Without suffix .dat! The suffix is added in Spectrum() constructor.
  return prefix + "_" + spectypestring(spectype) + "_dens_"  + name;
}

void open_files_spec_RealFreq(speclist &sl,
                       BaseSpectrum &spec, 
                       string name,
                       string prefix,
                       SPECTYPE spectype)
{
  nrglog('c', "Spectrum (real freq): " << 
    prefix << "-" << spectype << "-" <<  name);
  SpectrumPtr sp(new SpectrumRealFreq(name, 
			      get_specfile(prefix, spectype, name), 
			      spectype));
  spec.set(sp, spectype);
  sl.push_back(spec);
}

void open_files_spec_Temp(speclist &sl,
                       BaseSpectrum &spec, 
                       string name,
                       string prefix,
                       SPECTYPE spectype)
{
  nrglog('c', "Spectrum (temperature): " << 
    prefix << "-" << spectype << "-" <<  name);
  SpectrumPtr sp(new SpectrumTemp(name, 
			      get_specfile(prefix, spectype, name), 
			      spectype));
  spec.set(sp, spectype);
  sl.push_back(spec);
}

// open_files_spec() opens the output files and establishes the data structures
// for storing spectral information.
void open_files_spec(speclist &sl,
		     BaseSpectrum &spec, 
		     string name,
		     MatrixElements &op1,
		     MatrixElements &op2,
		     string prefix)
{
   spec.name = name;
   spec.op1 = &op1;
   spec.op2 = &op2;

   if (prefix == "gt") {
     if (nrgrun)
	open_files_spec_Temp(sl, spec, name, prefix, SPEC_GT);
     return;
   }

   if (prefix == "i1t") {
     if (nrgrun) 
	open_files_spec_Temp(sl, spec, name, prefix, SPEC_I1T);
     return;
   }

   if (prefix == "i2t") {
     if (nrgrun)
	open_files_spec_Temp(sl, spec, name, prefix, SPEC_I2T);
     return;
   }

   if (prefix == "chit") {
     if (nrgrun) 
	open_files_spec_Temp(sl, spec, name, prefix, SPEC_CHIT);
     return;
   }

   // If we did not return from this funciton by this point, what we are
   // computing is the spectral function. There are several possibilities 
   // in this case..

   if (nrgrun && P::finite) 
     open_files_spec_RealFreq(sl, spec, name, prefix, SPEC_FT);
   
   if (dmnrgrun && P::dmnrg)
     open_files_spec_RealFreq(sl, spec, name, prefix, SPEC_DMNRG);
   
   if (dmnrgrun && P::cfsold)
     open_files_spec_RealFreq(sl, spec, name, prefix, SPEC_CFSold);

   if (dmnrgrun && P::cfs)
     open_files_spec_RealFreq(sl, spec, name, prefix, SPEC_CFS);

   if (dmnrgrun && P::cfsgt)
     open_files_spec_RealFreq(sl, spec, name, prefix, SPEC_CFSgt);

   if (dmnrgrun && P::cfsls)
     open_files_spec_RealFreq(sl, spec, name, prefix, SPEC_CFSls);

   if (dmnrgrun && P::fdmold)
     open_files_spec_RealFreq(sl, spec, name, prefix, SPEC_FDMold);

   if (dmnrgrun && P::fdm)
     open_files_spec_RealFreq(sl, spec, name, prefix, SPEC_FDM);

   if (dmnrgrun && P::fdmgt)
     open_files_spec_RealFreq(sl, spec, name, prefix, SPEC_FDMgt);

   if (dmnrgrun && P::fdmls)
     open_files_spec_RealFreq(sl, spec, name, prefix, SPEC_FDMls);
}

// TO DO: replace macros with function calls!

#define OPENSPEC(stringtoken, spectra, infix, rectype1, rectype2, Spin) \
    { BaseSpectrum spec; \
      spec.spin = Spin; \
      const string name = SDNAME(NAME(op1), NAME(op2), Spin); \
      if (stringtoken.find(name)) { \
	open_files_spec(spectra, spec, name, op1.second, op2.second, infix); \
        oprecalc::rectype1.insert(NAME(op1)); \
        oprecalc::rectype2.insert(NAME(op2)); \
      } \
    }

#define LOOPOVER(set1, set2, job) \
   FOREACH(CustomOp::value_type &op1, set1) \
     FOREACH(CustomOp::value_type &op2, set2) { job; }
   
namespace oprecalc {

  /* The following lists hold the names of operators which need to be
   recomputed. The default behavior is to recompute all the operators
   that are required to calculate the requested spectral densities, see
   function open_files(). In addition, singlet operators are always
   recomputed in the first NRG run, so that we can calculate the
   expectation values. In addition, if fdmexpv=true, the singlet operators
   are also recomputed in the second run if fdmexpvn=-1. */
  set<string> s, p, g, d, v, t, q, ot;

  void clear(void) {
    s.clear();
    p.clear();
    g.clear();
    d.clear();
    v.clear();
    t.clear();
    q.clear();
    ot.clear();
  }

  void report(ostream &F, string name, const set<string> &x) {
    F << name << "=[";
    for_each(x.begin(), x.end(), F << _1 << ' ');
    F << "]" << endl;
  }

  void report(ostream &F = cout) {
    F << "Computing the following operators:" << endl;
    report(F, "s", s);
    report(F, "p" ,p);
    report(F, "g", g);
    report(F, "d", d);
    report(F, "v", v);
    report(F, "t", t);
    report(F, "q", q);
    report(F, "ot", ot);
  }

  bool do_s(const string &name) {
    if (nrgrun) return true;
    if (P::fdmexpv && STAT::N <= P::fdmexpvn) return true; // Calculate <O> using FDM algorithm
    return s.count(name);
  }

  bool do_g(const string &name) {
    if (nrgrun) return true;
    if (P::fdmexpv && STAT::N <= P::fdmexpvn) return true; // Calculate <O> using FDM algorithm
    return g.count(name);
  }

  bool do_p(const string &name) { return p.count(name); }
  bool do_d(const string &name) { return d.count(name); }
  bool do_v(const string &name) { return v.count(name); }
  bool do_t(const string &name) { return t.count(name); }
  bool do_q(const string &name) { return q.count(name); }
  bool do_ot(const string &name) { return ot.count(name); }

  // Reset lists of operators which need to be iterated.
  void reset_operator_lists_and_open_spectrum_files() {
     oprecalc::clear();

     // Correlators (singlet operators of all kinds).
     string_token sts(P::specs);
     LOOPOVER(a.ops,  a.ops,  OPENSPEC(sts, spectraS, "corr", s, s, 0));
     LOOPOVER(a.opsp, a.opsp, OPENSPEC(sts, spectraS, "corr", p, p, 0));
     LOOPOVER(a.opsg, a.opsg, OPENSPEC(sts, spectraS, "corr", g, g, 0));
     LOOPOVER(a.ops,  a.opsg, OPENSPEC(sts, spectraS, "corr", s, g, 0));
     LOOPOVER(a.opsg, a.ops,  OPENSPEC(sts, spectraS, "corr", g, s, 0));
     
     // Global susceptibilities (global singlet operators).
     string_token stchit(P::specchit);
     LOOPOVER(a.ops, a.ops,   OPENSPEC(stchit, spectraCHIT, "chit", s, s, 0));
     LOOPOVER(a.ops, a.opsg,  OPENSPEC(stchit, spectraCHIT, "chit", s, g, 0));
     LOOPOVER(a.opsg, a.ops,  OPENSPEC(stchit, spectraCHIT, "chit", g, s, 0));
     LOOPOVER(a.opsg, a.opsg, OPENSPEC(stchit, spectraCHIT, "chit", g, g, 0));

     // Dynamic spin susceptibilities (triplet operators).
     string_token stt(P::spect);
     LOOPOVER(a.opt, a.opt, OPENSPEC(stt, spectraT, "spin", t, t, 0));

     string_token stot(P::specot);
     LOOPOVER(a.opot, a.opot, OPENSPEC(stot, spectraOT, "orbspin", ot, ot, 0));

     const int varmin = (Sym->isfield() ? -1 : 0);
     const int varmax = (Sym->isfield() ? +1 : 0);

     // Spectral functions (doublet operators).
     string_token std(P::specd);
     for (int SPIN = varmin; SPIN <= varmax; SPIN += 2)
       LOOPOVER(a.opd, a.opd, 
		OPENSPEC(std, spectraD, "spec", d, d, SPIN));
     
     string_token stgt(P::specgt);
     for (int SPIN = varmin; SPIN <= varmax; SPIN += 2)
       LOOPOVER(a.opd, a.opd,
		OPENSPEC(stgt, spectraGT, "gt", d, d, SPIN));

     string_token sti1t(P::speci1t);
     for (int SPIN = varmin; SPIN <= varmax; SPIN += 2)
       LOOPOVER(a.opd, a.opd,
		OPENSPEC(sti1t, spectraI1T, "i1t", d, d, SPIN));

     string_token sti2t(P::speci2t);
     for (int SPIN = varmin; SPIN <= varmax; SPIN += 2)
       LOOPOVER(a.opd, a.opd,
		OPENSPEC(sti2t, spectraI2T, "i2t", d, d, SPIN));

     // Spectral functions (quadruplet operators).
     string_token stq(P::specq);
     LOOPOVER(a.opq, a.opq, OPENSPEC(stq, spectraQ, "specq", q, q, 0));

     oprecalc::report();
     
     cout << endl << "Computing the following spectra:" << endl;
     FOREACH(lsl::value_type &i, allspectra) 
       i->about();

  }
};

// Open output files and write headers
void open_output_files()
{
  // We dump all energies to separate files for NRG and DM-NRG
  // runs. This is a very convenient way to check if both runs produce
  // the same results.
  if (P::dumpenergies)
     F.open((nrgrun ? FN_ENERGIES_NRG : FN_ENERGIES_DMNRG) );

  if (nrgrun) {
    open_Ftd(Ftd);

    if (P::dumpannotated)
      Fannotated.open(FN_ANNOTATED);
  }
   
  list<string> ops;
  FOREACH(CustomOp::value_type &op, a.ops)
     ops.push_back(NAME(op));
  FOREACH(CustomOp::value_type &op, a.opsg)
     ops.push_back(NAME(op));

  if (nrgrun) {
    // Singlet operator expectation values. Generate header with names
    // of operators.
    Fcustom.open(FN_CUSTOM);

    custom = new ExpvOutput(Fcustom, STAT::expv, ops);
  }
   
  if (dmnrgrun && P::fdmexpv) {
    Fcustomfdm.open(FN_CUSTOMFDM);

    customfdm = new ExpvOutput(Fcustomfdm, STAT::fdmexpv, ops);
  }
   
  oprecalc::reset_operator_lists_and_open_spectrum_files();
}

// Close files. This has to be called explicitly, because there can be two
// separate runs (NRG and DMNRG). No error checking is done, nor do we test
// if the files are actually open.
void close_output_files()
{
   if (nrgrun) {
      F.close();
      Ftd.close();
      Fannotated.close();
      Fcustom.close();
   }
   
   if (dmnrgrun) {
      Fcustomfdm.close();
   }
}

// DM-NRG: initialization of the density matrix -----------------------------

// Calculate grand canonical partition function at current NRG energy
// shell. This is not the same as the true partition function of the
// full problem! Instead this is the Z_N that is used to initialize
// the density matrix, i.e. rho = 1/Z_N \sum_{l} exp{-beta E_l} |l;N>
// <l;N|.  calc_grand_canonical_Z() is also used to calculate
// STAT::Zft, that is used to compute the spectral function with the
// conventional approach, as well as STAT::Zgt for G(T) calculations,
// STAT::Zchit for chi(T) calculations.
double calc_grand_canonical_Z(const DiagInfo &diag, 
			      double temperature = P::T)
{
  bucket ZZ;
  LOOP_const(diag, is) {
     const unsigned int dim = NRSTATES(is);
     const Invar I = INVAR(is);

     for (unsigned int r = 0; r < dim; r++)
       ZZ += mult(I) * exp(-EIGEN(is).value(r) * (STAT::scale / temperature));
  }
  my_assert(ZZ >= 1.0);
  return ZZ;
}

// Calculate rho_N, the density matrix at the last NRG iteration. It is
// normalized to 1. Note: in CFS approach, we consider all states in the
// last iteration to be "discarded".
// For the details on the full Fock space approach see:
// F. B. Anders, A. Schiller, Phys. Rev. Lett. 95, 196801 (2005).
// F. B. Anders, A. Schiller, Phys. Rev. B 74, 245113 (2006).
// R. Peters, Th. Pruschke, F. B. Anders, Phys. Rev. B 74, 245114 (2006).
void init_rho(const DiagInfo &diag, DensMatElements &rho)
{
  nrglog('w', "init_rho()");
   
  const double ZZ = calc_grand_canonical_Z(diag);
  rho.clear();
   
  LOOP_const(diag, is) {
    // The size of the matrix is determined by NRSTATES().
    const unsigned int dim = NRSTATES(is);
    const Invar I = INVAR(is);
     
    rho[I] = Matrix(dim, dim);
    rho[I].clear();
    Matrix & rhoI = rho[I];
    for (unsigned int i = 0; i < dim; i++)
      rhoI(i, i) = exp( -EIGEN(is).value(i) * STAT::scale / P::T )/ZZ;
  }

  const double Tr = trace(rho);
  my_assert(num_equal(Tr, 1.0, 1e-8));
}

// Calculate the shell-N statistical sum as used in the FDM algorithm.
double calc_ZnD_one(int N, double T)
{
  ldouble ZZ = 0;
  for (Subs::const_iterator j = dm[N].begin(); j != dm[N].end(); ++j) {
    const Invar I = j->first;
    // Determine which states count as discarded in the FDM sense
    const unsigned int min = (LAST_ITERATION(N) ? 0 : j->second.kept);
    const unsigned int max = j->second.total;
    
    for (unsigned int i = min; i < max; i++) {
       const long double Eabs = j->second.absenergy[i]-STAT::GSenergy;
       my_assert(Eabs >= 0.0);
       const long double betaE = Eabs/T;
       ZZ += mult(I) * expl(-betaE);
    }
  }
  nrglog('w', "ZnD[" << N << "]=" << ZZ);
  return ZZ;
}

// Calculate rho(N), the shell-N density matrix, computed using
// the discarded states at shell N.
// Must be called AFTER calc_ZnD().
// A. Weichselbaum, J. von Delft, Phys. Rev. Lett. 99, 076402 (2007)
// T. A. Costi, V. Zlatic, Phys. Rev. B 81, 235127 (2010)
// H. Zhang, X. C. Xie, Q. Sun, Phys. Rev. B 82, 075111 (2010)
void init_rho_FDM(DensMatElements &rho, int N)
{
  nrglog('w', "init_rho_FDM(" << N << ")");
   
  double T = P::T;
  double ZZ = STAT::ZnD[N];
   
  long double tr1 = 0.0;
  long double tr2 = 0.0;
  for (Subs::const_iterator j = dm[N].begin(); j != dm[N].end(); ++j) {
    const Invar I = j->first;
    const unsigned int min = (LAST_ITERATION(N) ? 0 : j->second.kept);
    const unsigned int max = j->second.total;
    
    rho[I] = Matrix(max, max);
    rho[I].clear();
    Matrix & rhoI = rho[I];
    for (unsigned int i = min; i < max; i++) {
       const long double Eabs = j->second.absenergy[i]-STAT::GSenergy;
       my_assert(Eabs >= 0.0);
       const long double betaE = Eabs/T;
       long double val1 = expl(-betaE) / ZZ;
       const long double ratio = STAT::wn[N] / ZZ;
       long double val2 = expl(-betaE) * ratio;
       val1 = isfinite(val1) ? val1 : 0.0;
       val2 = isfinite(val2) ? val2 : 0.0;
       rhoI(i, i) = double(val2);
       tr1 += mult(I) * val1;
       tr2 += mult(I) * val2;
    }
  }
   
  // Sanity checks: the trace should be equal to the total weight of
  // the shell-N contribution to the full density matrix.
  const double diff = (tr2-STAT::wn[N])/STAT::wn[N]; // relative error
  if (isfinite(tr1) && isfinite(diff)) {
     if (!num_equal(tr1, 1.0))
       my_warning("tr1=%24.16Lf", tr1); // This is not fatal! (but weird)
     if (!num_equal(diff, 0.0, 1e-8)) {
	my_warning("diff=%24.16lf", diff); // This is more serious..
	my_assert(STAT::wn[N] < 1e-12); // ..but OK if small enough overall.
     }
  }
}

/*** Truncation ***/

// Determine the number of states to be retained.
// Returns Emax - the highest energy to still be retained.
t_eigen highest_retained_energy(const STDEVEC & energies)
{
  my_assert(energies[0] == 0.0); // check for the subtraction of Egs
  const unsigned int totalnumber = energies.size();
   
  unsigned int nrkeep;

  if (P::keepenergy <= 0.0) {
    nrkeep = P::keep;
  } else {
    double keepenergy = P::keepenergy;

    // We add 1 for historical reasons. We thus keep states with E<=Emax,
    // and one additional state which has E>Emax.
    nrkeep = 1 + count_if(energies.begin(), energies.end(),
			  _1 <= keepenergy);

    nrkeep = CLIP(nrkeep, P::keepmin, P::keep);
  }
       
  // Check for near degeneracy and ensure that the truncation occurs in a
  // "gap" between nearly-degenerate clusters of eigenvalues.
  if (P::safeguard > 0.0) {
    unsigned int cnt_extra = 0;
    while (nrkeep < totalnumber &&
	   (energies[nrkeep] - energies[nrkeep-1]) <= P::safeguard &&
	   cnt_extra < P::safeguardmax) {
       nrkeep++;
       cnt_extra++;
    }
    if (cnt_extra) 
      debug("Safeguard: keep additional " << cnt_extra <<" states");
  }
  
  nrkeep = CLIP(nrkeep, 1, totalnumber);
  return energies[nrkeep-1];
}

/* This function doesn't really perform the truncation in the sense of 
 deleting matrix elements. It just computes the number of states to
 keep for each invariant subspace. Truncation is performed by
 nrg_trucate_perform().
 */
void nrg_truncate_prepare(DiagInfo &diag)
{
  nrglog('@', "@ nrg_truncate_prepare()");

  t_eigen Emax = highest_retained_energy(STAT::energies);

  unsigned int nrkept = 0; // counter of states actually retained
  unsigned int nrkeptmult = 0; // counter of states actually retained, 
                        // taking into account the multiplicity 

  LOOP(diag, is) {
    const Invar I = INVAR(is);
    const unsigned int multiplicity = mult(INVAR(is));
    const unsigned int nrc = NRSTATES(is); // number of (calculated) states in the subspace
    my_assert(nrc > 0);

    // Count the number of elements to keep
    unsigned int count = count_if(EIGEN(is).value.begin(),
				  EIGEN(is).value.end(),
				  _1 <= Emax);
    my_assert(count <= nrc);

    // In CFS=true, we keep all states in the last iteration. 
    if (cfs_flags() && LAST_ITERATION() && !P::lastalloverride)
       count = nrc;

    /* Exception: in the last iteration of the density-matrix run, we may
    override strategy=kept and keep all (computed) states if
    lastall=true. This is only relevant if CFS=false and DMNRG=true. */
    if (dmnrgrun && LAST_ITERATION() && P::lastall)
       count = nrc;

    if (count == nrc && EIGEN(is).value(nrc-1) != Emax) {
       /* We determined that all calculated states in this invariant
        subspace will be retained (and that perhaps additional should
        have been computed).
      
        RMAX() gives actual dimensionality of the invariant subspace
        (i.e. the maximal number of eigenpairs), while the variable nrc
        given by NRSTATES() is the number of actually calculated states, which
        is only equal to the dimensionality if in the computational
        scheme used all the states are retained. */
       const unsigned int truedim = RMAX(is);

       /* When using partial diagonalisation, the value of diagratio may
        be too low and states within some energy window may be missing
        from some invariant subspaces, but not from others. This might
        systematically bias the results, so we signal a warning if this
        occurs. */
       if (nrc < truedim) {
         NRG::notenough = true;
       }
    }
     
    // In the last iteration we count as "discarded" all states! Note the
    // semantic difference of the term discard applied in this context..
    const unsigned int lower = (LAST_ITERATION() ? 0 : count);
    
    // loop over all discarded states
    for (unsigned int i = lower; i < nrc; i++) {
       // absolute energy at current shell
       t_eigen absE = (STAT::Egs + EIGEN(is).value[i]) * STAT::scale;
       // sum of all previous energy shifts
       absE += STAT::totalenergy;
       
       // NOTE: when nrg_truncate_prepare() is called from
       // nrg_do_diag(), Egs*scale has not yet been added to
       // totalenergy. The accumulation is performed in the beginning
       // of the procedure nrg_after_diag().
    }

    diag[I].truncate_prepare(count);
       
    nrkept += count;
    nrkeptmult += count * multiplicity;
  }
   
  nrgdump3(Emax, nrkept, nrkeptmult) << endl;
}

// Calculate partial statistical sums, ZnD, and the grand canonical Z
// (STAT::ZZ). calc_ZnD() must be called before the second NRG run.
void calc_ZnD()
{
  // Partial statistical sums, computed with respect to absolute excitation
  // energies! This implies that the absolute energies have to be used
  // in the calc_generic_FDM() function, since 1/ZnD are used as prefactors.
  STAT::ZnD = std::vector<double>(P::Nmax, 0.0);
  for (unsigned int N = P::Ninit; N < P::Nmax; N++) 
     STAT::ZnD[N] = calc_ZnD_one(N, P::T);

  // Grand-canonical partition function (over all shells!!), i.e. \sum
  // \exp(-beta E) for all (!!) states in the complete Fock space, where
  // E is the absolute energy.
  bucket Z;
  for (unsigned int N = P::Ninit; N < P::Nmax; N++) {
     Z += pow(double(P::combs), int(P::Nmax-N-1)) * STAT::ZnD[N];
  }
  STAT::ZZ = Z;
  cout << "Grand canonical Z= " << HIGHPREC(Z) << endl;
  cout << "Zft (last shell)=  " << HIGHPREC(STAT::Zft) << endl;

  // Weights for specific shells, w_n
  STAT::wn = std::vector<double>(P::Nmax, 0.0);

  bucket sumwn;
  for (unsigned int N = P::Ninit; N < P::Nmax; N++) {
     // This is w_n defined after Eq. (8) in the WvD paper.
     const double wn = pow(double(P::combs), int(P::Nmax-N-1)) * 
        STAT::ZnD[N] / Z;
     nrglog('w', "wn[" << N << "]=" << HIGHPREC(wn));
     STAT::wn[N] = wn;
     sumwn += wn;
  }
  nrglog('w', "sumwn=" << sumwn << " sumwn-1=" << sumwn-1.0);

  // Check the sum-rule.
  my_assert(num_equal(sumwn, 1.0));

  // These ratios enter the terms for the spectral function.
  STAT::wnfactor = std::vector<double>(P::Nmax, 0.0);
  for (unsigned int N = P::Ninit; N < P::Nmax; N++) {
     const double w = STAT::wn[N]/STAT::ZnD[N];
     STAT::wnfactor[N] = (!isnan(w) ? w : 0.0);
     nrglog('w', "wnfactor[" << N << "]=" << STAT::wnfactor[N]);
  }
}

// Actually truncate matrices. Drop subspaces with no states kept.
void nrg_truncate_perform(DiagInfo &diag)
{
  for (DiagInfo::iterator i = diag.begin(); i != diag.end(); ++i)
    i->second.truncate_perform(); // Truncate subspace to appropriate size
}

// scaled = true -> output scaled energies (i.e. do not multiply
// by the rescale factor)
inline t_eigen scaled_energy(t_eigen e, 
			     bool scaled = true,
			     bool absolute = false)
{
   return e * (scaled ? 1.0 : STAT::scale) 
     + (absolute ? STAT::totalenergy: 0.0);
}

/* Store (eigenvalue/quantum numbers) pairs at given NRG iteration.
 Eigenenergies are useful for determining RG flows and checking for
 possible phase transitions between various different ground states. */

void dump_annotated(const DiagInfo &diag, 
		    bool scaled = true,
		    bool absolute = false)
{
  typedef pair<t_eigen, Invar> EI_PAIR;
  typedef std::vector<EI_PAIR> EI_VEC;
  EI_VEC seznam;
  seznam.reserve(10000);

  LOOP_const(diag, is)
    FOREACH(t_eigen e, EIGEN(is).value) 
       seznam.push_back(make_pair(e, INVAR(is)));
   
  sort(seznam.begin(), seznam.end());
   
  unsigned int len = MIN(seznam.size(), P::dumpannotated);
  // If states are clustered, we dump the full cluster
  for (unsigned int i = len; i < seznam.size()-1; i++) {
     if (my_fcmp(seznam[i].first, seznam[i-1].first, P::grouptol) == 0) { 
	  // If the next state has an energy within P::grouptol, add it to
	  // the list.
	  len++;
     } else {
	  break;
     }
  }
  my_assert(len <= seznam.size());
	
  Fannotated << setprecision(P::dumpprecision);
  if (P::dumpgroups) {
     // Group by degeneracies
     for (unsigned int i = 0; i < len; ) {
	Fannotated << scaled_energy(seznam[i].first, scaled, absolute);
	std::vector<string> QNstrings;
	int total_degeneracy = 0; // Total number of levels (incl multiplicity)
	const unsigned int i0 = i;
	while (i < len &&
	       my_fcmp(seznam[i].first, seznam[i0].first, P::grouptol) == 0) {
	   QNstrings.push_back(tostring(seznam[i].second));
	   total_degeneracy += mult(seznam[i].second);
	   i++;
	}
	sort(QNstrings.begin(), QNstrings.end());
	const int nr = QNstrings.size(); // Number of degenerate multiplets
	my_assert(nr >= 1);
	for (int j = 0; j < nr; j++) {
	   Fannotated << (j == 0 ? " " : ", ") << "(" << QNstrings[j] << ")";
	}
	Fannotated << " [" << total_degeneracy << "]\n";
     }
  } else {
     for (EI_VEC::iterator i = seznam.begin(); i != seznam.end(); i++) {
	Fannotated << scaled_energy(i->first, scaled, absolute) << " "
	           << i->second << "\n";
     }
  }

  // Consecutive iterations are separated by an empty line 
  Fannotated << endl;
}

/* recalc_singlet() recalculates irreducible matrix elements of a
 singlet operator (nold) and stores them in a new matrix (nnew).
 This implementation is generic for all the symmetry types! */
class recalc_singlet
{
 public:
 void operator()(const DiagInfo &diag,
		 const MatrixElements &nold,
		 MatrixElements &nnew,
		 int parity)
 {
  std::vector<Recalc> recalc_table(P::combs);
  const InvarVec In = input_subspaces();

  if (Sym->islr()) {
     my_assert(parity == 1 || parity == -1);
  } else {
     my_assert(parity == 1);
  }

  LOOP_const(diag, is1) {
    const Invar I1 = INVAR(is1);

    Invar Ip = I1;
    if (parity == -1)
      Ip.InvertParity();

    for (unsigned int i = 1; i <= P::combs; i++) {
       Recalc r;
       r.i1 = r.ip = i;
       r.factor = 1.0;

       Invar ancI = I1;
       ancI.combine(In[i]);

       r.IN1 = r.INp = ancI;
       if (parity == -1) 
	 r.INp.InvertParity();

       recalc_table[i-1] = r; // mind the -1 shift!
    }

    recalc_general(diag, nold, nnew, Ip, I1,
                   &recalc_table[0], P::combs, Sym->InvarSinglet);
  } // loop over is
 }
};

class recalc_singlet_even
{
 public:
 void operator()(const DiagInfo &diag, 
		 const MatrixElements &nold, MatrixElements &nnew)
 {
    recalc_singlet()(diag, nold, nnew, 1);
 }
};

class recalc_singlet_odd
{
 public:
 void operator()(const DiagInfo &diag,
		 const MatrixElements &nold, MatrixElements &nnew)
 {
    recalc_singlet()(diag, nold, nnew, -1);
 }
};

// Wrapper routine for recalculations. Called from nrg_recalculate_operators().
template <class RecalcFnc>
void recalc_common(RecalcFnc recalc_fnc,
		   DiagInfo &dg,
		   CustomOp::value_type &op, 
		   string tip, bool (*testfn)(const string &))
{  
   if (testfn(NAME(op))) {
      TIME("recalc " + tip);
      nrglog('0', "Recalculate " << tip << " " << NAME(op) );
      MatrixElements opstore;
      opstore.swap(op.second);
      op.second.clear();
      recalc_fnc(dg, opstore, op.second);
      if (tip == "g") {
	 Sym->recalc_global(dg, NAME(op), op.second);
      }      
   } else {
      op.second.clear(); // save memory!
   }   
}

/* We trim the matrices containing the irreducible matrix elements of the
 operators to the sizes that are actually required in the next iterations.
 This saves memory and leads to better cache usage in recalc_general()
 recalculations. Note: this is only needed for strategy=all; copying is
 avoided for strategy=kept. */
void nrg_trim_matel(DiagInfo &dg, MatrixElements &op)
{
   FOREACH(MatrixElements::value_type &m, op) {
      const Invar & I1 = m.first.first;
      const Invar & I2 = m.first.second;

      // Current matrix dimensions
      const unsigned int size1 = m.second.size1();
      const unsigned int size2 = m.second.size2();
      if (size1 == 0 || size2 == 0)
	continue;

      // Target matrix dimensions
      const unsigned int nr1 = dg[I1].getnr();
      const unsigned int nr2 = dg[I2].getnr();
      my_assert(nr1 <= size1 && nr2 <= size2);

      if (nr1 == size1 && nr2 == size2) // Trimming not necessary!!
	continue;
      
      matrix_range<Matrix> m2(m.second, range(0, nr1), range(0, nr2));
      Matrix m2new = m2;
      m.second.swap(m2new);
   }
}
      
void nrg_trim_op(DiagInfo &dg, CustomOp &allops)
{
   FOREACH(CustomOp::value_type &op, allops)
     nrg_trim_matel(dg, op.second);
}

void nrg_trim_matrices(DiagInfo &dg)
{
   nrg_trim_op(dg, a.ops);
   nrg_trim_op(dg, a.opsp);
   nrg_trim_op(dg, a.opsg);
   nrg_trim_op(dg, a.opd);
   nrg_trim_op(dg, a.opt);
   nrg_trim_op(dg, a.opot);
   nrg_trim_op(dg, a.opq); // !!
}

void nrg_trim_eigenvectors(DiagInfo &diag)
{
   LOOP(diag, is) {
      const unsigned int nrblocks = is.second.blocks.size();
      for (unsigned int i = 0; i < nrblocks; i++) {
	 Matrix zeromatrix = Matrix(0, 0);
	 is.second.blocks[i].swap(zeromatrix);
      }
   }
}

// Z_S is the appropriate statistical sum
void nrg_measure_singlet1(const DiagInfo &dg,
			  const CustomOp::value_type &op,
			  double Z_S)
{
   const string name = NAME(op);
   const t_expv expv = calc_trace_singlet(dg, op.second)/Z_S;
   cout << "<" << name << ">=";
   output_val(cout, expv);  
   cout << endl;
   STAT::expv[name] = expv;
}

// Measure thermodynamic expectation values of singlet operators
void nrg_measure_singlet(const DiagInfo &dg)
{
  nrglog('@', "@ nrg_measure_singlet()");

  const double Z_S = calc_Z(dg);
   
  FOREACH(const CustomOp::value_type &op, a.ops)
     nrg_measure_singlet1(dg, op, Z_S);
  FOREACH(const CustomOp::value_type &op, a.opsg)
     nrg_measure_singlet1(dg, op, Z_S);
  
  custom->field_values();
}

// Expectation values using FDM algorithm
void nrg_measure_singlet1_fdm(const DiagInfo &dg,
			      const CustomOp::value_type &op)
{
   const string name = NAME(op);

   if (P::fdmexpvn == -1) {
      // Old algorithm: take contributions from all NRG iterations.
      // Note that rhoFDM is not used here, but STAT::wn and STAT::ZnD
      // are.
      const double wn = STAT::wn[STAT::N];
      double wnfactor = wn/STAT::ZnD[STAT::N];
      if (isnan(wnfactor)) {
	 wnfactor = 0.0;
      }
   
      const t_expv expv = calc_trace_fdm_discarded(dg, op.second);
      cout << "contribution to <" << name << ">=";
      output_val(cout, wnfactor * expv);
      cout << endl;

      STAT::fdmexpv[name] += wnfactor * expv;
   } else {
      // New algorithm: evaluate expectation value at iteration N.
      // Note that STAT::wn and STAT::ZnD are not needed here, but
      // rhoFDM is.
      if (STAT::N == P::fdmexpvn) {
	 const t_expv expv = calc_trace_fdm_kept(dg, op.second);
	 STAT::fdmexpv[name] = expv;
	 cout << "<" << name << ">_fdm=" << expv << endl;
      }
   }
}

void nrg_measure_singlet_fdm(const DiagInfo &dg)
{
  FOREACH(const CustomOp::value_type &op, a.ops)
     nrg_measure_singlet1_fdm(dg, op);
  FOREACH(const CustomOp::value_type &op, a.opsg)
     nrg_measure_singlet1_fdm(dg, op);
}

void check_operator_sumrules(DiagInfo &diag)
{
  nrglog('@', "@ check_operator_sumrules()");
   
  // We check sum rules wrt some given spin (+1/2, by convention).
  // For non-spin-polarized calculations, this is irrelevant (0).
  const int SPIN = (Sym->isfield() ? 1 : 0);
  FOREACH(CustomOp::value_type &op, a.opd)
     doublet_check_norm(op, diag, SPIN);

  FOREACH(CustomOp::value_type &op, a.opq)
     quadruplet_check_norm(op, diag, 0);
}

// Wrappers
class recalc_doublet
{
 public:
 void operator() (DiagInfo &diag,
		  MatrixElements &cold,
		  MatrixElements &cnew) 
 {
    Sym->recalc_doublet(diag, cold, cnew);
 }
};

class recalc_triplet
{
 public:
 void operator() (DiagInfo &diag,
		  MatrixElements &cold,
		  MatrixElements &cnew) 
 {
    Sym->recalc_triplet(diag, cold, cnew);
 }
};

class recalc_orb_triplet
{
 public:
 void operator() (DiagInfo &diag,
		  MatrixElements &cold,
		  MatrixElements &cnew) 
 {
    Sym->recalc_orb_triplet(diag, cold, cnew);
 }
};

class recalc_quadruplet
{
 public:
 void operator() (DiagInfo &diag,
		  MatrixElements &cold,
		  MatrixElements &cnew) 
 {
    Sym->recalc_quadruplet(diag, cold, cnew);
 }
};

void init_fdmexpv()
{
  FOREACH(CustomOp::value_type &op, a.ops) {
     string name = op.first;
     STAT::fdmexpv[name] = 0.0;
  }
}

// Recalculate operator matrix representations
void nrg_recalculate_operators(DiagInfo &dg)
{
  nrglog('@', "@ nrg_recalculate_operators()");

  FOREACH(CustomOp::value_type &op, a.ops)
     recalc_common(recalc_singlet_even(), dg, op, "s", oprecalc::do_s);

  FOREACH(CustomOp::value_type &op, a.opsp)
     recalc_common(recalc_singlet_odd(), dg, op, "p", oprecalc::do_p);

  FOREACH(CustomOp::value_type &op, a.opsg)
     recalc_common(recalc_singlet_even(), dg, op, "g", oprecalc::do_g);

  FOREACH(CustomOp::value_type &op, a.opd) 
     recalc_common(recalc_doublet(), dg, op, "d", oprecalc::do_d);
   
  FOREACH(CustomOp::value_type &op, a.opt) 
     recalc_common(recalc_triplet(), dg, op, "t", oprecalc::do_t);

  FOREACH(CustomOp::value_type &op, a.opot)
     recalc_common(recalc_orb_triplet(), dg, op, "ot", oprecalc::do_ot);

  FOREACH(CustomOp::value_type &op, a.opq)
     recalc_common(recalc_quadruplet(), dg, op, "q", oprecalc::do_q);
}

// Calculate spectral densities
void nrg_spectral_densities(DiagInfo &diag)
{    
   nrglog('@', "@ nrg_spectral_densities()");

   TIME_SECTION("spec s",
     FOREACH(speclist::value_type &i, spectraS) 
       calc_correlator_wrap(i, diag);
   );

   TIME_SECTION("spec chit",
     FOREACH(speclist::value_type &i, spectraCHIT) 
       calc_correlator_wrap(i, diag);
   );

   TIME_SECTION("spec d",
     FOREACH(speclist::value_type &i, spectraD) 
       calc_specdens_wrap(i, diag);
   );

  TIME_SECTION("spec t",
    FOREACH(speclist::value_type &i, spectraT)
       calc_spinsusc_wrap(i, diag);
  );

  TIME_SECTION("spec ot",
    FOREACH(speclist::value_type &i, spectraOT)
       calc_orbsusc_wrap(i, diag);
  );

  TIME_SECTION("spec q",
     FOREACH(speclist::value_type &i, spectraQ) 
       calc_specdensquad_wrap(i, diag);
   );

  TIME_SECTION("spec gt",
    FOREACH(speclist::value_type &i, spectraGT)
       calc_specdens_wrap(i, diag);
  );	       

  TIME_SECTION("spec i1t",
    FOREACH(speclist::value_type &i, spectraI1T)
       calc_specdens_wrap(i, diag);
  );
   
  TIME_SECTION("spec i2t",
    FOREACH(speclist::value_type &i, spectraI2T)
       calc_specdens_wrap(i, diag);
  );	       
}    

/* We calculate thermodynamic quantities before truncation to make better
 use of the available states. Here we compute quantities which are defined
 for all symmetry types. Other calculations are performed by calculate_TD
 member functions defined in symmetry.cc. */
void nrg_calculate_TD(DiagInfo &diag, double additional_factor = 1.0)
{
  nrglog('@', "@ nrg_calculate_TD()");
  
  TD::T = STAT::Teff;

   // Rescale factor for energies. The energies are expressed in
   // units of omega_N, thus we need to appropriately rescale them to
   // calculate the Boltzmann weights at the temperature scale Teff
   // (Teff=scale/betabar).
  double rescale_factor = P::betabar * additional_factor;
   
  bucket Z, E, E2; // Statistical sum, Tr[beta H], Tr[(beta H)^2]
    
  LOOP(diag, is) {
    bucket sumZ, sumE, sumE2;
    const unsigned int count = NRSTATES(is);
    for (unsigned int r = 0; r < count; r++) {
      const double betaE = rescale_factor * EIGEN(is).value(r);
      const double expo = exp(-betaE);
      sumE  += betaE * expo;
      sumE2 += sqr(betaE) * expo;
      sumZ  += expo;
    }

    // Take into account the multiplicity
    const int multip = mult(INVAR(is));
    Z    += multip * sumZ;
    E    += multip * sumE;
    E2   += multip * sumE2;
  }
  STAT::Z = Z;

  // Energy, energy squared, heat capacity, free energy, entropy
  TD::E  = STAT::E  = E/Z; // beta <H>
  TD::E2 = STAT::E2 = E2/Z; // beta^2 <H^2>
  TD::C  = STAT::C  = STAT::E2 - sqr(STAT::E); // C/k_B=beta^2(<H^2>-<H>^2)
  TD::F  = STAT::F  = -log(Z); // F/(k_B T)=-ln(Z)
  TD::S  = STAT::S  = STAT::E - STAT::F; // S/k_B=beta<H>+ln(Z)

  // Call quantum number specific calculation routine
  Sym->calculate_TD(diag, rescale_factor);

  if (nrgrun)
     TD::save_TD_quantities(Ftd);
}

void nrg_calculate_spectral(DiagInfo &diag)
{
  nrglog('@', "@ nrg_calculate_spectral()");

  // Zft is used in the spectral function calculations using the
  // conventional approach. We calculate it here, in order to avoid
  // recalculations later on.
  STAT::Zft = calc_grand_canonical_Z(diag);
   
  if (string(P::specgt) != "" || string(P::speci1t) != "" ||
      string(P::speci2t) != "") {
     double GTtemperature = P::gtp * STAT::scale;
     STAT::Zgt = calc_grand_canonical_Z(diag, GTtemperature);
  }
  if (string(P::specchit) != "") {
    double CHITtemperature = P::chitp * STAT::scale;
    STAT::Zchit = calc_grand_canonical_Z(diag, CHITtemperature);
  }
   
  if (dmnrgrun && !LAST_ITERATION()) {
     loadRho(STAT::N, FN_RHO, rho);
     if (P::checkrho)
       check_trace_rho(rho); // Check if Tr[rho]=1, i.e. the normalization
                             // condition.
     if (P::fdm) 
       loadRho(STAT::N, FN_RHOFDM, rhoFDM);
  }   

  nrg_spectral_densities(diag);

  if (nrgrun) 
    nrg_measure_singlet(diag);
   
  if (dmnrgrun && P::fdmexpv) 
    nrg_measure_singlet_fdm(diag);
}

// Perform calculations of physical quantities. Called prior to NRG
// iteration (if calc0=true) and after each NRG step.
void nrg_perform_measurements(DiagInfo &diag)
{
  nrglog('@', "@ nrg_perform_measurements()");

  nrg_calculate_TD(diag);
  if (nrgrun && P::dumpannotated)
     dump_annotated(diag, P::dumpscaled, P::dumpabs);
}

/* Make a list of subspaces for the new iteration. Generic implementation:
 use the quantum number differences in array In[] (obtained by a call to
 function input_subspaces), make a list of all possible subspaces and
 remove duplicates. */
void nrg_make_subspaces_list(list<Invar> &subspaces)
{
  BOOST_FOREACH(DiagInfo::value_type &i, diagprev)
    if (NRSTATES(i)) {
      const Invar I = INVAR(i);
      InvarVec In = input_subspaces(); // make a new copy of subspaces list
      for (unsigned int i = 1; i <= P::combs; i++) {
	In[i].inverse(); // IMPORTANT!
        In[i].combine(I);
        if (Sym->Invar_allowed(In[i]))
          subspaces.push_back(In[i]);
      }
    }

  subspaces.sort();
  subspaces.unique();
}

/* Define recalculation strategy 
 all: Recalculate using all vectors
 kept: Recalculate using vectors kept after truncation
 VERY IMPORTANT: Override in the case of CFS (in the second run) */
bool do_recalc_kept()
{
   return string(P::strategy) == "kept" && !(cfs_flags() && dmnrgrun);
}

bool do_recalc_all()
{
   return !do_recalc_kept();
}

t_eigen EigenvaluePrev(const Invar &I, int r)
{
   return diagprev[I].value(r);
}

// Diagonalise the Hamiltonian.
void diag_h(Matrix &h, Eigen &diagI)
{
  check_is_matrix_upper(h);
  diagonalise_h(h, diagI);
  my_assert(diagI.matrix0.size1() <= h.size1());
  my_assert(diagI.matrix0.size2() == h.size2());
}

void nrg_prepare_task_for_diag(const Invar &I, Matrix &h)
{
   nrglog('@', "@ nrg_prepare_task_for_diag()");

   InvarVec & In = a.ancestors[I];
   Rmaxvals & qq = qsrmax[I];
      
   const unsigned int dim = qq.total();
   
   nrglog('i', endl << "Subspace (" << I << ") dim=" << dim); // skip a line
   logancestors(I, In, qsrmax[I]);
   
   h = Matrix(dim, dim);
   h.clear();
   
   double scalefactor = (!P::substeps ? sqrt(P::Lambda) : pow(P::Lambda, 1./(2.*P::channels)));
      
   // H_{N+1}=\lambda^{1/2} H_N+\xi_N (hopping terms)
   for (unsigned int i = 1; i <= P::combs; i++) {
      const unsigned int offset = qq.offset(i);
      const unsigned int rmx = qq.rmax(i);
      for (unsigned int r = 0; r < rmx; r++)
	h(offset + r, offset + r) = scalefactor * EigenvaluePrev(In[i], r);
   }

   // Symmetry-type-specific matrix initialization steps.
   Sym->makematrix(h, qq, I, In);
   
   if (logletter('m'))
     dump_matrix(h);
}

void nrg_diagonalisations_serial()
{
   nrglog('@', "@ nrg_diagonalisations_serial()");

   for (std::vector<Invar>::iterator i = NRG::tasks.begin(); i != NRG::tasks.end(); i++) {
      Matrix h;
      nrg_prepare_task_for_diag(*i, h);
      diag_h(h, diag[*i]);
   }
}

#ifdef _OPENMP
void nrg_diagonalisations_OpenMP()
{
   nrglog('@', "@ nrg_diagonalisations_OpenMP()");
   nrglog('(', "OpenMP diag");

   unsigned int nr = NRG::tasks.size();
   unsigned int itask;

   #pragma omp parallel for schedule(dynamic)
   for (itask = 0; itask < nr; itask++) {
      Invar I = NRG::tasks[itask];
      Matrix h;
      nrg_prepare_task_for_diag(I, h);
      int thid = omp_get_thread_num();
      #pragma omp critical 
      {
	 nrglog('(', "Diagonalizing " << I << " size=" << h.size1() 
		<< " (task " << itask+1 << "/" << nr << ", thread " << thid << ")");
      }      
      diag_h(h, diag[I]);
   }
}
#endif

#ifdef NRG_MPI
void mpi_sync_params();

void check_status(mpi::status &status)
{
   if (status.error()) {
      cerr << "MPI communication error. rank=" << myrank << endl;
      mpienv->abort(1);
   }
}

// NOTE: MPI is limited to message size of 2GB (or 4GB). For big
// problems we thus need to send objects line by line.

//#define MPI_WHOLEMATRIX
#define MPI_LINEBYLINE

#ifdef MPI_WHOLEMATRIX
 #define mpi_send_matrix mpi_send_matrix_wholematrix
 #define mpi_receive_matrix mpi_receive_matrix_wholematrix
 #define mpi_send_eigen mpi_send_eigen_whole
 #define mpi_receive_eigen mpi_receive_eigen_whole
#endif

#ifdef MPI_LINEBYLINE
 #define mpi_send_matrix mpi_send_matrix_linebyline
 #define mpi_receive_matrix mpi_receive_matrix_linebyline
 #define mpi_send_eigen mpi_send_eigen_linebyline
 #define mpi_receive_eigen mpi_receive_eigen_linebyline
#endif

void mpi_send_matrix_wholematrix(int dest, Matrix &m)
{
   mpiw->send(dest, TAG_MATRIX, m);
}

void mpi_receive_matrix_wholematrix(int source, Matrix &m)
{
   mpi::status status;
   status = mpiw->recv(source, TAG_MATRIX, m);
   check_status(status);
}

void mpi_send_matrix_linebyline(int dest, const Matrix &m)
{
   unsigned int size1 = m.size1();
   mpiw->send(dest, TAG_MATRIX_SIZE, size1);
   unsigned int size2 = m.size2();
   mpiw->send(dest, TAG_MATRIX_SIZE, size2);
   
   if (logletter('M'))
     cout << "Sending matrix of size " << size1 << " x " << size2 
      << " line by line to " << dest << endl;
   
   for (unsigned int i = 0; i < size1; i++) {
      ublas::vector<t_matel> vec = ublas::vector<t_matel>(size2);
      matrix_row<const Matrix> mr = matrix_row<const Matrix>(m, i);
      vec = mr;
      mpiw->send(dest, TAG_MATRIX_LINE, vec);
   }
}

void mpi_receive_matrix_linebyline(int source, Matrix &m)
{
   mpi::status status;
   
   unsigned int size1;
   status = mpiw->recv(source, TAG_MATRIX_SIZE, size1);
   check_status(status);

   unsigned int size2;
   status = mpiw->recv(source, TAG_MATRIX_SIZE, size2);
   check_status(status);

   if (logletter('M')) 
     cout << "Receiving matrix of size " << size1 << " x " << size2 
      << " line by line from " << source << endl;
   
   m = Matrix(size1, size2);
   
   for (unsigned int i = 0; i < size1; i++) {
      ublas::vector<t_matel> vec;
      status = mpiw->recv(source, TAG_MATRIX_LINE, vec);
      check_status(status);
      my_assert(vec.size() == size2);
      matrix_row<Matrix>(m, i) = vec;
   }
}

void mpi_send_eigen_whole(int dest, const Eigen &eig)
{
   mpiw->send(dest, TAG_EIGEN, eig);
}

void mpi_receive_eigen_whole(int source, Eigen &eig)
{
   mpi::status status;
   status = mpiw->recv(source, TAG_EIGEN, eig);
   check_status(status);
}

void mpi_send_eigen_linebyline(int dest, const Eigen &eig)
{
   Eigen eigmock; // empty Eigen
   mpiw->send(dest, TAG_EIGEN, eigmock);
   
   if (logletter('M')) 
     cout << "Sending eigen from " << myrank << " to " << dest << endl;
   
   mpiw->send(dest, TAG_EIGEN_INT, eig.nr);
   mpiw->send(dest, TAG_EIGEN_INT, eig.rmax);
   mpiw->send(dest, TAG_EIGEN_INT, eig.nrpost);

   mpiw->send(dest, TAG_EIGEN_VEC, eig.value);
   mpiw->send(dest, TAG_EIGEN_VEC, eig.value_raw);
   mpiw->send(dest, TAG_EIGEN_VEC, eig.absenergy);

   mpi_send_matrix_linebyline(dest, eig.matrix0);
}

void mpi_receive_eigen_linebyline(int source, Eigen &eig)
{
   if (logletter('M')) 
     cout << "Receiving eigen from " << source << " on " << myrank << endl;

   mpi::status status;
   
   Eigen eigmock;
   status = mpiw->recv(source, TAG_EIGEN, eigmock);
   check_status(status);

   status = mpiw->recv(source, TAG_EIGEN_INT, eig.nr);
   check_status(status);
   status = mpiw->recv(source, TAG_EIGEN_INT, eig.rmax);
   check_status(status);
   status = mpiw->recv(source, TAG_EIGEN_INT, eig.nrpost);
   check_status(status);

   status = mpiw->recv(source, TAG_EIGEN_VEC, eig.value);
   check_status(status);
   status = mpiw->recv(source, TAG_EIGEN_VEC, eig.value_raw);
   check_status(status);
   status = mpiw->recv(source, TAG_EIGEN_VEC, eig.absenergy);
   check_status(status);

   mpi_receive_matrix_linebyline(source, eig.matrix0);
}

// Read results from a slave process.
Invar read_from(int source)
{
   if (logletter('M'))
     cout << "Reading results from " << source << endl;
   
   mpi::status readst;
   Eigen eig;
   mpi_receive_eigen(source, eig);
   
   Invar Irecv;
   readst = mpiw->recv(source, TAG_INVAR, Irecv);
   check_status(readst);
   
   if (logletter('M')) {
     cout << "Received results for subspace " << Irecv;
     cout << " [nr=" << eig.getnr() << ", dim=" << eig.getrmax() << "]" << endl;
   }
   
   // Some consistency checks
   my_assert(eig.matrix0.size1() == eig.getnr());
   my_assert(eig.matrix0.size2() == eig.getrmax());
   my_assert(eig.matrix0.size1() <= eig.matrix0.size2());
   my_assert(eig.getrmax() == qsrmax[Irecv].total());

   diag[Irecv] = eig;
   return Irecv;
}

void nrg_diagonalisations_MPI()
{
   nrglog('@', "@ nrg_diagonalisations_MPI()");

   // Synchronise parameters
   mpi_sync_params();
   
   // List of all the remaining tasks
   int nrtasks = NRG::tasks.size();
   list<Invar> todo;
   for (std::vector<Invar>::iterator i = NRG::tasks.begin(); i != NRG::tasks.end(); i++) {
      todo.push_back(*i);
   }
   //copy(NRG::tasks.begin(), NRG::tasks.end(), todo.begin());
   my_assert(todo.size() == nrtasks);
   
   // List of finished tasks.
   list<Invar> done;
   
   // List of the available computation nodes (including the master,
   // which is always at the very beginnig of the deque).
   deque<int> nodes;
   for (int i = 0; i < mpiw->size(); i++)
     nodes.push_back(i);
   
   if (logletter('M')) {
      cout << "nrtasks=" << nrtasks << endl;
      cout << "nrnodes=" << mpiw->size() << endl;
   }   
   
   while (!todo.empty()) {
      my_assert(!nodes.empty());
      
      // i is the node to which the next job will be scheduled
      int i;
      if (todo.size() == 1) {
	 // If a single task is left undone, do it on the master node
	 // to avoid the unnecessary network copying.
	 i = 0;
      } else {
	 i = nodes.back();
	 nodes.pop_back();
      }
   
      Invar I;
      if (i == 0) {
	 // On master, we take short jobs from the end of the list.
	 I = todo.back();
	 todo.pop_back();
      } else {
	 // On slaves, we take long jobs from the beginning of the
	 // list.
	 I = todo.front();
	 todo.pop_front();
      }
      
      Matrix h;
      nrg_prepare_task_for_diag(I, h);
      if (logletter('M'))
	cout << "Scheduler: job " << I << " (dim=" << h.size1() << ")"
       	     << " on node " << i << endl;
      
      if (i == 0) {
	 // On master, diagonalize immediately.
	 diag_h(h, diag[I]);
	 nodes.push_back(0);
	 done.push_back(I);
      } else {
	 mpiw->send(i, TAG_DIAG, 0);
	 mpi_send_matrix(i, h);
	 mpiw->send(i, TAG_INVAR, I);
      }
      
      // Check for terminated jobs
      boost::optional<mpi::status> status;
      while (status = mpiw->iprobe(mpi::any_source, TAG_EIGEN)) {
	 if (logletter('M')) {
	    cout << "Receiveing results from " << status->source() << endl;
	 }	 
	 Invar Irecv = read_from(status->source());
	 done.push_back(Irecv);
	 // The node is now available for new tasks!
	 nodes.push_back(status->source());
      }
   }
   
   // Keep reading results sent from the slave processes until all
   // tasks have been completed.
   int nrdone = done.size();
   while (nrdone != nrtasks) {
      if (logletter('M')) 
	cout << "So far: " << nrdone << "/" << nrtasks << endl;
      mpi::status status = mpiw->probe(mpi::any_source, TAG_EIGEN);
      Invar Irecv = read_from(status.source());
      done.push_back(Irecv);
      nrdone++;
   }
}
#endif

// Build matrix H(ri;r'i') in each subspace and diagonalize it
void nrg_diagonalisations()
{
   nrglog('@', "@ nrg_diagonalisations()");

   // This needs to be called here, because class Timing is not
   // thread-safe.
   TIME("diag");
   
   // Call init() again here, because NRG::diagratio might have
   // changed!
   sP.init();
   
   diag.clear();

#ifdef NRG_MPI
   nrg_diagonalisations_MPI();
#else
 #ifdef _OPENMP
   nrg_diagonalisations_OpenMP();
 #else
   nrg_diagonalisations_serial();
 #endif
#endif
}

// Determine the list of invariant subspaces in which diagonalisations need
// to be performed.
void nrg_determine_tasks()
{
   nrglog('@', "@ nrg_determine_tasks()");

   // Make a list of all subspaces to consider.
   list<Invar> subspaces;
   nrg_make_subspaces_list(subspaces);
   
   // Auxiliary information: ancestor subspaces and their dimensions.
   qsrmax.clear();
   a.ancestors.clear();

   /* Container holding all the subspaces that
      appear in the new iteration. */
   NRG::tasks.clear();

   FOREACH(Invar &I, subspaces) {
      // Determine which subspaces contribute to the Hamiltonian being built
      InvarVec In = input_subspaces();
      for (unsigned int i = 1; i <= P::combs; i++)
	In[i].combine(I); // In is the list of differences wrt I
      a.ancestors[I] = In;
      
      // Determine the range(s) of index r
      qsrmax[I].determine_ranges(I, In);
      // nr is actually the size of the Hamiltonian submatrix!
      const unsigned int nr = qsrmax[I].total();
            
      // Note that NRG::tasks is not the same list as 'subspaces',
      // since some possible subspaces may have dimension 0 and thus
      // do not really exist.
      if (nr) 
	NRG::tasks.push_back(I);
   }
}

struct SelectInvar 
{
   Invar operator()(pair<unsigned int, Invar> &p) {
      return p.second;
   }
};

// Sort NRG::tasks in the decreasing order of the submatrix size.
// As a side effect, we compute some statistics about matrix sizes.
void sort_task_list()
{
   typedef pair<unsigned int, Invar> pairSI; // (size, Invar) pair
   typedef std::vector<pairSI> pairSIvec; // vector of such pairs
   pairSIvec tasks_with_sizes;
   
   unsigned int max_size = 0;
   unsigned int min_size = UINT_MAX;
   unsigned int nr = 0;

   for (std::vector<Invar>::iterator i = NRG::tasks.begin(); i != NRG::tasks.end(); i++) {
      nr++;
      const unsigned int size = qsrmax[*i].total();
      tasks_with_sizes.push_back(make_pair(size, *i));
      max_size = MAX(max_size, size);
      min_size = MIN(min_size, size);
   }
   cout << "Stats: nr=" << nr << " min=" << min_size << " max=" << max_size << endl;
   
   // Sort in the *decreasing* order!
   sort(tasks_with_sizes.rbegin(), tasks_with_sizes.rend());
   
   // Report matrix sizes
   if (logletter('S')) 
      for (pairSIvec::iterator i = tasks_with_sizes.begin(); i != tasks_with_sizes.end(); i++)
	 cout << "size(" << i->second << ")=" << i->first << endl;
   
   // Update the task list NRG::tasks with the sorted list of subspaces
   transform(tasks_with_sizes.begin(),
	     tasks_with_sizes.end(),
	     NRG::tasks.begin(),
	     SelectInvar());
}

// Recalculate irreducible matrix elements for Wilson chains.
// Called from nrg_after_diag().
void nrg_recalc_f(DiagInfo &diag) 
{
   nrglog('@', "@ nrg_recalc_f()");
   TIME("recalc f");
   
   if (!P::substeps) {
      for (unsigned int i = 0; i < P::channels; i++)
	for (unsigned int j = 0; j < P::perchannel; j++)
	  a.opch[i][j].clear(); // Clear all channels
      // In principle, one could also use recalc_doublet() function to
      // simplify the code.
      // OTOH, recalc_irreduc() is probably better because it does not
      // accumulated floating point round-off errors.
      Sym->recalc_irreduc(diag);
   } else {
     int Ntrue, M;
     tie(Ntrue, M) = get_Ntrue_M(STAT::N);

     for (unsigned int i = 0; i < P::channels; i++) {
	if (i == M) {
	   for (unsigned int j = 0; j < P::perchannel; j++)
	     a.opch[M][j].clear(); // Clear channel M
	   
	   Sym->recalc_irreduc_substeps(diag, M);
	} else {
	   for (unsigned int j = 0; j < P::perchannel; j++) {
	      MatrixElements & f = a.opch[i][j];
	      MatrixElements opstore;
	      opstore.swap(f);
	      f.clear();
	      Sym->recalc_doublet(diag, opstore, f);
	   }
	}
     }
   }
}

void nrg_dump_f()
{
   cout << endl;
   for (unsigned int i = 0; i < P::channels; i++) {
      for (unsigned int j = 0; j < P::perchannel; j++) {
	 cout << "<f> dump, i=" << i << " j=" << j << endl;
	 dump_matrix_elements(a.opch[i][j]);
      }
   }
   cout << endl;
}  

// The absenergy[] values are shifted so that the ground state corresponds
// to zero. This is required in the FDM approach for calculating the 
// spectral functions. This is different from subtract_groundstate_energy().
// Called from nrg_do_diag() when diag is loaded from a stored file during
// the second pass of the NRG iteration. Note that the values in absenergy[]
// are not yet shifted when ZnD is calculated in calc_ZnD_one, but it is
// already shifted when calc_generic_FDM(), calc_trace_fdm_discarded/kept() 
// are called.
void shift_abs_energies(DiagInfo &diag) 
{
  LOOP(diag, is) {
    const unsigned int dim = NRSTATES(is);
    for (unsigned int r = 0; r < dim; r++) {
      EIGEN(is).absenergy[r] -= STAT::GSenergy;
      my_assert(EIGEN(is).absenergy[r] >= 0);
    }
  }
}

/* NRG diagonalisation driver: calls nrg_diagionalisations() or
 load_transformations(), as necessary, and performs the truncation. All other
 calculations are done in nrg_after_diag(). Called from nrg_iterate(). */
void nrg_do_diag()
{
   nrglog('@', "@ nrg_do_diag()");

   infostring();
   show_coefficients();
   nrg_determine_tasks();

   sort_task_list();
   
   NRG::diagratio = P::diagratio;
   
   do {
      NRG::notenough = false;
      
      if (nrgrun) {
	 if (!(P::resume && STAT::N <= P::laststored)) {
	    nrg_diagonalisations(); // compute in first run
	    copy_unshifted_eigenvalues(diag);
	 } else {
	    cout << "Loading stored transformation matrices" << endl;
	    load_transformations(STAT::N, diag); // or read from disk
	 }
      }
      
      if (dmnrgrun) {
	 load_transformations(STAT::N, diag); // read from disk in second run
	 
	 // IMPORTANT: subtract the absolute (!) GS energy in the
	 // abs_energy vector. The overall (all shells, all invariant
	 // subspaces) lowest abs_energy will thus be equal to zero.
	 shift_abs_energies(diag);
	 if (P::removefiles)
	   remove_transformation_files(STAT::N);
      }

      ms.check("after diag");
      
      find_groundstate(diag);
      subtract_groundstate_energy(diag);
      copy_sort_energies(diag, STAT::energies);
      find_clusters(STAT::energies, P::fixeps, STAT::cluster_mapping);
      bool recopy_required = fix_splittings(diag);
      // fix_splittings() returns true if any changes had been made. If so,
      // we determine the clusters again (just to see what the result is).
      if (recopy_required) {
        copy_sort_energies(diag, STAT::energies);
        find_clusters(STAT::energies, P::fixeps, STAT::cluster_mapping);
      }
      nrg_truncate_prepare(diag);
      
      ms.check("after trunc");

      if (NRG::notenough) {
	 cerr << "Insufficient number of states computed." << endl;

	 if (P::restart) {
	    NRG::diagratio = min(NRG::diagratio * P::restartfactor, 1.0);

	    cout << endl << "Restarting this iteration step. " 
	      << "diagratio=" << NRG::diagratio << endl << endl;
	 }      
      }
   } while(nrgrun && P::restart && NRG::notenough);
}

typedef map<string, double> mapSD;
std::vector<mapSD> td_data;

// Store all the results from the TD calculation in the form of a map
// (string -> double).
void store_td()
{
   vecoutptr::iterator i;
   mapSD td;
   for (i = allfields.begin(); i != allfields.end(); i++) {
      string name = (*i)->name();
      double value = (*i)->rawvalue();
      td[name] = value;
   }
   td_data.push_back(td);
}

// These are absolute energies, but *not* referrenced to absoltue 0.
// Must be called after STAT::totalenergy has been updated, but 
// before store_transformations().
void calc_abs_energies(const int N, DiagInfo &diag)
{
  LOOP(diag, is) {
    const int nr = NRSTATES(is); // all states!
    EIGEN(is).absenergy = EVEC(nr);
    for (int i = 0; i < nr; i++) 
      EIGEN(is).absenergy[i] = STAT::totalenergy + EIGEN(is).value[i] * STAT::scale;
  }
}

// Perform processing after a successful NRG step
void nrg_after_diag()
{
   nrglog('@', "@ nrg_after_diag()");

   // Contribution to the total energy.
   STAT::totalenergy += STAT::Egs * STAT::scale;
   if (nrgrun) 
      cout << "Total energy=" << STAT::totalenergy << "  Egs=" << STAT::Egs << endl;

   if (nrgrun)
     calc_abs_energies(STAT::N, diag);

   if (P::dm && nrgrun) {
      // Store eigenenergies and eigenvectors in all subspaces.
      // NOTE: we store before the truncation!
      if (!(P::resume && STAT::N <= P::laststored)) 
	 store_transformations(STAT::N, diag);
   }
  
   // Logging of ALL states (not only those that remain after truncation)
   if (P::dumpenergies) 
     dumptofile(diag, F);

   // Measurements are performed before the truncation!
   nrg_perform_measurements(diag);

   // Consistency checks
   LOOP_const(diag, i) {
     my_assert(EIGEN(i).matrix0.size1() <= EIGEN(i).matrix0.size2());
     my_assert(EIGEN(i).matrix0.size2() == qsrmax[INVAR(i)].total());
   }
   
   split_in_blocks(diag);
   ms.check("after split");

   if (do_recalc_all()) { // Either ...
      nrg_recalculate_operators(diag);
      nrg_calculate_spectral(diag);
   }

   // Actual truncation occurs at this point
   nrg_truncate_perform(diag);
   
   // Store information about subspaces and states
   unsigned int nrall = 0;
   unsigned int nrkept = 0;
   LOOP_const(diag, i) {
     const Invar I = INVAR(i);
     dm[STAT::N][I] = DimSub(NRSTATES(i), RMAX(i));
     nrall += RMAX(i);
     nrkept += NRSTATES(i);
     dm[STAT::N][I].rmax = qsrmax[I];
     dm[STAT::N][I].eigenvalue = EIGEN(i).value;
     dm[STAT::N][I].absenergy = EIGEN(i).absenergy;
   }

   double ratio = double(nrkept)/nrall;
   cout << "Kept: " << nrkept << " out of " << nrall 
        << ", ratio=" << setprecision(3) << ratio << endl;
   cout << setprecision(COUT_PRECISION);
   
   if (!LAST_ITERATION()) {
      nrg_recalc_f(diag);
      if (P::dump_f) 
	 nrg_dump_f();
   }
   
   if (do_recalc_kept()) { // ... or
      nrg_recalculate_operators(diag);
      nrg_calculate_spectral(diag);
   }
   
   if (P::checksumrules)
      check_operator_sumrules(diag);
   
   ms.check("after recalc");

   // Free up memory that contains information we no longer need
   nrg_trim_matrices(diag);
   nrg_trim_eigenvectors(diag);
   ms.check("after trim");
   diagprev.swap(diag); // IMPORTANT: we need to retain the eigenenergies!
   
   // Store TD data (all outfields)
   store_td();
}

// Perform one iteration step
void nrg_iterate()
{
   nrg_do_diag();
   nrg_after_diag();
     
#ifdef HAS_MEMORY_USAGE
   cout << "Memory used: " << long(ms.used()/1024) << " MB" << "  ";
#endif
   cout << "Time elapsed: " << long(t.total()) << " s" << endl;
}

void docalc0ht(int extra_steps) 
{
   my_assert(extra_steps >= 0);
   for (int i = -extra_steps; i <= -1; i++) {
      STAT::set_N(P::Ninit - 1 + i);
      double E_rescale_factor = pow(P::Lambda, i/2.0);
      nrg_calculate_TD(diagprev, E_rescale_factor);
   }
}

// Perform calculations with quantities from 'data' file
void docalc0()
{
   nrglog('@' , "@ docalc0()");

   cout << endl << "Before NRG iteration";
   cout << " (N=" << STAT::N << ")" << endl;
   nrg_perform_measurements(diagprev);
   nrg_calculate_spectral(diagprev);
   
   // Logging of ALL states (prior to truncation!)
   if (P::dumpenergies) 
      dumptofile(diagprev, F);
   
   if (P::checksumrules)
     check_operator_sumrules(diagprev);
}

// ****************************  Main NRG loop ****************************

void nrg_main_loop()
{
  nrglog('@', "@ nrg_main_loop()");

  // N denotes the order of the Hamiltonian. N=0 corresponds to H_0, i.e.
  // the initial Hamiltonian (cf. Krishna-Murty I) 
  for (unsigned int N = P::Ninit; N < P::Nmax; N++) {
    if (nrgrun && P::forcestop == int(N)) {
       cout << endl << "*** Stop request at iteration " << P::forcestop << endl;
       exit(1);
    }     
    STAT::set_N(N);
    nrg_iterate();
  }
}

// Estimate the Kondo temperature (Wilson's definition). This
// calculation only makes sense for the SIAM and the Kondo model and
// even then only if Tmin << TKW, B=0, and Sz^2 is computed.
double calc_TKW_do()
{
   const int len = td_data.size();
   assert(len > 2);
   const double T0chiT0 = td_data[len-1]["<Sz^2>"];
   const double y = T0chiT0 + 0.07;
   
   // Linear interpolation
   double T = 0.0;
   for (int i = len-1; i >= 1; i--) {
      const double x0 = td_data[i]["T"]; // lower T
      const double x1 = td_data[i-1]["T"]; // higher T
      const double y0 = td_data[i]["<Sz^2>"]; // lower Tchi(T)
      const double y1 = td_data[i-1]["<Sz^2>"]; // higher Tchi(T)
      if (y0 <= y && y <= y1) {
	 T = x0 + (x1-x0)/(y1-y0)*(y-y0);
	 break;
      }
   }
   
   return T;
}

void calc_TKW()
{
   if (outfield_exists("<Sz^2>")) {
      double TKW = calc_TKW_do();
      cout << "TKW=" << TKW << endl;
   }
}

// Total number of states (symmetry taken into account)
int count_states(const DiagInfo &dg)
{
  int states = 0;
  LOOP_const(dg, i)
    states += Sym->mult(INVAR(i)) * NRSTATES(i);
  return states;
}

// Count non-empty subspaces
int count_subspaces(const DiagInfo &dg)
{
   int subspaces = 0;
   LOOP_const(dg, i)
     if (NRSTATES(i))
       subspaces++;
   return subspaces;
}

// Dump information about states.
void states_report(const DiagInfo &dg, ostream &fout = cout)
{
  fout << "Number of invariant subspaces: " << count_subspaces(dg) << endl;
  LOOP_const(dg, is)
    if (NRSTATES(is)) {
       fout << "(" << INVAR(is) << ") " << NRSTATES(is) << " states" << endl;
       fout << is.second.value << endl;
    }     
  fout << "Number of states (multiplicity taken into account): " 
       << count_states(dg) << endl << endl;
}

void start_run()
{
  nrglog('@', "@ start_run()");

  states_report(diagprev);
  open_output_files();

   // If setting calc0 is set to true, a calculation of TD quantities is
  // performed before starting the NRG iteration. 
  if (nrgrun && P::calc0) {
     docalc0ht(P::tdht);
     STAT::set_N(P::Ninit - 1); // in usual case, this will result in N=-1
     docalc0();
  }
 
  nrg_main_loop();
   
  if (nrgrun) // leave files open in dmnrg run
     close_output_files();
   
  cout << endl << "** Iteration completed." << endl << endl;
}

// === AFTER THE NRG ITERATION HAD COMPLETED =================================

// Processing performed both after NRG and after DM runs.
void finalize_common(void)
{
  nrglog('@', "@ finalize_common()");

  /*** Calculate continuous spectral density functions ***/
  TIME("broaden");
  FOREACH(lsl::value_type &i, allspectra) 
     i->spec_continuous_clear();
}

// Save a dump of all subspaces, with dimension info, etc.
void dump_subspace_information()
{
  ofstream O(FN_SUBSPACES);
  for (int N = P::Ninit; N < P::Nmax; N++) {
    O << "Iteration " << N << endl;
    O << "len_dm=" << dm[N].size() << endl;
    for (Subs::const_iterator i = dm[N].begin(); i != dm[N].end(); ++i) {
      const Invar I = i->first;
      O << "I=" << I << " len=" << i->second.eigenvalue.size()
        << " kept=" << i->second.kept << " total=" << i->second.total << endl;
      my_assert(i->second.total == i->second.eigenvalue.size());
    }
    O << endl;
  }
}

// Called after the first NRG run.
void finalize_nrg(void)
{
  nrglog('@', "@ finalize_nrg()");

  finalize_common();

  cout << endl << "Total energy: " << HIGHPREC(STAT::totalenergy) << endl;

  // True ground state energy is just the value of totalenergy at the end
  // of the iteration. This is the energy of the lowest energy state in the
  // last iteration. All other states (incl. from previous shells)
  // obviously have higher energies.
  STAT::GSenergy = STAT::totalenergy;
   
  calc_TKW();

  if (P::fdm) // fix 13.12.2014
    calc_ZnD();

  if (P::dumpsubspaces)
     dump_subspace_information();

  // In the last iteration of the first run we initialize the (reduced)
  // density matrix. NOTE: diagprev <-> diag, see the last line in
  // nrg_after_diag() !!
  if (P::dm) {
     init_rho(diagprev, rho);
     if (P::fdm) {
       // Initialize the shell-N density matrix. Note that calc_ZnD()
       // has already been called above. For other shells,
       // init_rho_FDM() is called from dmnrg.h.
       init_rho_FDM(rhoFDM, STAT::N);
     }
  }
}

// Called after the second NRG run.
void finalize_dmnrg(void)
{
  nrglog('@', "@ finalize_dmnrg()");

  finalize_common();
   
  if (P::fdmexpv) {
     cout << endl << "FDM expectation values" << endl;
     FOREACH(STAT::t_mapexpv::value_type &v, STAT::fdmexpv) {
	cout << "<" << v.first << ">=";
	output_val(cout, v.second);
	cout << endl;
     }
     customfdm->field_values(P::T); // save to file
  }

  // These two should match if value_raw and value vectors are
  // handled correctly. GSenergy was computed in the first NRG run,
  // while totalenergy is recomputed in the second DMNRG run.
  my_assert(num_equal(STAT::GSenergy, STAT::totalenergy));
}
  
/************************ MAIN ****************************************/

void outOfMemory()
{
  cerr << "Memory exhausted." << endl;
  exit(1);
}

// Info on program, version, symmetry type, compiler, etc.
// This data is important for regression testing!
void print_about_message(ostream &F) 
{
  F << PACKAGE_STRING << " - (c) " << PACKAGE_BUGREPORT << endl;
  F << "Timestamp: " << __TIMESTAMP__ << endl;
  F << "Compiled on " << __DATE__ << " at " << __TIME__ << endl;

  F << "Compiled-in symmetry types: ";
  FOREACH(sym_map::value_type &x, all_syms)
    F << x.first << " ";
  F << endl;
   
  F << endl;
}

// OpenMP parallelization support
void init_openMP()
{
#ifdef _OPENMP
   cout << "[OMP] Max. number of threads: " << omp_get_max_threads() << endl;
   cout << "[OMP] Number of processors: "   << omp_get_num_procs() << endl;
   cout << "[OMP] Dynamic thread adjustment: "   << omp_get_dynamic() << endl;
   cout << "[OMP] Nested parallelism: "     << omp_get_nested() << endl;
   cout << endl;
#endif
}

// Called after the symmetry type is determined from the data file.
void set_symmetry(const string &sym_string)
{
   my_assert(Sym == 0); // only once!
   if (all_syms.count(sym_string) != 1) {
     cout << "Unknown symmetry " << sym_string << endl;
     exit(1);
   }
   Sym = all_syms[sym_string];
   my_assert(Sym != 0);
   Sym->init();
   
   testInvar(Sym->InvarSinglet);
   cout << "singlet QNs=(" << Sym->InvarSinglet << ") mult=" << Sym->mult(Sym->InvarSinglet) << endl;
}

// Parse the header of the data file and 1) check datafile version;
// 2) determine the symmetry type.
void parse_datafile_header(istream &fdata, 
			   int expected_version = 8) // version number
{
   int dataversion = -1;
   while (fdata.peek() == '#') {
      fdata.ignore(); // ignore '#'
      if (fdata.peek() == '!') {
	 fdata.ignore(); // ignore '!'
	 fdata >> dataversion;
	 cout << "V" << dataversion << endl;
      } else {
	 string line;
	 getline(fdata, line);
	 string::size_type pos = line.find("symtype", 1);
	 if (pos == string::npos) {
	    // Comment line
	    cout << ">> " << line << endl;
	 } else {
	    // Symmetry type declaration
	    string::size_type p = line.find_last_of(" \t");
	    if (p != string::npos && p < line.size()-1) {
               sym_string = line.substr(p+1); // global variable
	       cout << "SYMMETRY TYPE: " << sym_string << endl;
	       if (!Sym) 
	         set_symmetry(sym_string); 
	    }
	 }
      }
      if (fdata.peek() == '\n') fdata.ignore();
   }
   my_assert(dataversion == expected_version);
}

// Read all initial energies and matrix elements
void read_data()
{
  cout << endl;

  ifstream fdata("data");
  if (!fdata) 
    my_error("Can't load initial data.");

  parse_datafile_header(fdata);
  read_nr_channels(fdata);
  read_nr_dots(fdata);
  read_Nmax(fdata);

  int nsubs; // Number of invariant subspaces
  fdata >> nsubs;
  my_assert(nsubs > 0);
  
  cout << "channels=" << P::channels << " impurities=" << P::dots
       << " Nmax=" << P::Nmax << " subspaces=" << nsubs << endl;
   
  // Note: we are reading diagprev, not diag, since this is
  // information from the previous (0-th) NRG step

  skip_comments(fdata);
  read_energies(fdata, diagprev, nsubs);

  skip_comments(fdata);
  read_ireducf(fdata, diagprev);

  while (1) {
    /* skip white space */
    while (!fdata.eof() && isspace(fdata.peek())) fdata.get();
    if (fdata.eof()) break;

    char ch = fdata.get();
    string opname;
    getline(fdata, opname);
    if (ch != '#') {
       debug("Reading <||" << opname << "||> (" << ch << ")");
    }
    switch (ch) {
     case '#':
       // ignore embedded comment lines
       break;

     case 'e':
       read_gs_energy(fdata);
       break;
       
     case 's':
       read_matrix_elements(fdata, a.ops[opname], diagprev);
       break;
       
     case 'p':
       read_matrix_elements(fdata, a.opsp[opname], diagprev);
       break;

     case 'g':
       read_matrix_elements(fdata, a.opsg[opname], diagprev);
       break;

     case 'd':
       read_matrix_elements(fdata, a.opd[opname], diagprev);
       break;

     case 't':
       read_matrix_elements(fdata, a.opt[opname], diagprev);
       break;
       
     case 'o':
       read_matrix_elements(fdata, a.opot[opname], diagprev);
       break;

     case 'q':
       read_matrix_elements(fdata, a.opq[opname], diagprev);
       break;
     
     case 'z':
       // Really read just in the first run!
       xi.read(fdata, STAT::runtype == RUN_NORMALNRG);
       zeta.read(fdata, STAT::runtype == RUN_NORMALNRG);
       break;

     case 'Z':
       delta.read(fdata, STAT::runtype == RUN_NORMALNRG);
       kappa.read(fdata, STAT::runtype == RUN_NORMALNRG);
       break;
       
     case 'X':
       xiR.read(fdata, STAT::runtype == RUN_NORMALNRG);
       zetaR.read(fdata, STAT::runtype == RUN_NORMALNRG);
       break;

     case 'T':
       ep.read(fdata);
       em.read(fdata);
       u0p.read(fdata);
       u0m.read(fdata);
       break;

     default:
       my_error("Unknown block %c in data file.", ch);
    }
  }
}

// Misc checks for validity of the input parameters and data file.
void check_validity()
{
   if (P::substeps)
     my_assert(sym_string == "QS" || sym_string == "QSZ" || 
	       sym_string == "SPSU2" || sym_string == "SPU1");
}

// Determine Nmax from the length of the coefficient tables! Modify
// it for substeps==true. Call after tridiagonalization routines (if
// not using the tables computed by initial.m).
void determine_Nmax()
{
  int length_coef_table = xi.max(0);
  my_assert(length_coef_table == P::Nmax);

  if (P::substeps)
     P::Nmax = P::channels * length_coef_table;

  cout << "length_coef_table=" << length_coef_table << " Nmax=" << P::Nmax << endl;
}

void start_calculation()
{
  nrglog('@', "@ start_calculation()");

  STAT::runtype = RUN_NORMALNRG;
  a.cleanup();
  read_data(); // read input data for NRG iteration
  check_validity();
   
  if (string(P::tri) == "cpp") {
#ifdef TRIDIAG
    tridiag();
#else
    my_error("Tridiagonalisation code not compiled in.");
#endif // TRIDIAG
  }
  
  cout << endl;
  determine_Nmax();
  cout << endl;

  // In and QN are global variables. They contain information about
  // how the invariant subspaces at consecutive iteration steps are
  // combined.
  In.resize(P::combs + 1);
  QN.resize(P::combs + 1);
  Sym->set_channels(P::channels);
  Sym->set_substeps(P::substeps);
  Sym->set_spin(P::spin);
  // This call actually initializes the data structures.
  Sym->load();
  if (logletter('Q')) {
     for (int i = 1; i <= P::combs; i++) {
	cout << "In[" << i << "]=(" << In[i] << ")" << endl;
     }
     for (int i = 1; i <= P::combs; i++) {
	cout << "QN[" << i << "]=(" << QN[i] << ")" << endl;
     }
  }

  dm = AllSteps(P::Nmax);
  
  // Perform first run: calculate everything and store unitary matrices (if
  // so desired) for calculating density matrix for DM-NRG.
  start_run();
  finalize_nrg();
    
  if (string(P::stopafter) == "nrg") {
     cout << endl << "*** Stop after the first NRG sweep requested." << endl;
     exit(1);
  }
   
  // Do the density-matrix NRG run.
  if (P::dm) {
    calc_densitymatrix(rho);
    if (P::fdm)
      calc_fulldensitymatrix(rhoFDM);

    if (string(P::stopafter) == "rho") {
       cout << endl << "*** Stop after DM calculation requested." << endl;
       exit(1);
    }
     
    STAT::runtype = RUN_DMNRG; // before read_data is called!
    a.cleanup();
    read_data();

    if (P::fdmexpv)
       init_fdmexpv();
     
    start_run();
    finalize_dmnrg();
  }
}
  
// Copy parameters to an object that is synchronised accross all
// processes. Warning: init() has to be called at the beginning of
// the program (after parsing the parameters in P), but also before
// each series of diagonalizations, because NRG::diagratio might have
// changed!
void sharedparam::init()
{
   diagroutine = P::diagroutine;
   diagratio = NRG::diagratio;
   dsyevrlimit = P::dsyevrlimit;
   logall = P::logall;
   log = P::log;
}
   
#ifdef NRG_MPI
int mpidebuglevel = 0;

void mpidebug(string str)
{
   if (mpidebuglevel > 0)
     cout << "MPI process " << myrank << " " << str << endl;
}

void mpi_sync_params()
{
   // Synchronize global parameters
   if (myrank == 0) {
      sP.init();
      for (int i = 1; i < mpiw->size(); i++) {
	 mpiw->send(i, TAG_SYNC, 0);
      }
   }   
   
   mpi::broadcast(*mpiw, sP, 0);
}
#endif

#ifdef MKL
#include <mkl_service.h>
#include <mkl_types.h>
#endif 

// What is the last iteration completed in the previous NRG runs?
void init_laststored()
{
   if (P::resume) {
      P::laststored = -1;
      for (int N = P::Ninit; N < P::Nmax; N++) {
	 const string fn = unitaryfn(N);
	 ifstream F(fn.c_str()); // open file
	 if (F.good()) { // successful?
	    P::laststored = N;
	 }
      }
      cout << "Last unitary file found: " << P::laststored << endl;
   }
}

// main() for the master process
int main_master(int argc, char *argv[])
{
  // Master process does most of the i/o and passes calculations to
  // the slaves.

  print_about_message(cout);

#ifdef NRG_MPI
   cout << "Parallelization using MPI: Running on " << mpiw->size() << " processors." << endl << endl;
#else
   cout << "No MPI: single node calculation." << endl << endl;
#endif

   init_openMP();

#ifdef MKL
   MKLVersion version;
   mkl_get_version(&version);
   cout << "Using Intel MKL library " << 
     version.MajorVersion << "." << version.MinorVersion << "." << version.UpdateVersion << endl;
   cout << "Processor optimization: " << version.Processor << endl;
   int max_threads = mkl_get_max_threads();
   
// Portability hack
# ifdef MKL_DOMAIN_BLAS
   #define NRG_MKL_DOMAIN_BLAS MKL_DOMAIN_BLAS
#else
   #define NRG_MKL_DOMAIN_BLAS MKL_BLAS
#endif
   
   int blas_max_threads = mkl_domain_get_max_threads(NRG_MKL_DOMAIN_BLAS);
   int dynamic = mkl_get_dynamic();
   cout << "max_threads=" << max_threads <<
     " blas_max_threads=" << blas_max_threads << 
     " dynamic=" << dynamic << endl << endl;
#endif
   
  set_new_handler(outOfMemory);
  cout << setprecision(COUT_PRECISION);
  parse_cmnd_line_read_parameters(argc, argv);
  validateparameters();
  calculate_invariants();
  init_laststored();
  dump_parameters();
  sP.init(); // copy parameters for MPI

#ifdef NRG_MPI
  for (int i = 1; i < mpiw->size(); i++) {
     mpiw->send(i, TAG_HELLO, 0);
  }
#endif
   
  start_calculation();

#ifdef NRG_MPI
  cout << "Master done. Terminating slave processes." << endl;
  for (int i = 1; i < mpiw->size(); i++) {
     mpiw->send(i, TAG_EXIT, 0);
  }
  cout << "Master exiting." << endl;
#endif
   
  if (P::done) {
     // Indicate completion by creating a flag file 'DONE'.
     ofstream D("DONE");
     if (!D) {
	cerr << "Can't create DONE." << endl;
     }
  }
     
  return 0;
}

#ifdef NRG_MPI
// Handle a diagonalisation request:
void slave_diag()
{
   mpi::status status;

   const int MASTER = 0;
   
   // 1. receive the matrix and the subspace identification
   mpidebug("recv");
   Matrix m;
   mpi_receive_matrix(MASTER, m);
   Invar I;
   status = mpiw->recv(MASTER, TAG_INVAR, I);
   check_status(status);
   
   // 2. preform the diagonalisation
   mpidebug("diag_h");
   Eigen eig;
   diag_h(m, eig);
   
   // 3. send back the results
   mpidebug("send");
   mpi_send_eigen(MASTER, eig);
   mpiw->send(MASTER, TAG_INVAR, I);
}

// main() for the slave processes
int main_slave(int argc, char *argv[])
{
  set_new_handler(outOfMemory);
   
  cout << "MPI slave rank " << myrank << endl;

  const int MASTER = 0;
   
  bool done = false;
  while (!done) {
     boost::optional<mpi::status> probestatus;
     if (probestatus = mpiw->iprobe(MASTER, mpi::any_tag)) {
	// A message can be received.
	int task;
	mpi::status status;
	status = mpiw->recv(MASTER, mpi::any_tag, task);
	check_status(status);
	
	if (logletter('M'))
	  cout << "Slave " << myrank << " received message with tag " << status.tag() << endl;

	switch (status.tag()) {
	 case TAG_HELLO:
	   mpidebug("ready");
	   break;
	 case TAG_EXIT:
	   mpidebug("exiting");
	   done = true;
	   break;
	 case TAG_DIAG:
	   mpidebug("diag");
	   slave_diag();
	   break;
	 case TAG_SYNC:
	   mpidebug("sync");
	   mpi_sync_params();
	   break;
	 default:
	   cerr << "MPI error: unknown tag on " << myrank << endl;
	   break;
	} // switch
     } else {
	// No message received. We sleep for a while to reduce the
	// load on the computer. (OpenMPI "feature" workaround)
	usleep(100);
     }
  } // while(!done)
   
  return 0;
}
#endif

int main(int argc, char *argv[])
{
  std::cerr.rdbuf(std::cout.rdbuf()); // alias cerr to cout
   
#ifdef NRG_MPI
  mpi::environment env(argc, argv);
  mpi::communicator world;

  mpienv = &env;
  mpiw = &world;
  myrank = mpiw->rank();

  if (myrank == 0) {
     return main_master(argc, argv);
  } else {
     return main_slave(argc, argv);
  }
#else
  return main_master(argc, argv);
#endif  // NRG_MPI
}
