// portabil.h - Portability code. Low-level stuff belongs here.
// Copyright (C) 2005-2015 Rok Zitko

#ifndef _portabil_h_
#define _portabil_h_

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <cmath>
#include <string>
#include <complex>

using namespace std;

// Disable some annoying compiler warnings
#define BE_QUIET

#if defined(BE_QUIET) && defined(__INTEL_COMPILER)
#pragma warning(disable:191) // type qualifier is meaningless on cast type
#pragma warning(disable:810) // ... may loose significant bits
#pragma warning(disable:383) // value copied to temporary, reference to temporary used
#pragma warning(disable:1572) // floating-point comparisons are unreliable
#pragma warning(disable:3500) // field initializers are a C++11 feature
#endif

// Autotools definitions, generated by "configure" script
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// Square raises x to the power of two
#ifndef HAVE_SQR
#define sqr(x) ((x)*(x))
#endif

// Backtrace. The code needs to be linked using the -rdynamic flag on
// systems using GNU ld to make the function names available to the
// program.
#ifdef HAVE_EXECINFO_H
 #include <execinfo.h>

 #include <cstdio>
 #include <cstdlib>
#endif
void print_trace(void)
{
#ifdef HAVE_EXECINFO_H
   const unsigned int max_trace_len = 100;

   void *array[max_trace_len];
   size_t size;
   char **strings;
   size_t i;
        
   size = backtrace (array, max_trace_len);
   strings = backtrace_symbols (array, size);
        
   printf ("Obtained %zd stack frames.\n", size);
   
   for (i = 0; i < size; i++)
     printf ("%s\n", strings[i]);
   
   free (strings);
#endif
}

#define assert_isfinite(x) finite_test_fnc(x, __FILE__, __LINE__)

bool my_isfinite(double x)
{
   return isfinite(x);
}

typedef complex<double> cmpl;

bool my_isfinite(cmpl z)
{
   return isfinite(z.real()) && isfinite(z.imag());
}

inline int isfinite(cmpl z)
{
   return (isfinite(z.real()) && isfinite(z.imag()) ? 1 : 0);
}

template <typename T>
T finite_test_fnc(T x, const char *file, int line)
{
   if (!my_isfinite(x)) {
      cerr << "Finiteness assertion failed." << endl;
      cerr << "File " << file << ", line " << line << "." << endl;
      cerr << "Exiting." << endl; 
      cout << "#### EXITING DUE TO FAILED ASSERTION." << endl;
      print_trace();
      exit(1);
   }
   return x;
}

#define my_assert(x) do { \
   if (!(x)) { \
      cerr << "Assertion failed." << endl; \
      cerr << "File " << __FILE__ << ", line " << __LINE__ << "." << endl; \
      const char sAssert[] = #x; \
      cerr << sAssert << endl; \
      cerr << "Exiting." << endl; \
      cout << "#### EXITING DUE TO FAILED ASSERTION." << endl; \
      print_trace(); \
      exit(1); \
   } \
} while(0)

// Assert a==b
#define my_assert_equal(a, b) do { \
  if (!(a==b)) { \
      cerr << "Assertion of equality failed." << endl; \
      cerr << "File " << __FILE__ << ", line " << __LINE__ << "." << endl; \
      const char sAssert_a[] = #a; \
      const char sAssert_b[] = #b; \
      cerr << sAssert_a << "=" << a << endl; \
      cerr << sAssert_b << "=" << b << endl; \
      cerr << "Exiting." << endl; \
      cout << "#### EXITING DUE TO FAILED ASSERTION." << endl; \
      print_trace(); \
      exit(1); \
   } \
} while(0)

#define my_warning(...) do { \
   cerr << "Warning." << endl; \
   cerr << "File " << __FILE__ << ", line " << __LINE__ << "." << endl; \
   char buf[256]; \
   sprintf(buf, __VA_ARGS__); \
   cerr << buf << endl; \
   cerr << "Continuing." << endl; \
} while(0)

#define my_error(...)  do { \
   cerr << "Error." << endl; \
   cerr << "File " << __FILE__ << ", line " << __LINE__ << "." << endl; \
   char buf[256]; \
   sprintf(buf, __VA_ARGS__); \
   cerr << buf << endl; \
   cerr << "Exiting" << endl; \
   cout << "#### EXITING DUE TO ERROR." << endl; \
   print_trace(); \
   exit(1); \
} while(0)

#define my_assert_not_reached() do { \
   cerr << "Should never be reached." << endl; \
   cerr << "File " << __FILE__ << ", line " << __LINE__ << "." << endl; \
   cerr << "Exiting." << endl; \
   cout << "#### EXITING DUE TO FAILED ASSERTION." << endl; \
   print_trace(); \
   exit(1); \
} while(0)

// Workaround
#ifndef __TIMESTAMP__
#define __TIMESTAMP__ ""
#endif

// Misc. macros

#ifndef ARRAYLENGTH
# define ARRAYLENGTH(x) (sizeof(x) / sizeof(*(x)))
#endif

#ifndef MIN
# define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

#ifndef MAX
# define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

#ifndef CLIP
# define CLIP(x, xmin, xmax) MIN(MAX(x, xmin), xmax)
#endif

// *** Timing

#include <time.h>                                                                  
#include <sys/time.h>

// Return time in seconds                                                          
double gettime (void)                                                              
{                                                                                  
#ifdef HAVE_LIBRT                                                                  
// Low-level timing code pinched from glib-2.18.2                                  
// Other choices: CLOCK_MONOTONIC, CLOCK_PROCESS_CPUTIME_ID                        
#define USE_CLOCK CLOCK_REALTIME                                                   
#define NSEC_PER_SEC 1000000000                                                    
                                                                                   
   struct timespec tv;                                                             
                                                                                   
   clock_gettime(USE_CLOCK, &tv);                                                  
                                                                                   
   return tv.tv_sec + (double)tv.tv_nsec / NSEC_PER_SEC;                           
#else                                                                              
   struct timeval tv;
   
   int res = gettimeofday(&tv, NULL);
   const double MILLION = 1.0e6;
   return (double)(tv.tv_sec+tv.tv_usec/MILLION);
#endif                                                                             
}               

// *** Memory usage

#if defined(__APPLE__) && defined(__MACH__)
#include <mach/mach_init.h>
#include <mach/task.h>

// Source: http://blog.kuriositaet.de/?p=257
int getmem (unsigned int *rss, unsigned int *vs)
{
    task_t task = MACH_PORT_NULL;
    struct task_basic_info t_info;
    mach_msg_type_number_t t_info_count = TASK_BASIC_INFO_COUNT;

    if (KERN_SUCCESS != task_info(mach_task_self(),
       TASK_BASIC_INFO, (task_info_t)&t_info, &t_info_count))
    {
        return -1;
    }
    *rss = t_info.resident_size;
    *vs  = t_info.virtual_size;
    return 0;
}

int memoryused() {
  unsigned int rss;
  unsigned int vs;
  
  getmem(&rss, &vs);
  
  return rss/1024; // result in kB block of memory!
}
#define HAS_MEMORY_USAGE
#endif

#if defined(__linux)
#include <string.h>
#include <stdio.h>
// Returns the number of kB blocks of memory used by this process (self).        
int memoryused() {                                                         
   int used = 0;
   
   // See /usr/src/linux/Documentation/filesystems/proc.txt                       
   if (FILE *info = fopen("/proc/self/status", "r")) {                            
     char buf[512];                                                               
     int val;                                                                     
                                                                                   
     while (fgets(buf, sizeof(buf), info)) {                                      
       if (!strncmp(buf, "VmSize:", 7) && sscanf(buf+7, "%d", &val))              
         used = val;                                                              
     }                                                                            
     fclose(info);                                                                
   }
   
   return used;
}
#define HAS_MEMORY_USAGE                                                           
#endif                                                                             
                                                                                   
// Fall-back
#if !defined(HAS_MEMORY_USAGE)
int memoryused() { return 0; }
#endif

// More considerate way of opening output files: 
// - check if we really want to open the file (non-empty filename)
// - check for open errors
// - output the header
void safeopen(ofstream &F, string filename, bool binary = false)
{
   if (filename == "") 
     return;
   
   ios::openmode flag = ios::out;
   if (binary)
     flag |= ios::binary;
   
   F.open(filename.c_str(), flag);
   if (!F) 
     my_error("safeopen: Can't open %s", filename.c_str());
}

// Open a file for reading (with error checking)
void isafeopen(ifstream &F, string filename)
{
   if (filename == "")
     return;
   
   F.open(filename.c_str());
   if (!F)
     my_error("isafeopen: Can't open %s", filename.c_str());
}

#endif
