#ifndef _spec_cc_
#define _spec_cc_

// Calculate (finite temperature) spectral function 1/Pi Im << op1^\dag(t)
// op2(0) >>. Required spin direction is determined by 'SPIN'. For SPIN=0
// both spin direction are equivalent. For QSZ, we need to differentiate
// the two.

template <class FactorFnc, class CheckSpinFnc>
void calc_generic(const MatrixElements &op1, 
		      const MatrixElements &op2,
		      SpectrumPtr s, 
		      const DiagInfo &diag, 
		      SPECTYPE spectype, 
		      int SPIN,
		      double sign, // 1 for bosonic, -1 for fermionic operators
		      FactorFnc factorfnc, // functors, see above!
		      CheckSpinFnc checkspinfnc)
{
  ChainSpectrum * cs;
  if (real_spectrum(spectype)) {
    cs = new ChainSpectrumBinning;
  } else if (temp_spectrum(spectype)) {
    cs = new ChainSpectrumTemp;
  } else {
    my_error("should not be reached");
  }

  std::vector<Twoinvar> tasks;

  // Strategy: we loop through all subspace pairs and check whether
  // they have non-zero irreducible matrix elements.
  
  LOOP_const(diag, isp) {
    const Invar Ip = INVAR(isp);
    
    LOOP_const(diag, is1) {
      const Invar I1 = INVAR(is1);

      const Twoinvar II = make_pair(I1, Ip);
      if (op1.count(II)) {
        bool contributes = checkspinfnc(I1, Ip, SPIN);
        if (!contributes)
          continue;
        tasks.push_back(II);
      }
    }
  }

  const unsigned int nrtasks = tasks.size();
  if (nrtasks == 0)
     cout << "WARNING: nrtasks=0 in calc_generic() for " << s->name() << endl;
   
  for (unsigned int itask = 0; itask < nrtasks; itask++) {
    const Twoinvar II = tasks[itask];
    Invar I1, Ip;
    tie(I1, Ip) = II;

    const Eigen & diagIp = diag.find(Ip)->second;
    const Eigen & diagI1 = diag.find(I1)->second;

    const unsigned int dimp = get_loop_dim(spectype, diagIp, Ip);
    const unsigned int dim1 = get_loop_dim(spectype, diagI1, I1);

    my_assert(op2.count(II) == 1);
      
    // Note that rho is always "diagonal in subspace indexes", i.e.
    // rho(I1,I2)=delta_(I1,I2) rho_I1. There is thus only one multiplicity
    // factor in play here.
    const t_factor spinfactor = factorfnc(Ip, I1);
    my_assert(!num_equal(spinfactor, 0.0)); // bug trap    

    const Matrix & op1II = op1.find(II)->second;
    const Matrix & op2II = op2.find(II)->second;

    if (logletter('g')) {
       cout << "calc_generic: " << s->name() << " " << spectype << " ";
       nrgdump2(I1, Ip) << endl;
    }

    switch (spectype) {
    case SPEC_FT:
      calc_generic_FT(diagIp, diagI1, op1II, op2II, 
                      sign, spinfactor, cs);
      break;

    case SPEC_DMNRG:
      calc_generic_DMNRG(diagIp, diagI1, dimp, dim1, op1II, op2II, 
                         sign, spinfactor, cs, Ip, I1);
      break;

    case SPEC_CFSold:
      calc_generic_CFSold(diagIp, diagI1, dimp, dim1, op1II, op2II, 
                       sign, spinfactor, cs, Ip, I1);
      break;

    case SPEC_CFS:
      calc_generic_CFSgt(diagIp, diagI1, dimp, dim1, op1II, op2II, 
                       sign, spinfactor, cs, Ip, I1);
      calc_generic_CFSls(diagIp, diagI1, dimp, dim1, op1II, op2II, 
                       sign, spinfactor, cs, Ip, I1);
      break;

    case SPEC_CFSgt:
      calc_generic_CFSgt(diagIp, diagI1, dimp, dim1, op1II, op2II, 
                       sign, spinfactor, cs, Ip, I1);
      break;

    case SPEC_CFSls:
      calc_generic_CFSls(diagIp, diagI1, dimp, dim1, op1II, op2II, 
                       sign, spinfactor, cs, Ip, I1);
      break;

    case SPEC_FDMold:
      calc_generic_FDMold(diagIp, diagI1, dimp, dim1, op1II, op2II, 
                       sign, spinfactor, cs, Ip, I1);
      break;

    case SPEC_FDM:
      calc_generic_FDMgt(diagIp, diagI1, dimp, dim1, op1II, op2II, 
                       sign, spinfactor, cs, Ip, I1);
      calc_generic_FDMls(diagIp, diagI1, dimp, dim1, op1II, op2II, 
                       sign, spinfactor, cs, Ip, I1);
      break;

    case SPEC_FDMgt:
      calc_generic_FDMgt(diagIp, diagI1, dimp, dim1, op1II, op2II, 
                       sign, spinfactor, cs, Ip, I1);
      break;

    case SPEC_FDMls:
      calc_generic_FDMls(diagIp, diagI1, dimp, dim1, op1II, op2II, 
                       sign, spinfactor, cs, Ip, I1);
      break;

    case SPEC_GT:
      calc_generic_GT(diagIp, diagI1, op1II, op2II,
                      sign, spinfactor, cs, Ip, I1, 0);
      break;

    case SPEC_I1T:
      calc_generic_GT(diagIp, diagI1, op1II, op2II, 
                      sign, spinfactor, cs, Ip, I1, 1);
      break;

    case SPEC_I2T:
      calc_generic_GT(diagIp, diagI1, op1II, op2II, 
                      sign, spinfactor, cs, Ip, I1, 2);
      break;

    case SPEC_CHIT:
      calc_generic_CHIT(diagIp, diagI1, op1II, op2II, 
                        sign, spinfactor, cs, Ip, I1);
      break;

    default:
      my_assert_not_reached();
    }
  }

  if (temp_spectrum(spectype)) {
     // Report on the temperature-dependent quantities.
     // Note that T of the data point is not exactly equal to current Teff
     cout << spectypestring(spectype) << "[" << s->name() << "]=" 
          << cs->total_weight() << endl;
  }
  
  if (logletter('*')) 
     cout << "Merging " << s->opname << " " << s->spectype << endl;
  s->merge(cs);
}

// Wrappers for spectral function calculations. They differ in the
// FactorFnc() and CheckSpinFnc() functors. The first defines the spin
// factor for a pair of Invar arguments, the second checks if the
// combinations of subspaces contributes to the given spectral
// functions.
void calc_specdens_wrap(BaseSpectrum &sc, DiagInfo &diag)
{
   calc_generic(*(sc.op1), *(sc.op2), sc.spec, diag, sc.spectype, sc.spin,
		S_FERMIONIC, SpecdensFactorFnc(), SpecdensCheckSpinFnc());
}

void calc_specdensquad_wrap(BaseSpectrum &sc, DiagInfo &diag)
{
   calc_generic(*(sc.op1), *(sc.op2), sc.spec, diag, sc.spectype, sc.spin, // support J_z ?
		S_FERMIONIC, SpecdensquadFactorFnc(), trivialCheckSpinFnc());
		//SpecdensquadCheckSpinFnc());
}

void calc_correlator_wrap(BaseSpectrum &sc, DiagInfo &diag)
{
   calc_generic(*(sc.op1), *(sc.op2), sc.spec, diag, sc.spectype, 0,
		S_BOSONIC, CorrelatorFactorFnc(), trivialCheckSpinFnc());
}

void calc_spinsusc_wrap(BaseSpectrum &sc, DiagInfo &diag)
{
   calc_generic(*(sc.op1), *(sc.op2), sc.spec, diag, sc.spectype, 0,
		S_BOSONIC, SpinSuscFactorFnc(), trivialCheckSpinFnc());
}

void calc_orbsusc_wrap(BaseSpectrum &sc, DiagInfo &diag)
{
   calc_generic(*(sc.op1), *(sc.op2), sc.spec, diag, sc.spectype, 0,
		S_BOSONIC, OrbSuscFactorFnc(), trivialCheckSpinFnc());
}

#endif // _spec_cc_
