void calc_generic_FDMold(const Eigen & diagIp, const Eigen & diagI1, 
		      unsigned int dimp, unsigned int dim1,
		      const Matrix & op1II, const Matrix & op2II,
		      double sign, t_factor spinfactor,
		      ChainSpectrum * cs,
		      const Invar &Ip, const Invar &I1)
{
   using namespace STAT;

   const double wn = STAT::wn[N];
   
   double wnfactor = wn/ZnD[N];
   if (isnan(wnfactor)) {
     wnfactor = 0.0;
   }

   const Matrix & rhoNIp = rhoFDM[Ip]; // hand optimised out of the loops
   const Matrix & rhoNI1 = rhoFDM[I1];

   // Convention: k-loops over retained states, l-loops over
   // discarded states.
        
   my_assert(diagI1.getnr() == diagI1.getrmax()); // required for CFS and FDM
   my_assert(diagIp.getnr() == diagIp.getrmax());
   
   unsigned int ret1 = dim1;
   unsigned int retp = dimp;
   const unsigned int all1 = diagI1.getnr(); // all states
   const unsigned int allp = diagIp.getnr();
   
   if (LAST_ITERATION()) {
     // All states are considered as discarded in the last iteration
     // of the FDM. (Note the different semantics!!)
     ret1 = 0;
     retp = 0;
   }

   // p. 16 in Costi, Zlatic, PRB  81, 235127 (2010)
   // NOTE THAT THE ORDER OF THE INDEXES OF op2II IS REVERSED WRT THE CITED
   // PAPER.
   
   double T = P::T;
   
   // Contribution 1: sizeable at high temperatures
   for (unsigned int l1 = ret1; l1 < all1; l1++) { // D
      const t_eigen El = diagI1.absenergy(l1);
	 
      for (unsigned int lp = retp; lp < allp; lp++) { // D
	 const t_eigen Elp = diagIp.absenergy(lp);
	 
	 DELTA d;
	 d.energy = El - Elp;
	 
	 d.weight = spinfactor * CONJ_ME(op1II(l1, lp)) * op2II(l1, lp)
	   * ( exp(-El/T) - sign * exp(-Elp/T) );
	 cs->add(d.energy, wnfactor * d.weight);
      }
   }
  
   if (LAST_ITERATION()) {
     return; // No further contributions in the last iteration!
     // It is not strictly necessary to return at this point, since
     // retp=ret1=0 anyway, thus the loops do not execute..
   }
   
   // Contribution 2, positive frequency excitations. No contribution at very low
   // temperatures.
   for (unsigned int l1 = ret1; l1 < all1; l1++) { // D
      const t_eigen El = diagI1.absenergy(l1);
	 
      for (unsigned int kp = 0; kp < retp; kp++) { // K
	 const t_eigen Ek = diagIp.absenergy(kp);
	    
	 DELTA d;
	 d.energy = El - Ek;
	 my_assert(d.energy > 0); // always positive
	 d.weight = spinfactor * CONJ_ME(op1II(l1, kp)) * op2II(l1, kp)
	   * exp(-El/T);
	 cs->add(d.energy, wnfactor * d.weight);
      }      
   }
   
   // Contribution 3, negative frequency excitations. No contribution at very low
   // temperatures.
   for (unsigned int k1 = 0; k1 < ret1; k1++) { // K
      const t_eigen Ek = diagI1.absenergy(k1);
	 
      for (unsigned int lp = retp; lp < allp; lp++) { // D
	 const t_eigen El = diagIp.absenergy(lp);
	    
	 DELTA d;
	 d.energy = Ek - El;
	 my_assert(d.energy < 0); // always negative
	 d.weight = spinfactor * CONJ_ME(op1II(k1, lp)) * op2II(k1, lp)
	   * (-sign * exp(-El/T));
	 cs->add(d.energy, wnfactor * d.weight);
      }
   }
   
   // Contribution 4, negative frequency excitations. Always sizeable contribution,
   // both at low and high temperatures. Always negative.
   for (unsigned int rk = 0; rk < ret1; rk++) { // K
      const t_eigen Ek = diagI1.absenergy(rk);
      
      for (unsigned int rl = retp; rl < allp; rl++) { // D
	 const t_eigen El = diagIp.absenergy(rl);
	 
	 DELTA d;
	 d.energy = Ek - El;
	 my_assert(d.energy <= 0.0); // always negative!
	 weight_bucket sum;
	 for (unsigned int rkp = 0; rkp < ret1; rkp++) // K
	   sum += CONJ_ME( op1II(rkp, rl) ) * rhoNI1(rkp, rk);
	 
	 d.weight = (-sign) * spinfactor * t_weight(sum); // sign == +1
	 d.weight *= op2II(rk, rl);
	 cs->add(d.energy, d.weight);
      }
   }
      
   // Contribution 5, positive frequency excitations. Always sizeable contribution,
   // both at low and high temperatures. Always positive.
   for (unsigned int rl = ret1; rl < all1; rl++) { // D
      const t_eigen El = diagI1.absenergy(rl);
	 
      for (unsigned int rk = 0; rk < retp; rk++) { // K
	 const t_eigen Ek = diagIp.absenergy(rk);
	 
	 DELTA d;
	 d.energy = El - Ek;
	 my_assert(d.energy >= 0.0); // always positive!

	 weight_bucket sum;
	 for (unsigned int rkp = 0; rkp < retp; rkp++) // K, was: dimp
	   sum += op2II(rl, rkp) * rhoNIp(rkp, rk); // no sign here!

	 d.weight = spinfactor * CONJ_ME(op1II(rl, rk)) * t_weight(sum);
	 cs->add(d.energy, d.weight);
      }
   }
}
